/*------------------------------------------------------------------------
 * curses/termcap interface routines
 */
#include "tcapcommon.h"

/*------------------------------------------------------------------------
 * tcap_set_inp_nl() - set input nl options on/off
 */
int tcap_set_inp_nl (TERMINAL *sp, int bf)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_inp_nl(sp->term, bf);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_out_nl() - set output nl options on/off
 */
int tcap_set_out_nl (TERMINAL *sp, int bf)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_out_nl(sp->term, bf);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_raw() - set raw options on/off
 */
int tcap_set_raw (TERMINAL *sp, int bf)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_raw(sp->term, bf);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_cbreak() - set cbreak options on/off
 */
int tcap_set_cbreak (TERMINAL *sp, int bf)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_cbreak(sp->term, bf);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_echo() - set echo options on/off
 */
int tcap_set_echo (TERMINAL *sp, int bf)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_echo(sp->term, bf);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_qiflush() - set flush-on-interrupt options on/off
 */
int tcap_set_qiflush (TERMINAL *sp, int bf)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_qiflush(sp->term, bf);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_meta() - set 8-bit options on/off
 */
int tcap_set_meta (TERMINAL *sp, int bf)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_meta(sp->term, bf);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_get_curx() - get current x position
 */
int tcap_get_curx (const TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	return sp->scrn->curr_col;
}

/*------------------------------------------------------------------------
 * tcap_get_cury() - get current y position
 */
int tcap_get_cury (const TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	return sp->scrn->curr_row;
}

/*------------------------------------------------------------------------
 * tcap_set_curx() - set current x position
 */
int tcap_set_curx (TERMINAL *sp, int x)
{
	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		sp->scrn->curr_col = x;
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_set_cury() - set current y position
 */
int tcap_set_cury (TERMINAL *sp, int y)
{
	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		sp->scrn->curr_row = y;
	}
	TERMINAL_UNLOCK(sp);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_napms() - sleep for given milliseconds
 */
int tcap_napms (int ms)
{
	if (ms < 0)
		return (-1);

	term_nap(ms);

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_set_use_env() - specify whether to use LINES & COLS env vars
 */
void tcap_set_use_env (int bf)
{
	tcap_use_env_vars = bf;
}

/*------------------------------------------------------------------------
 * tcap_get_use_env() - get whether to use LINES & COLS env vars
 */
int tcap_get_use_env (void)
{
	return (tcap_use_env_vars);
}

/*------------------------------------------------------------------------
 * tcap_set_term_path() - set path to search for term files
 */
void tcap_set_term_path (const char *path)
{
	tcap_term_path = path;
}

/*------------------------------------------------------------------------
 * tcap_get_term_path() - get path to search for term files
 */
const char * tcap_get_term_path (void)
{
	return (tcap_term_path);
}
