/*------------------------------------------------------------------------
 * mode functions
 */
#include "tcapcommon.h"

/*------------------------------------------------------------------------
 * tcap_get_baud_rate() - get current baudrate
 */
int	tcap_get_baud_rate	(const TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	return term_get_baud_rate(sp->term);
}

/*------------------------------------------------------------------------
 * tcap_get_kill_char() - get current kill char
 */
int	tcap_get_kill_char	(const TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	return term_get_kill_char(sp->term);
}

/*------------------------------------------------------------------------
 * tcap_get_quit_char() - get current quit char
 */
int	tcap_get_quit_char	(const TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	return term_get_quit_char(sp->term);
}

/*------------------------------------------------------------------------
 * tcap_get_intr_char() - get current intr char
 */
int	tcap_get_intr_char	(const TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	return term_get_intr_char(sp->term);
}

/*------------------------------------------------------------------------
 * tcap_get_erase_char() - get current erase char
 */
int	tcap_get_erase_char	(const TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	return term_get_erase_char(sp->term);
}

/*------------------------------------------------------------------------
 * tcap_get_wait() - get current wait mode
 */
int	tcap_get_wait		(const TERMINAL *sp)
{
	if (sp == 0)
		return (-1);

	return term_get_wait(sp->term);
}

/*------------------------------------------------------------------------
 * tcap_set_baud_rate() - set current baud-rate
 */
int	tcap_set_baud_rate	(TERMINAL *sp, int br)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_baud_rate(sp->term, br);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_kill_char() - set current kill char
 */
int	tcap_set_kill_char	(TERMINAL *sp, int ch)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_kill_char(sp->term, ch);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_quit_char() - set current quit char
 */
int	tcap_set_quit_char	(TERMINAL *sp, int ch)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_quit_char(sp->term, ch);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_intr_char() - set current intr char
 */
int	tcap_set_intr_char	(TERMINAL *sp, int ch)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_intr_char(sp->term, ch);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_erase_char() - set current erase char
 */
int	tcap_set_erase_char	(TERMINAL *sp, int ch)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_erase_char(sp->term, ch);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_wait() - set current wait mode
 */
int	tcap_set_wait		(TERMINAL *sp, int bf)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_wait(sp->term, bf);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_flush_inp() - flush input queue
 */
int tcap_flush_inp (TERMINAL *sp)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_flush_inp(sp->term);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_flush_out() - flush output queue
 */
int tcap_flush_out (TERMINAL *sp)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_flush_out(sp->term);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}

/*------------------------------------------------------------------------
 * tcap_set_typeahead() - set typeahead processing
 */
int tcap_set_typeahead (TERMINAL *sp, int fildes)
{
	if (sp == 0)
		return (-1);

	/* NOT IMPLEMENTED */

	return (0);
}

/*------------------------------------------------------------------------
 * tcap_set_halfdelay() - set halfdelay mode
 */
int tcap_set_halfdelay (TERMINAL *sp, int tenths)
{
	int rc;

	if (sp == 0)
		return (-1);

	if (tenths < 1 || tenths > 255)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_set_read_mode(sp->term, tenths * 100);
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}
