/*------------------------------------------------------------------------
 * window size routines
 */
#include "tcapcommon.h"

/*------------------------------------------------------------------------
 * tcap_winsize_check() - check size of window
 */
int tcap_winsize_check (const TERMINAL *sp)
{
	int changed = FALSE;

	if (sp != 0)
	{
		int old_rows	= sp->tcap->ints.maxrows;
		int old_cols	= sp->tcap->ints.maxcols;
		int new_rows;
		int new_cols;

		if (term_window_size_get(sp->term, &new_rows, &new_cols) == 0)
		{
			sp->tcap->ints.maxrows = new_rows;
			sp->tcap->ints.maxcols = new_cols;
		}

		if (old_rows != sp->tcap->ints.maxrows ||
		    old_cols != sp->tcap->ints.maxcols)
		{
			changed = TRUE;
		}
	}

	return (changed);
}

/*------------------------------------------------------------------------
 * tcap_winsize_reset() - reset window size
 */
int tcap_winsize_reset (TERMINAL *sp, int rows, int cols)
{
	int rc;

	if (sp == 0)
		return (-1);

	TERMINAL_LOCK(sp);
	{
		rc = term_window_size_set(sp->term, rows, cols);
		if (rc == 0)
		{
			sp->tcap->ints.maxrows = rows;
			sp->tcap->ints.maxcols = cols;
			tcap_mouse_invalidate(sp);
		}
	}
	TERMINAL_UNLOCK(sp);

	return (rc);
}
