/*------------------------------------------------------------------------
 * definitions for box & line-drawing characters
 */
#ifndef	WBOX_H
#define	WBOX_H

/*------------------------------------------------------------------------
 * C++ stuff
 */
#ifdef __cplusplus
extern "C" {
#endif

/*@ +++API+++ */
/*------------------------------------------------------------------------
 * block letter size
 */
#define BLK_WIDTH	8		/* width  of a block letter */
#define BLK_HEIGHT	5		/* height of a block letter */

/*------------------------------------------------------------------------
 *	box types
 */
#define ASCII		0		/* ascii char boxes			*/
#define	SHSV		1		/* single horiz single vert	*/
#define	DHDV		2		/* double horiz double vert	*/
#define	DHSV		3		/* double horiz single vert	*/
#define	SHDV		4		/* single horiz double vert	*/
#define BLOCK		5		/* block chars				*/

/* I can never remember the order (HV / VH), so here's the other way */

#define	SVSH		SHSV	/* single vert single horiz	*/
#define	DVDH		DHDV	/* double vert double horiz	*/
#define	SVDH		DHSV	/* single vert double horiz	*/
#define	DVSH		SHDV	/* double vert single horiz	*/

/*------------------------------------------------------------------------
 * box char definitions
 *
 * A 3x3 box looks like this:
 *
 *	TL	TC	TR
 *	ML	MC	MR
 *	BL	BC	BR
 *
 * A 5x5 box looks like this:
 *
 *	TL	HO	TC	HO	TR
 *	VE		VE		VE
 *	ML	HO	MC	HO	MR
 *	VE		VE		VE
 *	BL	HO	BC	HO	BR
 */
#define	B_BL		0		/* bottom left		*/
#define B_BC		1		/* bottom center	*/
#define B_BR		2		/* bottom right		*/

#define B_ML		3		/* middle left		*/
#define B_MC		4		/* middle center	*/
#define B_MR		5		/* middle right		*/

#define B_TL		6		/* top    left		*/
#define B_TC		7		/* top    center	*/
#define B_TR		8		/* top    right		*/

#define B_VE		9		/* vert   line		*/
#define B_HO		10		/* horiz  line		*/

/*------------------------------------------------------------------------
 * box chars table
 */
extern const unsigned char box_chrs[6][11];

/*------------------------------------------------------------------------
 * ACS compatability definitions
 *
 * These definitions follow the VT100 naming convention
 */

/* box characters */

#define ACS_ULCORNER	(chtype)( box_chrs[SHSV][B_TL] | A_ALTCHARSET )
#define ACS_LLCORNER	(chtype)( box_chrs[SHSV][B_BL] | A_ALTCHARSET )
#define ACS_URCORNER	(chtype)( box_chrs[SHSV][B_TR] | A_ALTCHARSET )
#define ACS_LRCORNER	(chtype)( box_chrs[SHSV][B_BR] | A_ALTCHARSET )
#define ACS_RTEE		(chtype)( box_chrs[SHSV][B_MR] | A_ALTCHARSET )
#define ACS_LTEE		(chtype)( box_chrs[SHSV][B_ML] | A_ALTCHARSET )
#define ACS_BTEE		(chtype)( box_chrs[SHSV][B_BC] | A_ALTCHARSET )
#define ACS_TTEE		(chtype)( box_chrs[SHSV][B_TC] | A_ALTCHARSET )
#define ACS_PLUS		(chtype)( box_chrs[SHSV][B_MC] | A_ALTCHARSET )

/* line characters */

#define ACS_HLINE		(chtype)( box_chrs[SHSV][B_HO] | A_ALTCHARSET )
#define ACS_VLINE		(chtype)( box_chrs[SHSV][B_VE] | A_ALTCHARSET )

/* special characters */

#define ACS_BLOCK		(chtype)( 0xdb | A_ALTCHARSET )
#define ACS_DIAMOND		(chtype)( 0xb2 | A_ALTCHARSET )
#define ACS_CKBOARD		(chtype)( 0xb0 | A_ALTCHARSET )
#define ACS_DEGREE		(chtype)( 0xf8 | A_ALTCHARSET )
#define ACS_PLMINUS		(chtype)( 0xf1 | A_ALTCHARSET )
#define ACS_BULLET		(chtype)( 0xf9 | A_ALTCHARSET )
#define ACS_S1			(chtype)( '-'  | A_ALTCHARSET )
#define ACS_S3			(chtype)( '-'  | A_ALTCHARSET )
#define ACS_S7			(chtype)( '_'  | A_ALTCHARSET )
#define ACS_S9			(chtype)( '_'  | A_ALTCHARSET )
#define ACS_LE			(chtype)( 0xf3 | A_ALTCHARSET )
#define ACS_GE			(chtype)( 0xf2 | A_ALTCHARSET )
#define ACS_PI			(chtype)( 0xe3 | A_ALTCHARSET )

/* extended (non-VT100) characters */

#define ACS_LARROW		(chtype)( '<'  | A_ALTCHARSET )
#define ACS_RARROW		(chtype)( '>'  | A_ALTCHARSET )
#define ACS_UARROW		(chtype)( '^'  | A_ALTCHARSET )
#define ACS_DARROW		(chtype)( 'v'  | A_ALTCHARSET )
#define ACS_LANTERN		(chtype)( 0xe8 | A_ALTCHARSET )
#define ACS_BOARD		(chtype)( 0xb2 | A_ALTCHARSET )

/*@ ---API--- */
/*------------------------------------------------------------------------
 * C++ stuff
 */
#ifdef __cplusplus
}
#endif

#endif	/* WBOX_H */
