/*------------------------------------------------------------------------
 * screen dump routines
 */
#include "wincommon.h"

/*------------------------------------------------------------------------
 * scr_dump() - dump curscr to a file
 */
int scr_dump (const char *filename)
{
	FILE *	fp;
	int		rc;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (filename == 0 || *filename == 0)
		return (ERR);

	/*--------------------------------------------------------------------
	 * open the file
	 */
	fp = fopen(filename, "wb");
	if (fp == 0)
		return (ERR);

	/*--------------------------------------------------------------------
	 * save curscr
	 */
	rc = putwin(curscr, fp);

	/*--------------------------------------------------------------------
	 * close the file
	 */
	fclose(fp);

	return (rc);
}

/*------------------------------------------------------------------------
 * scr_restore() - read in curscr from a file
 */
int scr_restore (const char *filename)
{
	FILE *			fp;
	unsigned int	m;
	int				maxy;
	int				maxx;
	int				begy;
	int				begx;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (filename == 0 || *filename == 0)
		return (ERR);

	/*--------------------------------------------------------------------
	 * open the file
	 */
	fp = fopen(filename, "rb");
	if (fp == 0)
		return (ERR);

	/*--------------------------------------------------------------------
	 * check for magic number
	 */
	m = get_4byte(fp);
	if (m != WIN_MAGIC)
	{
		fclose(fp);
		return (ERR);
	}

	/*--------------------------------------------------------------------
	 * header info
	 */
	maxy = get_4byte(fp);
	maxx = get_4byte(fp);
	begy = get_4byte(fp);
	begx = get_4byte(fp);

	/*--------------------------------------------------------------------
	 * check if header matches curscr parameters
	 */
	if (maxy != getmaxy(curscr) || maxx != getmaxx(curscr))
	{
		fclose(fp);
		return (ERR);
	}

	/*--------------------------------------------------------------------
	 * read file into curscr
	 */
	win_load(curscr, fp);

	/*--------------------------------------------------------------------
	 * close the file
	 */
	fclose(fp);

	return (OK);
}

/*------------------------------------------------------------------------
 * scr_init() - read a file into stdscr
 */
int scr_init (const char *filename)
{
	FILE *			fp;
	unsigned int	m;
	int				maxy;
	int				maxx;
	int				begy;
	int				begx;

	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (filename == 0 || *filename == 0)
		return (ERR);

	/*--------------------------------------------------------------------
	 * open the file
	 */
	fp = fopen(filename, "rb");
	if (fp == 0)
		return (ERR);

	/*--------------------------------------------------------------------
	 * check for magic number
	 */
	m = get_4byte(fp);
	if (m != WIN_MAGIC)
	{
		fclose(fp);
		return (ERR);
	}

	/*--------------------------------------------------------------------
	 * header info
	 */
	maxy = get_4byte(fp);
	maxx = get_4byte(fp);
	begy = get_4byte(fp);
	begx = get_4byte(fp);

	/*--------------------------------------------------------------------
	 * check if header info matches stdscr parameters
	 */
	if (maxy != getmaxy(stdscr) || maxx != getmaxx(stdscr))
	{
		fclose(fp);
		return (ERR);
	}

	/*--------------------------------------------------------------------
	 * read in stdscr & clear curscr
	 */
	win_load(stdscr, fp);
	clearok(curscr, TRUE);

	/*--------------------------------------------------------------------
	 * close the file
	 */
	fclose(fp);

	return (OK);
}

/*------------------------------------------------------------------------
 * scr_set() - restore a file & believe it it what is currently displayed
 */
int scr_set (const char *filename)
{
	int rc;

	/*--------------------------------------------------------------------
	 * restore curscr
	 */
	rc = scr_restore(filename);
	if (rc)
		return (ERR);

	/*--------------------------------------------------------------------
	 * copy its contents to stdscr
	 */
	copywin(curscr, trmscr, 0, 0, 0, 0, getmaxy(curscr), getmaxx(curscr),
		TRUE);

	return (OK);
}
