/*------------------------------------------------------------------------
 * current screen routines
 */
#include "xvtcommon.h"

/*------------------------------------------------------------------------
 * set current-screen margins
 */
void xvt_curscrn_set_margins (TERMDATA *td, int top, int bot)
{
	xvt_scrn_set_margins(CUR_SCRN(td), top, bot);
}

/*------------------------------------------------------------------------
 * clear the current-screen
 */
void xvt_curscrn_clear (TERMDATA *td)
{
	xvt_scrn_clear(CUR_SCRN(td));
	xvt_term_disp_fill(td, 0, 0, CUR_SCRN_COLS(td), CUR_SCRN_ROWS(td));
}

/*------------------------------------------------------------------------
 * set the current-screen to all 'c'
 */
void xvt_curscrn_set (TERMDATA *td, int c)
{
	int x;
	int y;

	for (y=0; y<CUR_SCRN_ROWS(td); y++)
	{
		CELL *	cp	= CUR_SCRN_LINE(td, y);

		for (x=0; x<CUR_SCRN_COLS(td); x++)
		{
			cp->ch = (unsigned char)c;
			xvt_term_disp_char(td, x, y, cp->ch, cp->fg, cp->bg, cp->at);
			cp++;
		}
	}
}

/*------------------------------------------------------------------------
 * clear chars on a line
 *
 * mode:
 *	-1	clear from BOL to cursor
 *	 0	clear entire line
 *	 1	clear from cursor to EOL
 */
void xvt_curscrn_chars_clr (TERMDATA *td, int mode)
{
	CELL *	cp	= CUR_SCRN_LINE(td, CUR_SCRN_Y(td));
	int		j;
	int		sx;
	int		sy;
	int		dx;
	int		dy;

	switch (mode)
	{
	case 1:
		/*----------------------------------------------------------------
		 * clear to EOL
		 */
		sx = CUR_SCRN_X(td);
		dx = CUR_SCRN_COLS(td) - CUR_SCRN_X(td);
		sy = CUR_SCRN_Y(td);
		dy = 1;

		xvt_term_disp_fill(td, sx, sy, dx, dy);

		for (j=CUR_SCRN_X(td); j<CUR_SCRN_COLS(td); j++)
		{
			cp[j].ch = ' ';
			cp[j].fg = CUR_SCRN_FG(td);
			cp[j].bg = CUR_SCRN_BG(td);
			cp[j].at = NORMAL;
		}
		break;

	case -1:
		/*----------------------------------------------------------------
		 * clear from BOL
		 */
		sx = 0;
		dx = CUR_SCRN_X(td) + 1;
		sy = CUR_SCRN_Y(td);
		dy = 1;

		xvt_term_disp_fill(td, sx, sy, dx, dy);

		for (j=0; j<CUR_SCRN_X(td); j++)
		{
			cp[j].ch = ' ';
			cp[j].fg = CUR_SCRN_FG(td);
			cp[j].bg = CUR_SCRN_BG(td);
			cp[j].at = NORMAL;
		}
		break;

	case 0:
		/*----------------------------------------------------------------
		 * clear entire line
		 */
		xvt_curscrn_lines_clr(td, CUR_SCRN_Y(td), 1);
		break;
	}
}

/*------------------------------------------------------------------------
 * erase n chars from current point
 */
void xvt_curscrn_chars_era (TERMDATA *td, int n)
{
	int		sx	= CUR_SCRN_X(td);
	int		sy	= CUR_SCRN_Y(td);
	int		dx	= n;
	int		dy	= 1;
	CELL *	cp	= CUR_SCRN_CHAR(td, sy, sx);

	if ((sx + dx) > CUR_SCRN_COLS(td))
		dx = CUR_SCRN_COLS(td) - sx + 1;

	xvt_term_disp_fill(td, sx, sy, dx, dy);

	for (; sx <= dx; sx++)
	{
		cp->ch = ' ';
		cp->fg = CUR_SCRN_FG(td);
		cp->bg = CUR_SCRN_BG(td);
		cp->at = NORMAL;
		cp++;
	}
}

/*------------------------------------------------------------------------
 * delete n chars from current point
 */
void xvt_curscrn_chars_del (TERMDATA *td, int n)
{
	if (n == 0)
		n = 1;

	if (n > CUR_SCRN_X(td) + 1)
		n = CUR_SCRN_X(td) + 1;

	{
		CELL *	line = CUR_SCRN_LINE(td, CUR_SCRN_Y(td));
		int		x;

		/*----------------------------------------------------------------
		 * move line over n chars
		 */
		for (x = CUR_SCRN_X(td); x < CUR_SCRN_COLS(td) - n; x++)
		{
			line[x].ch = line[x+n].ch;
			line[x].fg = line[x+n].fg;
			line[x].bg = line[x+n].bg;
			line[x].at = line[x+n].at;
		}

		/*----------------------------------------------------------------
		 * set last n chars on line to blank
		 */
		for (; x < CUR_SCRN_COLS(td); x++)
		{
			line[x].ch = ' ';
			line[x].fg = CUR_SCRN_FG(td);
			line[x].bg = CUR_SCRN_BG(td);
			line[x].at = CUR_SCRN_AT(td);
		}

		/*----------------------------------------------------------------
		 * now redisplay all changed chars
		 */
		for (x=CUR_SCRN_X(td); x<CUR_SCRN_COLS(td); x++)
		{
			xvt_term_disp_char(td, x, CUR_SCRN_Y(td),
				line[x].ch,
				line[x].fg,
				line[x].bg,
				line[x].at);
		}
	}
}

/*------------------------------------------------------------------------
 * insert n chars at current point
 */
void xvt_curscrn_chars_ins (TERMDATA *td, int n)
{
	if (n == 0)
		n = 1;

	if (n > CUR_SCRN_COLS(td) - CUR_SCRN_X(td))
		n = CUR_SCRN_COLS(td) - CUR_SCRN_X(td);

	{
		CELL *	line = CUR_SCRN_LINE(td, CUR_SCRN_Y(td));
		int		x;

		/*----------------------------------------------------------------
		 * copy line over n chars
		 */
		for (x = CUR_SCRN_COLS(td) - n - 1; x >= CUR_SCRN_X(td); x--)
		{
			line[x+n].ch = line[x].ch;
			line[x+n].fg = line[x].fg;
			line[x+n].bg = line[x].bg;
			line[x+n].at = line[x].at;
		}

		/*----------------------------------------------------------------
		 * blank out n chars
		 */
		for (x = CUR_SCRN_X(td) + n - 1; x >= CUR_SCRN_X(td); x--)
		{
			line[x].ch = ' ';
			line[x].fg = CUR_SCRN_FG(td);
			line[x].bg = CUR_SCRN_BG(td);
			line[x].at = CUR_SCRN_AT(td);
		}

		/*----------------------------------------------------------------
		 * now display all changed chars
		 */
		for (x=CUR_SCRN_X(td); x<CUR_SCRN_COLS(td); x++)
		{
			xvt_term_disp_char(td, x, CUR_SCRN_Y(td),
				line[x].ch,
				line[x].fg,
				line[x].bg,
				line[x].at);
		}
	}
}

/*------------------------------------------------------------------------
 * clear n lines in the current-screen
 */
void xvt_curscrn_lines_clr (TERMDATA *td, int y, int n)
{
	xvt_scrn_clear_lines(CUR_SCRN(td), y, n);
	xvt_term_disp_fill(td, 0, y, CUR_SCRN_COLS(td), n);
}

/*------------------------------------------------------------------------
 * copy n lines in the current-screen
 */
void xvt_curscrn_lines_cpy (TERMDATA *td, int yt, int yf, int n)
{
	xvt_scrn_copy_lines(CUR_SCRN(td), yt, yf, n);
	xvt_term_disp_copy(td,
		0, yt,
		CUR_SCRN_COLS(td), n,
		0, yf);
}

/*------------------------------------------------------------------------
 * delete n lines from current line
 */
void xvt_curscrn_lines_del (TERMDATA *td, int n)
{
	int	top		= CUR_SCRN_TOP(td);
	int	bot		= CUR_SCRN_BOT(td);
	int cur		= CUR_SCRN_Y(td) - top;
	int length	= (bot - cur) + 1;

	/*--------------------------------------------------------------------
	 * no delete if out of margin-box
	 */
	if (CUR_SCRN_Y(td) < top || CUR_SCRN_Y(td) > bot)
		return;

	/*--------------------------------------------------------------------
	 * ignore negative & zero requests
	 */
	if (n <= 0)
		return;

	/*--------------------------------------------------------------------
	 * if n > lines in box below current line, just clear it
	 */
	if (n >= length)
	{
		xvt_curscrn_lines_clr(td, cur, length);
		return;
	}

	/*--------------------------------------------------------------------
	 * copy the lines up & clear the vacated lines
	 */
	if (length > n)
	{
		xvt_curscrn_lines_cpy(td, CUR_SCRN_Y(td), CUR_SCRN_Y(td) + n,
			length - n);
	}
	xvt_curscrn_lines_clr(td, bot - n + 1, n);
}

/*------------------------------------------------------------------------
 * insert n lines at current line
 */
void xvt_curscrn_lines_ins (TERMDATA *td, int n)
{
	int	top		= CUR_SCRN_TOP(td);
	int	bot		= CUR_SCRN_BOT(td);
	int cur		= CUR_SCRN_Y(td) - top;
	int	range	= (bot - top) + 1;
	int length	= (bot - cur) + 1;

	/*--------------------------------------------------------------------
	 * no insert if out of margin-box
	 */
	if (CUR_SCRN_Y(td) < top || CUR_SCRN_Y(td) > bot)
		return;

	/*--------------------------------------------------------------------
	 * ignore negative & zero requests
	 */
	if (n <= 0)
		return;

	/*--------------------------------------------------------------------
	 * copy the lines up & clear the vacated lines
	 */
	if (CUR_SCRN_Y(td) == bot)
	{
		xvt_curscrn_lines_cpy(td, top, top+1, range - 1);
		xvt_curscrn_lines_clr(td, CUR_SCRN_Y(td), 1);
	}
	else
	{
		/*----------------------------------------------------------------
		 * if n > lines in box above current line, just clear it
		 */
		if (n >= length)
		{
			xvt_curscrn_lines_clr(td, CUR_SCRN_Y(td), length);
		}
		else
		{
			if (length > n)
			{
				xvt_curscrn_lines_cpy(td, CUR_SCRN_Y(td) + n, CUR_SCRN_Y(td),
					length - n);
			}
			xvt_curscrn_lines_clr(td, CUR_SCRN_Y(td), n);
		}
	}
}

/*------------------------------------------------------------------------
 * scroll screen n lines
 */
void xvt_curscrn_scroll (TERMDATA *td, int n)
{
	int save_y	= CUR_SCRN_Y(td);

	if (n > 0)
	{
		/*----------------------------------------------------------------
		 * scroll down n lines
		 */
		CUR_SCRN_Y(td) = CUR_SCRN_TOP(td);
		xvt_curscrn_lines_ins(td, n);
	}
	else if (n < 0)
	{
		/*----------------------------------------------------------------
		 * scroll up n lines
		 */
		CUR_SCRN_Y(td) = CUR_SCRN_BOT(td);
		xvt_curscrn_lines_ins(td, -n);
	}

	CUR_SCRN_Y(td) = save_y;
}
