/*------------------------------------------------------------------------
 * header for window routines
 */
#ifndef XVTTERM_H
#define XVTTERM_H

/*------------------------------------------------------------------------
 * misc defines
 */
#define XVT_DIE_LOOPS		10				/* max loops to wait for child	*/
											/* to die before bailing		*/

/*------------------------------------------------------------------------
 * number of params supported
 */
#define ESC_NUM_PARAMS	9					/* terminfo supports 9			*/

/*------------------------------------------------------------------------
 * data length needed to hold an icon
 */
#define XVT_ICON_LEN	(XVT_MAX_BITS * XVT_MAX_BITS / 8)

/*------------------------------------------------------------------------
 * screen modes
 */
#define SCRN_MODE_MONO		0
#define SCRN_MODE_COLOR		1

/*------------------------------------------------------------------------
 * printing modes
 */
#define PRT_MODE_NONE		0				/* not printing					*/
#define PRT_MODE_NORM		1				/* printing: normal				*/
#define PRT_MODE_ESC		2				/* printing: in esc seq			*/

/*------------------------------------------------------------------------
 * char-set info
 */
#define CHAR_SET_NONE		-1
#define CHAR_SET_G0			0
#define CHAR_SET_G1			1
#define CHAR_SET_G2			2
#define CHAR_SET_G3			3

#define NUM_CHAR_SETS		4

/*------------------------------------------------------------------------
 * term-data escape-info struct
 */
struct escinfo
{
	int					reading_esc_seq;		/* T if reading esc-seq		*/

	char				esc_chars[128];			/* esc-seq buffer			*/
	int					esc_char_cnt;			/* # of chars in esc_chars	*/

	int					code;					/* current code				*/

	int					n[ESC_NUM_PARAMS];		/* esc-seq params			*/
	int					n_index;				/* cur index of n[]			*/

	char				data_chars[1024 + 4];	/* data chars				*/
	int					data_cnt;				/* # chars of data read		*/

	int					vt_mode;				/* processing mode			*/
	int					vt_type;				/* data type (GCS & STP)	*/
};
typedef struct escinfo ESCINFO;

/*------------------------------------------------------------------------
 * forward reference for cmd-tbl
 */
typedef struct cmd_tbl	CMD_TBL;

/*------------------------------------------------------------------------
 * term-data struct
 */
struct termdata
{
	/*--------------------------------------------------------------------
	 * initialized flag
	 */
	int					init;					/* TRUE if initialized		*/
	int					detached;				/* TRUE if detached			*/

	/*--------------------------------------------------------------------
	 * program info
	 */
	XVT_DATA *			xd;						/* ptr to user's xvt info	*/

	char				org_label[128];			/* original window label	*/
	char				win_label[128];			/* current  window label	*/
	char				ico_label[128];			/* current  icon   label	*/

	char				dsp_var_str[128];		/* DISPLAY env string		*/
	char				env_var_str[128];		/* env var string			*/
	char				trm_var_str[128];		/* TERM env string			*/
	char				wid_var_str[128];		/* WINDOWID env string		*/

	int					output_debug_seq;		/* T if last output was seq	*/

	/*--------------------------------------------------------------------
	 * icon info
	 */
	unsigned char *		ico_bits;				/* ptr to icon definition	*/
	int					ico_width;				/* icon width  in pixels	*/
	int					ico_height;				/* icon height in pixels	*/

	/*--------------------------------------------------------------------
	 * screen info
	 */
	TERMSCRN *			reg_scrn;				/* ptr to regular   screen	*/
	TERMSCRN *			alt_scrn;				/* ptr to alternate screen	*/
	TERMSCRN *			cur_scrn;				/* ptr to current   screen	*/

	/*--------------------------------------------------------------------
	 * terminal emulator info
	 */
	const CMD_TBL *		cmd_tbl_ptr;			/* ptr to command table		*/

	const KEY_DATA *	keys_fn;				/* function key table		*/
	const KEY_DATA *	keys_ms;				/* misc     key table		*/
	const KEY_DATA *	keys_ct;				/* control  key table		*/
	const KEY_DATA *	keys_kp;				/* keypad   key table		*/
	const KEY_DATA *	keys_pf;				/* prefix   key table		*/
	const KEY_DATA *	keys_st_up;				/* state-up key table		*/
	const KEY_DATA *	keys_st_dn;				/* state-dn key table		*/

	/*--------------------------------------------------------------------
	 * char-set info
	 */
	int					g[NUM_CHAR_SETS];		/* Gx char sets				*/

	/*--------------------------------------------------------------------
	 * color info
	 */
	int					scrn_mode;				/* screen mode				*/

	/*--------------------------------------------------------------------
	 * mouse info
	 */
	int					mouse_x;				/* current  x-position		*/
	int					mouse_y;				/* current  y-position		*/
	int					buttons;				/* current  button state	*/
	int					prev_buttons;			/* previous button state	*/

	/*--------------------------------------------------------------------
	 * state info
	 */
	int					our_pid;				/* for debugging			*/
	int					our_tid;				/* sub-thread id			*/

	int					cursor_mode;			/* T if we have focus		*/
	int					read_pending;			/* T if last op was a read	*/

	int					map_kbd;				/* T to map kbd chars		*/

	int					enable_mouse_events;	/* T to send mouse events	*/
	int					disp_info;				/* T to display info on lbl	*/
	int					log_pid;				/* T to add pid to evt log	*/

	int					bail;					/* non-zero if bailing		*/

	/*--------------------------------------------------------------------
	 * child process info
	 */
	char **				argv;					/* arg vector				*/
	int					pid;					/* PID of child				*/
	int					pid_died;				/* pid from waitpid			*/
	int					exit_code;				/* exit code				*/
	PTY_DATA *			pd;						/* pty data					*/

	PRINTER *			prt;					/* printer struct			*/

	/*--------------------------------------------------------------------
	 * input processing
	 */
	unsigned char		input_buffer[1024];		/* input from child			*/
	unsigned char *		input_ptr;				/* cur pos in input_buffer	*/
	int					input_count;			/* # chars in input_buffer	*/

	int					run_interval;			/* run wait interval (ms)	*/
	int					die_interval;			/* die wait interval (ms)	*/

	/*--------------------------------------------------------------------
	 * escape-seq struct
	 */
	ESCINFO				esc;					/* escape-info struct		*/
	int					prt_mode;				/* printing mode			*/

	/*--------------------------------------------------------------------
	 * icon info
	 */
	unsigned char		icon_bits[XVT_ICON_LEN];	/* icon-bits data		*/
	int					icon_width;				/* width  of icon bits		*/
	int					icon_height;			/* height of icon-bits		*/

	/*--------------------------------------------------------------------
	 * display info
	 */
	TERMWIN *			tw;						/* ptr to window info		*/
};
typedef struct termdata	TERMDATA;

/*------------------------------------------------------------------------
 * current screen accessor macros
 */
#define CUR_SCRN(td)			( (td)->cur_scrn )

#define CUR_SCRN_COLS(td)		SCRN_COLS		(CUR_SCRN(td))
#define CUR_SCRN_ROWS(td)		SCRN_ROWS		(CUR_SCRN(td))
#define CUR_SCRN_TABS(td)		SCRN_TABS		(CUR_SCRN(td))

#define CUR_SCRN_TOP(td)		SCRN_TOP		(CUR_SCRN(td))
#define CUR_SCRN_BOT(td)		SCRN_BOT		(CUR_SCRN(td))

#define CUR_SCRN_DISP(td)		SCRN_DISP		(CUR_SCRN(td))
#define CUR_SCRN_LINE(td,y)		SCRN_LINE		(CUR_SCRN(td), y)
#define CUR_SCRN_CHAR(td,y,x)	SCRN_CHAR		(CUR_SCRN(td), y, x)

#define CUR_SCRN_TABSTOPS(td)	SCRN_TABSTOPS	(CUR_SCRN(td))
#define CUR_SCRN_TABSTOP(td,x)	SCRN_TABSTOP	(CUR_SCRN(td), x)

#define CUR_SCRN_MODE(td)		SCRN_MODE		(CUR_SCRN(td))
#define CUR_SCRN_FG(td)			SCRN_FG			(CUR_SCRN(td))
#define CUR_SCRN_BG(td)			SCRN_BG			(CUR_SCRN(td))
#define CUR_SCRN_AT(td)			SCRN_AT			(CUR_SCRN(td))

#define CUR_SCRN_CURCS(td)		SCRN_CURCS		(CUR_SCRN(td))
#define CUR_SCRN_TMPCS(td)		SCRN_TMPCS		(CUR_SCRN(td))
#define CUR_SCRN_SAVCS(td)		SCRN_SAVCS		(CUR_SCRN(td))

#define CUR_SCRN_INS(td)		SCRN_INS		(CUR_SCRN(td))
#define CUR_SCRN_ORG(td)		SCRN_ORG		(CUR_SCRN(td))
#define CUR_SCRN_LOCK(td)		SCRN_LOCK		(CUR_SCRN(td))

#define CUR_SCRN_X(td)			SCRN_X			(CUR_SCRN(td))
#define CUR_SCRN_Y(td)			SCRN_Y			(CUR_SCRN(td))

#define CUR_SCRN_SAVE_X(td)		SCRN_SAVE_X		(CUR_SCRN(td))
#define CUR_SCRN_SAVE_Y(td)		SCRN_SAVE_Y		(CUR_SCRN(td))

#define CUR_SCRN_CSV(td)		SCRN_CSV		(CUR_SCRN(td))
#define CUR_SCRN_CON(td)		SCRN_CON		(CUR_SCRN(td))

/*------------------------------------------------------------------------
 * internal functions
 */
extern void		xvt_term_raise				(TERMDATA *td);
extern void		xvt_term_lower				(TERMDATA *td);
extern void		xvt_term_get_size			(TERMDATA *td, int *w, int *h);
extern void		xvt_term_set_size			(TERMDATA *td, int  w, int  h);
extern void		xvt_term_get_position		(TERMDATA *td, int *x, int *y);
extern void		xvt_term_set_position		(TERMDATA *td, int  x, int  y);

extern void		xvt_term_resize				(TERMDATA *td, int cols, int rows);
extern void		xvt_term_repaint				(TERMDATA *td);
extern void		xvt_term_disp_fill			(TERMDATA *td,
												int dx, int dy,
												int dw, int dh);
extern void		xvt_term_disp_copy			(TERMDATA *td,
												int dx, int dy,
												int dw, int dh,
												int sx, int sy);
extern void		xvt_term_disp_char			(TERMDATA *td,
												int x, int y, int ch,
												int fg, int bg, int at);

extern void		xvt_term_screen_print		(TERMDATA *td);
extern void		xvt_term_screen_clipboard	(TERMDATA *td);
extern void		xvt_term_write_report		(TERMDATA *td, const char *fmt,
												...);
extern void		xvt_term_force_kbd_state	(TERMDATA *td, int it_is);
extern void		xvt_term_set_kbd_state		(TERMDATA *td, int it_is);
extern void		xvt_term_set_scrn_mode		(TERMDATA *td, int mode);
extern void		xvt_term_set_win_title		(TERMDATA *td, char *str);
extern void		xvt_term_set_ico_title		(TERMDATA *td, char *str);
extern void		xvt_term_set_icon			(TERMDATA *td,
												int width, int height,
												unsigned char *bits);
extern void		xvt_term_set_txt_cursor		(TERMDATA *td, int state);
extern void		xvt_term_set_x11_cursor		(TERMDATA *td, int shape);
extern void		xvt_term_set_icon_state		(TERMDATA *td, int state);
extern void		xvt_term_set_font_no		(TERMDATA *td, int n);
extern void		xvt_term_set_rev_video		(TERMDATA *td, int reverse);
extern void		xvt_term_bell				(TERMDATA *td);

extern void		xvt_term_printer_open		(TERMDATA *td);
extern void		xvt_term_printer_close		(TERMDATA *td);

extern int		xvt_term_loop_once			(TERMDATA *td, int ms);

extern void *	xvt_term_main_loop			(void *data);

extern int		xvt_term_setup_all			(TERMDATA *td);
extern void		xvt_term_close_all			(TERMDATA *td);

extern void		xvt_term_shutdown			(TERMDATA *td);
extern int		xvt_term_restart			(TERMDATA *td);

extern void		xvt_term_list_pointers		(FILE *fp, int verbose);

#endif /* XVTTERM_H */
