/*------------------------------------------------------------------------
 * header for X11 window routines
 */
#ifndef XVTTERMWIN_H
#define XVTTERMWIN_H

/*------------------------------------------------------------------------
 * button definitions
 */
#define BUT_L			0x01		/* left   button	*/
#define BUT_M			0x02		/* middle button	*/
#define BUT_R			0x04		/* right  button	*/
#define BUT_U			0x08		/* scroll up		*/
#define BUT_D			0x10		/* scroll down		*/

/*------------------------------------------------------------------------
 * mono attributes
 */
#define NORMAL			0x00
#define BOLD			0x01
#define UNDERLINE		0x02
#define BLINK			0x04
#define REVERSE			0x08

/*------------------------------------------------------------------------
 * font id's
 */
#define FONT_NONE		-1
#define FONT_NORM		0
#define FONT_BOLD		1

/*------------------------------------------------------------------------
 * These are "extensions" to the user-color table for all the
 * other colors used.  This provides one array with all the colors
 * that are used.
 */
#define XVT_CLR_FG		( XVT_NUM_COLORS + 0 )	/* foreground		*/
#define XVT_CLR_BG		( XVT_NUM_COLORS + 1 )	/* background		*/
#define XVT_CLR_PR		( XVT_NUM_COLORS + 2 )	/* pointer			*/
#define XVT_CLR_BD		( XVT_NUM_COLORS + 3 )	/* border			*/
#define XVT_CLR_CR		( XVT_NUM_COLORS + 4 )	/* text cursor		*/
#define XVT_CLR_SB		( XVT_NUM_COLORS + 5 )	/* scrollbar		*/
#define XVT_CLR_ST		( XVT_NUM_COLORS + 6 )	/* scrollbar trough	*/
#define XVT_CLR_SB_TS	( XVT_NUM_COLORS + 7 )	/* SB top shadow	*/
#define XVT_CLR_SB_BS	( XVT_NUM_COLORS + 8 )	/* SB bottom shadow	*/

#define CMS_COLORSIZE	( XVT_NUM_COLORS + 9 )

/*------------------------------------------------------------------------
 * x-data struct
 */
typedef struct x_data	X_DATA;

/*------------------------------------------------------------------------
 * window data struct
 */
struct termwin
{
	/*--------------------------------------------------------------------
	 * user configuration stuff (from XVT_DATA)
	 */
	int			reverse;				/* reverse video			*/

	int			border;					/* border width in pixels	*/
	int			margin;					/* margin width in pixels	*/
	int			extra;					/* border + margin			*/

	int			sb_enable;				/* enable scrollbar			*/
	int			sb_right;				/* scrollbar on right		*/
	int			sb_shadow;				/* scrollbar shadow			*/
	int			sb_size;				/* scrollbar size			*/

	int			cr_color;				/* XVT_CLR_CR or -1			*/

	/*--------------------------------------------------------------------
	 * color table stuff (these never change: server dictates color/mono)
	 */
	int			can_be_color;			/* T if color supported		*/

	/*--------------------------------------------------------------------
	 * window stuff (changes only if user changes the window)
	 */
	int			win_cols;				/* width  of win in cols	*/
	int			win_rows;				/* height of win in rows	*/

	/*--------------------------------------------------------------------
	 * window state (these constantly change)
	 */
	int			cur_font_no;			/* cur font number			*/
	int			cur_font_type;			/* cur font type			*/

	int			cur_fg;					/* cur fg display is in		*/
	int			cur_bg;					/* cur bg display is in		*/

	int			doing_resize;			/* TRUE if doing resize		*/
	int			iconified;				/* TRUE if iconified		*/

	int			keys;					/* cur state of prefix-keys	*/
	int			prev_action;			/* previous action			*/

	/*--------------------------------------------------------------------
	 * X11 stuff which is only known internally
	 */
	X_DATA *	X;						/* ptr to x-data struct		*/
};
typedef struct termwin TERMWIN;

/*------------------------------------------------------------------------
 * setup/close functions
 */
extern TERMWIN *	xvt_w_setup_all			(XVT_DATA *xd);
extern void			xvt_w_close_all			(TERMWIN *tw);
extern void			xvt_w_close_con			(TERMWIN *tw);
extern void			xvt_w_free				(TERMWIN *tw);

/*------------------------------------------------------------------------
 * misc window functions
 */
extern void			xvt_w_resize			(TERMWIN *tw, int cols, int rows);
extern void			xvt_w_raise				(TERMWIN *tw);
extern void			xvt_w_lower				(TERMWIN *tw);

/*------------------------------------------------------------------------
 * misc get/set functions
 */
extern int			xvt_w_get_xfd			(const TERMWIN *tw);
extern unsigned int	xvt_w_get_windowid		(const TERMWIN *tw);

extern void			xvt_w_get_size			(TERMWIN *tw, int *w, int *h);
extern void			xvt_w_set_size			(TERMWIN *tw, int  w, int  h);

extern void			xvt_w_get_position		(TERMWIN *tw, int *x, int *y);
extern void			xvt_w_set_position		(TERMWIN *tw, int  x, int  y);

extern void			xvt_w_set_rev_video		(TERMWIN *tw, int reverse);
extern void			xvt_w_chg_font_no		(TERMWIN *tw, int direction);
extern void			xvt_w_set_font_no		(TERMWIN *tw, int font_no);
extern void			xvt_w_set_font_type		(TERMWIN *tw, int font_type);
extern int			xvt_w_set_scrn_mode		(TERMWIN *tw, int mode);
extern void			xvt_w_set_cursor		(TERMWIN *tw, int shape);
extern void			xvt_w_set_icon_state	(TERMWIN *tw, int state);
extern void			xvt_w_set_win_title		(TERMWIN *tw, const char *str);
extern void			xvt_w_set_ico_title		(TERMWIN *tw, const char *str);
extern void			xvt_w_set_icon			(TERMWIN *tw,
												int width, int height,
												unsigned char *bits);

extern void			xvt_w_set_fg			(TERMWIN *tw, int fg, int force);
extern void			xvt_w_set_bg			(TERMWIN *tw, int bg, int force);

/*------------------------------------------------------------------------
 * keyboard state functions
 */
extern int			xvt_w_query_kbd			(TERMWIN *tw);

/*------------------------------------------------------------------------
 * drawing functions
 */
extern void			xvt_w_rect_fill			(TERMWIN *tw, int x, int y,
												int w, int h);

extern void			xvt_w_rect_copy			(TERMWIN *tw, int dx, int dy,
												int dw, int dh,
												int sx, int sy);

extern void			xvt_w_char_rect			(TERMWIN *tw, int x, int y);
extern void			xvt_w_char_line			(TERMWIN *tw, int x, int y);
extern void			xvt_w_char_disp			(TERMWIN *tw, int x, int y, int c);
extern void			xvt_w_char_bold			(TERMWIN *tw, int x, int y, int c);

/*------------------------------------------------------------------------
 * event functions
 */
extern int			xvt_w_event				(TERMWIN *tw, XVT_EVENT *xe);

/*------------------------------------------------------------------------
 * misc functions
 */
extern void			xvt_w_bell				(TERMWIN *tw);

/*------------------------------------------------------------------------
 * pointer functions
 */
extern int			xvt_w_pointer_value		(const char *name);
extern const char *	xvt_w_pointer_name		(int n);
extern int			xvt_w_pointer_code		(int n);

/*------------------------------------------------------------------------
 * clipboard functions
 */
extern int			xvt_w_clipboard_copy_to	(TERMWIN *tw, unsigned char *data,
												int len);

extern unsigned char *	xvt_w_clipboard_copy_from	(TERMWIN *tw, int *plen);

#endif /* XVTTERMWIN_H */
