/*------------------------------------------------------------------------
 * X Toolkit options
 */
#ifndef XVTTK_H
#define XVTTK_H

/*------------------------------------------------------------------------
 * option flags
 *
 * These flags are used in generating the help display and in generating
 * the toolkit options dump.
 */
#define TK_FLAG_NOTHERE		0x0001		/* not valid in "here" mode		*/
#define TK_FLAG_APPOVR		0x0002		/* "appname" overrides default	*/
#define TK_FLAG_ENVNAME		0x0004		/* default is env name(with $)	*/
#define TK_FLAG_PHONYDFLT	0x0008		/* default is phony				*/
#define TK_FLAG_NOHELP_0	0x0010		/* don't show in help level 0	*/
#define TK_FLAG_NOTINHELP	0x0080		/* don't show in help (testing)	*/

/*------------------------------------------------------------------------
 * toolkit option struct
 */
struct tk_options
{
	int				opt_type;			/* option type					*/
	int				opt_offset;			/* var offset in struct	or -1	*/
	int				opt_flags;			/* option flags					*/
	const char *	opt_string;			/* option string to look for	*/
	const char *	opt_help;			/* help msg						*/
	const char *	opt_desc;			/* description					*/
	const char *	opt_resname;		/* resource name (may be NULL)	*/
	const char *	opt_default;		/* default value (may be NULL)	*/
										/* Note: for bool & num entries	*/
										/* this is the actual value		*/
	const char **	opt_lines;			/* extra help    (may be NULL)	*/
};
typedef struct tk_options	TK_OPTIONS;

extern const TK_OPTIONS	xvt_tk_options[];

/*------------------------------------------------------------------------
 * functions
 */
extern const TK_OPTIONS *	xvt_tk_find_by_option	(const char *option);
extern const TK_OPTIONS *	xvt_tk_find_by_resname	(const char *resname);

extern char *				xvt_tk_data_value		(XVT_DATA *xd,
													const TK_OPTIONS *op,
													int use_dflt,
													char *buf);

extern void					xvt_tk_data_init		(XVT_DATA *xd,
													int set_dflt);

#endif /* XVTTK_H */
