/*------------------------------------------------------------------------
 * vt100 terminal emulator commands
 */
#include "xvtcommon.h"
#include "xvtvt100.h"

#define te		(&(td->esc))

/*------------------------------------------------------------------------
 * misc other defines
 */
#define PLURAL(n)		(n == 1 ? "" : "s")

/*------------------------------------------------------------------------
 * do any initialization
 */
void xvt_vt100_init (TERMDATA *td)
{
	/*--------------------------------------------------------------------
	 * setup all char sets
	 */  
	td->g[CHAR_SET_G0] = GCS_US;
	td->g[CHAR_SET_G1] = GCS_US;
	td->g[CHAR_SET_G2] = GCS_US;
	td->g[CHAR_SET_G3] = GCS_US;

	/*--------------------------------------------------------------------
	 * setup all key structs
	 */  
	td->keys_fn		= xvt_vt100_keys_fn_reg;
	td->keys_ct		= xvt_vt100_keys_ct_reg;
	td->keys_ms		= xvt_vt100_keys_ms_reg;
	td->keys_kp		= xvt_vt100_keys_kp_reg;
	td->keys_pf		= xvt_vt100_keys_pf_reg;
	td->keys_st_up	= xvt_vt100_keys_st_up;
	td->keys_st_dn	= xvt_vt100_keys_st_dn;
}

/*------------------------------------------------------------------------
 * output a character to the screen
 */
void xvt_vt100_char_output (TERMDATA *td, int ch)
{
	CELL *	cp;

	/*--------------------------------------------------------------------
	 * debug output
	 */
	{
		char	buf[2];

		buf[0] = ch;
		buf[1] = 0;
		xvt_log_output_chr(td, buf);
	}

	/*--------------------------------------------------------------------
	 * adjust character according to character set
	 */
	if (CUR_SCRN_TMPCS(td) != CHAR_SET_NONE)
	{
		ch = xvt_gcs_get_char(td->g[CUR_SCRN_TMPCS(td)], ch);
		CUR_SCRN_TMPCS(td) = CHAR_SET_NONE;
	}
	else
	{
		ch = xvt_gcs_get_char(td->g[CUR_SCRN_CURCS(td)], ch);
	}

	/*--------------------------------------------------------------------
	 * now display it
	 */
	if (! CUR_SCRN_INS(td))
	{
		/*----------------------------------------------------------------
		 * check if this will wrap to next line
		 */
		if (CUR_SCRN_X(td) >= CUR_SCRN_COLS(td))
		{
			if (td->xd->autowrap)
			{
				/*--------------------------------------------------------
				 * bump to next line
				 */
				CUR_SCRN_X(td) = 0;
				CUR_SCRN_Y(td)++;

				/*-----------------------------------------------------------
				 * now check if we have to scroll
				 */
				if (CUR_SCRN_Y(td) > CUR_SCRN_BOT(td))
				{
					CUR_SCRN_Y(td)--;
					xvt_curscrn_lines_ins(td, 1);
				}
				else if (CUR_SCRN_Y(td) >= CUR_SCRN_ROWS(td))
				{
					CUR_SCRN_Y(td)--;
				}
			}
			else
			{
				CUR_SCRN_X(td)--;
			}
		}

		/*----------------------------------------------------------------
		 * display it if it doesn't match char already on screen
		 */
		cp = CUR_SCRN_CHAR(td, CUR_SCRN_Y(td), CUR_SCRN_X(td));

		if (cp->at != CUR_SCRN_AT(td) ||
			cp->bg != CUR_SCRN_BG(td) ||
			cp->fg != CUR_SCRN_FG(td) ||
			cp->ch != ch)
		{
			cp->ch = (unsigned char)ch;
			cp->fg = CUR_SCRN_FG(td);
			cp->bg = CUR_SCRN_BG(td);
			cp->at = CUR_SCRN_AT(td);

			xvt_term_disp_char(td, CUR_SCRN_X(td), CUR_SCRN_Y(td), cp->ch,
				cp->fg, cp->bg, cp->at);
		}
	}
	else
	{
		xvt_curscrn_chars_ins(td, 1);

		cp = CUR_SCRN_CHAR(td, CUR_SCRN_Y(td), CUR_SCRN_X(td));
		cp->ch = (unsigned char)ch;
		cp->fg = CUR_SCRN_FG(td);
		cp->bg = CUR_SCRN_BG(td);
		cp->at = CUR_SCRN_AT(td);

		xvt_term_disp_char(td, CUR_SCRN_X(td), CUR_SCRN_Y(td), cp->ch,
			cp->fg, cp->bg, cp->at);
	}

	/*--------------------------------------------------------------------
	 * bump x cursor position
	 */
	CUR_SCRN_X(td)++;

	/*--------------------------------------------------------------------
	 * check margin bell
	 */
	if (td->read_pending && td->xd->mbell)
	{
		if (CUR_SCRN_X(td) == (CUR_SCRN_COLS(td) - td->xd->mbcols))
			xvt_term_bell(td);
	}

	/*--------------------------------------------------------------------
	 * now check out position again
	 */
	if (CUR_SCRN_X(td) == CUR_SCRN_COLS(td))
	{
		if (td->xd->autowrap)
		{
			xvt_vt100_CTL_LF(td);
			CUR_SCRN_X(td) = 0;
		}
		else
		{
			CUR_SCRN_X(td)--;
		}
	}

	td->read_pending = FALSE;
}

/*------------------------------------------------------------------------
 * process printing
 */
void xvt_vt100_char_print (TERMDATA *td, int ch)
{
	/*--------------------------------------------------------------------
	 * if currently reading an esc-seq, cache this char into the buffer
	 */
	if (td->prt_mode == PRT_MODE_ESC)
	{
		te->esc_chars[te->esc_char_cnt++] = ch;
		te->esc_chars[te->esc_char_cnt  ] = 0;

		/*----------------------------------------------------------------
		 * an alpha char will end esc-seq processing
		 */
		if (isalpha(ch))
		{
			/*------------------------------------------------------------
			 * check if this is an end-printing esc-seq.
			 */
			if (strcmp(te->esc_chars, "\033[4i") == 0)
			{
				te->reading_esc_seq	= FALSE;

				xvt_log_output_seq(td, "stop print pass-thru");
				xvt_term_printer_close(td);
			}
			else
			{
				td->prt_mode = PRT_MODE_NORM;
				xvt_prt_output_str(td->prt, te->esc_chars);
			}

			te->esc_chars[0]	= 0;
			te->esc_char_cnt	= 0;
			te->vt_mode			= ESC_TOP;
		}

		return;
	}

	/*--------------------------------------------------------------------
	 * not reading an esc-seq, check if this will start one
	 */
	if (ch == 0x1b)
	{
		te->esc_chars[te->esc_char_cnt++] = ch;
		te->esc_chars[te->esc_char_cnt  ] = 0;
		td->prt_mode = PRT_MODE_ESC;
		return;
	}

	/*--------------------------------------------------------------------
	 * just a reg char
	 */
	xvt_prt_output_chr(td->prt, ch);
}

/*========================================================================
 * static support functions
 *========================================================================
 */

/*------------------------------------------------------------------------
 * XVT_KBDMAP	set keyboard mapping
 */
static void xvt_vt100_kbdmap (TERMDATA *td, int state)
{
	if (state)
	{
		xvt_log_output_seq(td, "kbd-map %d on", state);

		/*----------------------------------------------------------------
		 * set kbd-mapping on
		 */
		td->map_kbd = state;
		td->keys_ct = xvt_vt100_keys_ct_map;
		td->keys_pf = xvt_vt100_keys_pf_map;
		td->keys_ms = xvt_vt100_keys_ms_map;
		td->keys_kp = xvt_vt100_keys_kp_map;

		/*----------------------------------------------------------------
		 * We don't set kbd-state on yet because the user may not
		 * be ready yet. We'll let him get it the first time the
		 * kbd is read.
		 */
#if 0
		xvt_term_set_kbd_state(td, -1);
#endif
	}
	else
	{
		xvt_log_output_seq(td, "kbd-map off");

		/*----------------------------------------------------------------
		 * set kbd-state off
		 */
		xvt_term_set_kbd_state(td, 0);

		/*----------------------------------------------------------------
		 * set kbd-mapping off
		 */
		td->map_kbd = 0;
		td->keys_ct = xvt_vt100_keys_ct_reg;
		td->keys_pf = xvt_vt100_keys_pf_reg;
		td->keys_ms = xvt_vt100_keys_ms_reg;
		td->keys_kp = xvt_vt100_keys_kp_reg;
	}
}

/*------------------------------------------------------------------------
 * DECCKM	application cursor keys on|off
 */
static void xvt_vt100_DECCKM (TERMDATA *td, int bf)
{
	/*--------------------------------------------------------------------
	 * Only set KP keys if kbd is not "mapped".
	 */
	if (! td->map_kbd)
		td->keys_kp = (bf ? xvt_vt100_keys_kp_apl : xvt_vt100_keys_kp_reg);
}

/*------------------------------------------------------------------------
 * DECNKM	application keypad keys on|off
 */
static void xvt_vt100_DECNKM (TERMDATA *td, int bf)
{
	/*--------------------------------------------------------------------
	 * Only set KP keys if kbd is not "mapped".
	 */
	if (! td->map_kbd)
		td->keys_kp = (bf ? xvt_vt100_keys_kp_cur : xvt_vt100_keys_kp_reg);
}

/*------------------------------------------------------------------------
 * ICON		process icon data
 */
static void xvt_vt100_ICON_BITS (TERMDATA *td)
{
	int		w;
	int		h;
	int		i;
	int		b;
	char *	c;

	/*--------------------------------------------------------------------
	 * reading an icon
	 *
	 * Icon data is as follows:
	 *
	 *	width  (in pixels) (must be multiple of 8)
	 *	height (in pixels) (must be multiple of 8)
	 *	((width * height) / 8) data-bytes of pixel data.
	 *
	 *	width	one byte  encoded as (w + ' ')
	 *	height	one byte  encoded as (h + ' ')
	 *	data	two bytes encoded as ((b >> 4) + ' ', (b & 0x0f) + ' ')
	 *
	 * We impose a limit of 64 x 64 pixels on the size.
	 */
	if (te->data_cnt <= 2)
		return;

	w = (te->data_chars[0] - ' ');
	h = (te->data_chars[1] - ' ');

	if (w > 64)	w = 64;
	if (h > 64)	h = 64;

	b = (w * h) / 8;

	if (((b * 2) + 2) != te->data_cnt)
		return;

	c = te->data_chars + 2;

	/*--------------------------------------------------------------------
	 * store icon data
	 */
	td->icon_width	= w;
	td->icon_height	= h;

	for (i=0; i<b; i++)
	{
		int	l = (*c++ - ' ') << 4;
		int r = (*c++ - ' ');

		td->icon_bits[i] = (l | r);
	}

	/*--------------------------------------------------------------------
	 * now set it
	 */
	xvt_term_set_icon(td, td->icon_width, td->icon_height, td->icon_bits);
}

/*========================================================================
 * control char routines
 *========================================================================
 */

/*------------------------------------------------------------------------
 * ENQ		enquire terminal status
 */
void xvt_vt100_CTL_ENQ (TERMDATA *td)
{
	xvt_vt100_CSI_DA(td);
}

/*------------------------------------------------------------------------
 * BL		ring the bell
 */
void xvt_vt100_CTL_BL (TERMDATA *td)
{
	xvt_term_bell(td);
}

/*------------------------------------------------------------------------
 * BS		backspace
 */
void xvt_vt100_CTL_BS (TERMDATA *td)
{
	if (CUR_SCRN_X(td) > 0)
	{
		CUR_SCRN_X(td)--;
	}
	else if (td->xd->revwrap)
	{
		if (CUR_SCRN_Y(td) > 0)
		{
			CUR_SCRN_Y(td)--;
			CUR_SCRN_X(td) = CUR_SCRN_COLS(td) - 1;
		}
	}
}

/*------------------------------------------------------------------------
 * HT		horizontal tab (just move, no writing of chars)
 */
void xvt_vt100_CTL_HT (TERMDATA *td)
{
	int x;

	for (x=0; x<CUR_SCRN_COLS(td); x++)
	{
		if (CUR_SCRN_TABSTOP(td, x))
			break;
	}

	if (x < CUR_SCRN_COLS(td))
		CUR_SCRN_X(td) = x;
}

/*------------------------------------------------------------------------
 * LF		line-feed
 */
void xvt_vt100_CTL_LF (TERMDATA *td)
{
	CUR_SCRN_Y(td)++;
	if (CUR_SCRN_Y(td) > CUR_SCRN_BOT(td))
	{
		CUR_SCRN_Y(td)--;
		xvt_curscrn_lines_ins(td, 1);
	}
	else if (CUR_SCRN_Y(td) >= CUR_SCRN_ROWS(td))
	{
		CUR_SCRN_Y(td)--;
	}
}

/*------------------------------------------------------------------------
 * FF		form-feed
 */
void xvt_vt100_CTL_FF (TERMDATA *td)
{
	xvt_curscrn_clear(td);
	CUR_SCRN_X(td) = 0;
	CUR_SCRN_Y(td) = 0;
}

/*------------------------------------------------------------------------
 * CR		carriage return
 */
void xvt_vt100_CTL_CR (TERMDATA *td)
{
	CUR_SCRN_X(td) = 0;
}

/*------------------------------------------------------------------------
 * SO		alt-chars on (switch to G1)
 */
void xvt_vt100_CTL_SO (TERMDATA *td)
{
	CUR_SCRN_CURCS(td) = CHAR_SET_G1;
	xvt_log_output_seq(td, "select G1 char-set");
}

/*------------------------------------------------------------------------
 * SI		alt-chars off (switch to G0)
 */
void xvt_vt100_CTL_SI (TERMDATA *td)
{
	CUR_SCRN_CURCS(td) = CHAR_SET_G0;
	xvt_log_output_seq(td, "select G0 char-set");
}

/*========================================================================
 * TOP mode commands
 *========================================================================
 */

/*------------------------------------------------------------------------
 * DECSC	save cursor position
 */
void xvt_vt100_TOP_DECSC (TERMDATA *td)
{
	xvt_log_output_seq(td, "save-cursor");

	CUR_SCRN_SAVE_X(td)	= CUR_SCRN_X(td);
	CUR_SCRN_SAVE_Y(td)	= CUR_SCRN_Y(td);
	CUR_SCRN_SAVCS(td)	= CUR_SCRN_CURCS(td);
}

/*------------------------------------------------------------------------
 * DECRC	restore cursor position
 */
void xvt_vt100_TOP_DECRC (TERMDATA *td)
{
	xvt_log_output_seq(td, "restore-cursor");

	CUR_SCRN_X(td)		= CUR_SCRN_SAVE_X(td);
	CUR_SCRN_Y(td)		= CUR_SCRN_SAVE_Y(td);
	CUR_SCRN_CURCS(td)	= CUR_SCRN_SAVCS(td);
}

/*------------------------------------------------------------------------
 * DECPAM	application keypad
 */
void xvt_vt100_TOP_DECPAM (TERMDATA *td)
{
	xvt_log_output_seq(td, "application keypad on");
	xvt_vt100_DECNKM(td, TRUE);
}

/*------------------------------------------------------------------------
 * DECPNM	normal keypad
 */
void xvt_vt100_TOP_DECPNM (TERMDATA *td)
{
	xvt_log_output_seq(td, "application keypad off");
	xvt_vt100_DECNKM(td, FALSE);
}

/*------------------------------------------------------------------------
 * RIS		reset terminal
 */
void xvt_vt100_TOP_RIS (TERMDATA *td)
{
	xvt_log_output_seq(td, "reset terminal");

	/*--------------------------------------------------------------------
	 * reset all screens & set current screen as reg screen
	 */
	xvt_scrn_reset(td->reg_scrn);
	xvt_scrn_reset(td->alt_scrn);
	td->cur_scrn = td->reg_scrn;

	/*--------------------------------------------------------------------
	 * set window title back
	 */
	xvt_term_set_win_title(td, td->org_label);
	xvt_term_set_ico_title(td, td->org_label);

	/*--------------------------------------------------------------------
	 * set normal keypad & cursor
	 */
	xvt_vt100_DECCKM(td, FALSE);
	xvt_vt100_DECNKM(td, FALSE);

	/*--------------------------------------------------------------------
	 * unmap the keyboard
	 */
	xvt_vt100_kbdmap(td, 0);

	/*--------------------------------------------------------------------
	 * reset tty modes
	 */
	xvt_pty_setmode(td->pd, &td->pd->tty);

	/*--------------------------------------------------------------------
	 * reset all char sets
	 */
	td->g[CHAR_SET_G0] = GCS_US;
	td->g[CHAR_SET_G1] = GCS_US;
	td->g[CHAR_SET_G2] = GCS_US;
	td->g[CHAR_SET_G3] = GCS_US;

	/*--------------------------------------------------------------------
	 * force screen to mono mode
	 */
	xvt_term_set_scrn_mode(td, FALSE);

	/*--------------------------------------------------------------------
	 * now clear the screen
	 */
	xvt_curscrn_clear(td);
}

/*------------------------------------------------------------------------
 * LS2		invoke G2 char set
 */
void xvt_vt100_TOP_LS2 (TERMDATA *td)
{
	xvt_log_output_seq(td, "select G2 char set");

	CUR_SCRN_CURCS(td) = CHAR_SET_G2;
}

/*------------------------------------------------------------------------
 * LS3		invoke G3 char set
 */
void xvt_vt100_TOP_LS3 (TERMDATA *td)
{
	xvt_log_output_seq(td, "select G3 char set");

	CUR_SCRN_CURCS(td) = CHAR_SET_G3;
}

/*------------------------------------------------------------------------
 * TOP_DEVEND		end device processing
 */
void xvt_vt100_TOP_DEVEND (TERMDATA *td)
{
	switch (te->vt_type)
	{
	case DEV_DCS:
		xvt_log_output_seq(td, "DCS");
		break;

	case DEV_PM:
		xvt_log_output_seq(td, "PM");
		break;

	case DEV_APC:
		xvt_log_output_seq(td, "APC");
		xvt_vt100_ICON_BITS(td);
		break;
	}
}

/*------------------------------------------------------------------------
 * TOP_DA			return terminal ID
 */
void xvt_vt100_TOP_DA (TERMDATA *td)
{
	xvt_log_output_seq(td, "send device attribtes");
	xvt_term_write_report(td, "\033[?1;2c");
}

/*------------------------------------------------------------------------
 * TOP_CLL			cursor to lower-left of screen
 */
void xvt_vt100_TOP_CLL (TERMDATA *td)
{
	xvt_log_output_seq(td, "cursor to lower-left of screen");
	CUR_SCRN_X(td) = 0;
	CUR_SCRN_Y(td) = CUR_SCRN_ROWS(td) - 1;
}

/*------------------------------------------------------------------------
 * TOP_HTS			tab set
 */
void xvt_vt100_TOP_HTS (TERMDATA *td)
{
	xvt_log_output_seq(td, "set tab at cur col");
	xvt_scrn_tab_set(CUR_SCRN(td), SCRN_TAB_SET_CUR);
}

/*------------------------------------------------------------------------
 * TOP_SS2			single shift select of G2 char set
 */
void xvt_vt100_TOP_SS2 (TERMDATA *td)
{
	xvt_log_output_seq(td, "single shift select of G2 char set");
	CUR_SCRN_TMPCS(td) = CHAR_SET_G2;
}

/*------------------------------------------------------------------------
 * TOP_SS3			single shift select of G3 char set
 */
void xvt_vt100_TOP_SS3 (TERMDATA *td)
{
	xvt_log_output_seq(td, "single shift select of G3 char set");
	CUR_SCRN_TMPCS(td) = CHAR_SET_G3;
}

/*------------------------------------------------------------------------
 * TOP_IND			index down
 */
void xvt_vt100_TOP_IND (TERMDATA *td)
{
	xvt_log_output_seq(td, "index down");

	if (CUR_SCRN_Y(td) >= CUR_SCRN_BOT(td))
	{
		xvt_curscrn_scroll(td, -1);
	}
	else
	{
		CUR_SCRN_Y(td)++;
	}
}

/*------------------------------------------------------------------------
 * TOP_RI			index up (reverse index)
 */
void xvt_vt100_TOP_RI (TERMDATA *td)
{
	xvt_log_output_seq(td, "index up");
	if (CUR_SCRN_Y(td) <= CUR_SCRN_TOP(td))
	{
		xvt_curscrn_scroll(td, 1);
	}
	else
	{
		CUR_SCRN_Y(td)--;
	}
}

/*------------------------------------------------------------------------
 * TOP_NEL			next line
 */
void xvt_vt100_TOP_NEL (TERMDATA *td)
{
	xvt_log_output_seq(td, "next line");
	xvt_vt100_CTL_LF(td);
	CUR_SCRN_X(td) = 0;
}

/*------------------------------------------------------------------------
 * TOP_MEML			memory lock
 */
void xvt_vt100_TOP_MEML (TERMDATA *td)
{
	xvt_log_output_seq(td, "memory lock");
	CUR_SCRN_LOCK(td) = CUR_SCRN_Y(td);
}

/*------------------------------------------------------------------------
 * TOP_MEMU			memory unlock
 */
void xvt_vt100_TOP_MEMU (TERMDATA *td)
{
	xvt_log_output_seq(td, "memory unlock");
	CUR_SCRN_LOCK(td) = -1;
}

/*========================================================================
 * DEC mode commands
 *========================================================================
 */

/*------------------------------------------------------------------------
 * DECALN	DEC alignment test (fill screen with E's)
 */
void xvt_vt100_DEC_DECALN (TERMDATA *td)
{
	xvt_log_output_seq(td, "DECALN");
	xvt_curscrn_set(td, 'E');
}

/*========================================================================
 * GCS mode commands
 *========================================================================
 */

/*------------------------------------------------------------------------
 * GCS_CODE		designate a char set
 */
void xvt_vt100_GCS_CODE (TERMDATA *td)
{
	int c	= te->code;
	int	s	= xvt_gcs_get_gcs_code(c);

	if (s < 0)
	{
		xvt_log_output_seq(td, "Set G%d to %c: unknown: ignored",
			te->vt_type, c);
	}
	else
	{
		td->g[te->vt_type] = s;
		xvt_log_output_seq(td, "Set G%d to %c: %s",
			te->vt_type, c, xvt_gcs_get_name(s));
	}
}

/*========================================================================
 * CSI mode commands
 *========================================================================
 */

/*------------------------------------------------------------------------
 * HPA	move the cursor to col x
 */
void xvt_vt100_CSI_HPA (TERMDATA *td)
{
	int x = te->n[0];

	if (--x < 0)	x = 0;

	xvt_log_output_seq(td, "move-cursor to col %d", x);

	if (x >= CUR_SCRN_COLS(td))	x = CUR_SCRN_COLS(td)-1;
	CUR_SCRN_X(td) = x;
}

/*------------------------------------------------------------------------
 * VPA	move the cursor to row y
 */
void xvt_vt100_CSI_VPA (TERMDATA *td)
{
	int y = te->n[0];

	if (--y < 0)	y = 0;

	xvt_log_output_seq(td, "move-cursor to row %d", y);

	if (y >= CUR_SCRN_ROWS(td))	y = CUR_SCRN_ROWS(td)-1;
	CUR_SCRN_Y(td) = y;
}

/*------------------------------------------------------------------------
 * CUP	move the cursor to (x,y)
 */
void xvt_vt100_CSI_CUP (TERMDATA *td)
{
	int y = te->n[0];
	int x = te->n[1];

	if (--y < 0)	y = 0;
	if (--x < 0)	x = 0;

	xvt_log_output_seq(td, "move-cursor to (%d,%d)", x, y);

	if (CUR_SCRN_ORG(td))
	{
		/*----------------------------------------------------------------
		 * "origin" mode means that the cursor is limited to moving
		 * within the scrolling region and line numbers are relative to
		 * the scrolling region.
		 */
		int nlines = (CUR_SCRN_BOT(td) - CUR_SCRN_TOP(td) + 1);

		if (y >= nlines)
			y = (nlines - 1);

		CUR_SCRN_Y(td) = CUR_SCRN_TOP(td) + y;
	}
	else
	{
		if (y >= CUR_SCRN_ROWS(td))
			y = CUR_SCRN_ROWS(td)-1;
		CUR_SCRN_Y(td) = y;
	}

	if (x >= CUR_SCRN_COLS(td))
		x = CUR_SCRN_COLS(td)-1;
	CUR_SCRN_X(td) = x;
}

/*------------------------------------------------------------------------
 * ECH	erase n chars (with current attrs)
 */
void xvt_vt100_CSI_ECH (TERMDATA *td)
{
	int n = te->n[0];

	if (n <= 0)
		n = 1;

	xvt_log_output_seq(td, "erase %d chars", n);
	xvt_curscrn_chars_era(td, n);
}

/*------------------------------------------------------------------------
 * ED	erase display (with current attrs)
 *
 * The current cursor position is not changed.
 */
void xvt_vt100_CSI_ED (TERMDATA *td)
{
	int mode = te->n[0];

	switch (mode)
	{
	case 0:
		/*----------------------------------------------------------------
		 * clear from cursor to EOS
		 */
		xvt_log_output_seq(td, "clear-to-eos");

		xvt_curscrn_chars_clr(td, 1);
		if (CUR_SCRN_Y(td) < CUR_SCRN_ROWS(td) - 1)
		{
			xvt_curscrn_lines_clr(td, CUR_SCRN_Y(td) + 1,
				(CUR_SCRN_ROWS(td) - (CUR_SCRN_Y(td) + 1)));
		}
		break;

	case 1:
		/*----------------------------------------------------------------
		 * clear from BOS to cursor
		 */
		xvt_log_output_seq(td, "clear-from-bos");

		xvt_curscrn_chars_clr(td, -1);
		if (CUR_SCRN_Y(td) > 0)
		{
			xvt_curscrn_lines_clr(td, 0, CUR_SCRN_Y(td) - 1);
		}
		break;

	case 2:
		/*----------------------------------------------------------------
		 * clear entire screen
		 */
		xvt_log_output_seq(td, "clear-screen");

		xvt_curscrn_clear(td);
		break;

	case 3:
		/*----------------------------------------------------------------
		 * erase saved lines
		 */
		xvt_log_output_seq(td, "erase saved lines: ignored");
		break;

	default:
		/*----------------------------------------------------------------
		 * invalid seq
		 */
		xvt_log_output_seq(td, "invalid ED mode %d", mode);
		break;
	}
}

/*------------------------------------------------------------------------
 * EL	erase line (with current attrs)
 *
 * The current cursor position is not changed.
 */
void xvt_vt100_CSI_EL (TERMDATA *td)
{
	int mode = te->n[0];

	switch (mode)
	{
	case 0:
		/*----------------------------------------------------------------
		 * clear to EOL
		 */
		xvt_log_output_seq(td, "clear-to-eol");
		xvt_curscrn_chars_clr(td, 1);
		break;

	case 1:
		/*----------------------------------------------------------------
		 * clear from BOL
		 */
		xvt_log_output_seq(td, "clear-from-bol");
		xvt_curscrn_chars_clr(td, -1);
		break;

	case 2:
		/*----------------------------------------------------------------
		 * clear entire line
		 */
		xvt_log_output_seq(td, "clear-entire-line");
		xvt_curscrn_chars_clr(td, 0);
		break;

	default:
		/*----------------------------------------------------------------
		 * invalid mode
		 */
		xvt_log_output_seq(td, "invalid EL mode %d", mode);
		break;
	}
}

/*------------------------------------------------------------------------
 * CUU	move the cursor up
 */
void xvt_vt100_CSI_CUU (TERMDATA *td)
{
	int n = te->n[0];

	if (n == 0)
		n = 1;

	xvt_log_output_seq(td, "up %d line%s", n, PLURAL(n));

	CUR_SCRN_Y(td) -= n;
	if (CUR_SCRN_Y(td) < 0)
		CUR_SCRN_Y(td) = 0;

	if (CUR_SCRN_ORG(td) && CUR_SCRN_Y(td) < CUR_SCRN_TOP(td))
		CUR_SCRN_Y(td) = CUR_SCRN_TOP(td);
}

/*------------------------------------------------------------------------
 * CNL	move the cursor up & go to col 1
 */
void xvt_vt100_CSI_CNL (TERMDATA *td)
{
	int n = te->n[0];

	if (n == 0)
		n = 1;

	xvt_log_output_seq(td, "up %d line%s & go to col 1", n, PLURAL(n));

	CUR_SCRN_Y(td) -= n;
	if (CUR_SCRN_Y(td) < 0)
		CUR_SCRN_Y(td) = 0;

	if (CUR_SCRN_ORG(td) && CUR_SCRN_Y(td) < CUR_SCRN_TOP(td))
		CUR_SCRN_Y(td) = CUR_SCRN_TOP(td);

	CUR_SCRN_X(td) = 0;
}

/*------------------------------------------------------------------------
 * CUD	move the cursor down
 */
void xvt_vt100_CSI_CUD (TERMDATA *td)
{
	int n = te->n[0];

	if (n == 0)
		n = 1;

	xvt_log_output_seq(td, "down %d line%s", n, PLURAL(n));

	CUR_SCRN_Y(td) += n;
	if (CUR_SCRN_Y(td) >= CUR_SCRN_ROWS(td))
		CUR_SCRN_Y(td) = CUR_SCRN_ROWS(td) - 1;

	if (CUR_SCRN_ORG(td) && CUR_SCRN_Y(td) > CUR_SCRN_BOT(td))
		CUR_SCRN_Y(td) = CUR_SCRN_BOT(td);
}

/*------------------------------------------------------------------------
 * CPL	move the cursor down & go to col 1
 */
void xvt_vt100_CSI_CPL (TERMDATA *td)
{
	int n = te->n[0];

	if (n == 0)
		n = 1;

	xvt_log_output_seq(td, "down %d line%s & go to col 1", n, PLURAL(n));

	CUR_SCRN_Y(td) += n;
	if (CUR_SCRN_Y(td) >= CUR_SCRN_ROWS(td))
		CUR_SCRN_Y(td) = CUR_SCRN_ROWS(td) - 1;

	if (CUR_SCRN_ORG(td) && CUR_SCRN_Y(td) > CUR_SCRN_BOT(td))
		CUR_SCRN_Y(td) = CUR_SCRN_BOT(td);

	CUR_SCRN_X(td) = 0;
}

/*------------------------------------------------------------------------
 * CUF	move the cursor right
 */
void xvt_vt100_CSI_CUF (TERMDATA *td)
{
	int n = te->n[0];

	if (n == 0)
		n = 1;

	xvt_log_output_seq(td, "right %d line%s", n, PLURAL(n));

	CUR_SCRN_X(td) += n;
	if (CUR_SCRN_X(td) >= CUR_SCRN_COLS(td))
		CUR_SCRN_X(td) = CUR_SCRN_COLS(td) - 1;
}

/*------------------------------------------------------------------------
 * CUB	move the cursor left
 */
void xvt_vt100_CSI_CUB (TERMDATA *td)
{
	int n = te->n[0];

	if (n == 0)
		n = 1;

	xvt_log_output_seq(td, "left %d line%s", n, PLURAL(n));

	CUR_SCRN_X(td) -= n;
	if (CUR_SCRN_X(td) < 0)
		CUR_SCRN_X(td) = 0;
}

/*------------------------------------------------------------------------
 * ICH	insert n characters at the current cursor position
 *
 * This moves the chars on the current line over one & inserts a
 * blank char at the current position (with current attributes).
 *
 * The current cursor position is not changed.
 */
void xvt_vt100_CSI_ICH (TERMDATA *td)
{
	int n = te->n[0];

	if (n < 1)
		n = 1;

	xvt_log_output_seq(td, "insert %d char%s", n, PLURAL(n));

	xvt_curscrn_chars_ins(td, n);
}

/*------------------------------------------------------------------------
 * DCH	delete n chars from where the cursor is
 *
 * This moves the rest of the line to the left one char & inserts a
 * blank (with current attrs) at the end of the line.
 *
 * The current cursor position is not changed.
 */
void xvt_vt100_CSI_DCH (TERMDATA *td)
{
	int n = te->n[0];

	if (n == 0)
		n = 1;

	xvt_log_output_seq(td, "delete %d char%s", n, PLURAL(n));

	xvt_curscrn_chars_del(td, n);
}

/*------------------------------------------------------------------------
 * AL	add n lines at the current cursor position (with current attrs)
 *
 * This results in all rows above the current line to be "scrolled" up.
 * The cursor is moved to the start of the current line.
 */
void xvt_vt100_CSI_AL (TERMDATA *td)
{
	int n = te->n[0];

	if (n == 0)
		n = 1;

	xvt_log_output_seq(td, "add %d line%s", n, PLURAL(n));

	xvt_curscrn_lines_ins(td, n);
}

/*------------------------------------------------------------------------
 * DL	delete n lines where the cursor is on
 *
 * This will add a blank line at the bottom of the screen
 * (with the current attrs).
 */
void xvt_vt100_CSI_DL (TERMDATA *td)
{
	int n = te->n[0];

	if (n == 0)
		n = 1;

	xvt_log_output_seq(td, "delete %d line%s", n, PLURAL(n));

	xvt_curscrn_lines_del(td, n);
}

/*------------------------------------------------------------------------
 * SM 	set mode
 */
void xvt_vt100_CSI_SM (TERMDATA *td)
{
	int i;

	for (i=0; i<=te->n_index; i++)
	{
		int n = te->n[i];

		switch (n)
		{
		case 4:			/* insert mode */
			xvt_log_output_seq(td, "insert mode on");
			CUR_SCRN_INS(td) = TRUE;
			break;

		case 20:		/* auto-nl */
			xvt_log_output_seq(td, "auto-nl mode on: ignored");
			break;

		default:
			xvt_log_output_seq(td, "invalid SM opt %d", n);
			break;
		}
	}
}

/*------------------------------------------------------------------------
 * RM	reset mode
 */
void xvt_vt100_CSI_RM (TERMDATA *td)
{
	int i;

	for (i=0; i<=te->n_index; i++)
	{
		int n = te->n[i];

		switch (n)
		{
		case 4:			/* insert mode */
			xvt_log_output_seq(td, "insert mode off");
			CUR_SCRN_INS(td) = FALSE;
			break;

		case 20:		/* auto-nl */
			xvt_log_output_seq(td, "auto-nl mode off: ignored");
			break;

		default:
			xvt_log_output_seq(td, "invalid RM opt %d", n);
			break;
		}
	}
}

/*------------------------------------------------------------------------
 * SGR_AT	set attribute
 */
static void xvt_vt100_SGR_AT (TERMDATA *td, int set, int n)
{
	/*--------------------------------------------------------------------
	 * REVERSE forces a switch to mono mode
	 */
	switch (n)
	{
	case REVERSE:
	case NORMAL:
		if (CUR_SCRN_MODE(td) == SCRN_MODE_COLOR)
		{
			xvt_log_output_seq(td, "switch to mono mode");
			xvt_term_set_scrn_mode(td, SCRN_MODE_MONO);
		}
		break;
	}

	/*--------------------------------------------------------------------
	 * change attribute
	 */
	if (n == 0)
	{
		CUR_SCRN_AT(td) = NORMAL;

		xvt_log_output_seq(td, "all attributes off");
	}
	else
	{
		if (set)
			CUR_SCRN_AT(td) |=  n;
		else
			CUR_SCRN_AT(td) &= ~n;

		{
			const char *at = "?";

			switch (n)
			{
			case BOLD:		at = "bold";		break;
			case UNDERLINE:	at = "underline";	break;
			case BLINK:		at = "blink";		break;
			case REVERSE:	at = "reverse";		break;
			}

			xvt_log_output_seq(td, "attr %s %s", at,
				set ? "on" : "off");
		}
	}
}

/*------------------------------------------------------------------------
 * SGR_FG	set foreground color
 */
static void xvt_vt100_SGR_FG (TERMDATA *td, int fg)
{
	/*--------------------------------------------------------------------
	 * now switch to color mode if currently in mono
	 */
	if (CUR_SCRN_MODE(td) == SCRN_MODE_MONO)
	{
		xvt_log_output_seq(td, "switch to color mode");
		xvt_term_set_scrn_mode(td, SCRN_MODE_COLOR);
	}

	/*--------------------------------------------------------------------
	 * now if actually in color, set it
	 */
	if (CUR_SCRN_MODE(td) == SCRN_MODE_COLOR)
	{
		CUR_SCRN_FG(td) = color_code(fg);
	}

	/*--------------------------------------------------------------------
	 * debug output
	 */
	xvt_log_output_seq(td, "fg:%s", color_name(fg));
}

/*------------------------------------------------------------------------
 * SGR_BG	set background color
 */
static void xvt_vt100_SGR_BG (TERMDATA *td, int bg)
{
	/*--------------------------------------------------------------------
	 * now switch to color mode if currently in mono
	 */
	if (CUR_SCRN_MODE(td) == SCRN_MODE_MONO)
	{
		xvt_log_output_seq(td, "switch to color mode");
		xvt_term_set_scrn_mode(td, SCRN_MODE_COLOR);
	}

	/*--------------------------------------------------------------------
	 * now if actually in color, set it
	 */
	if (CUR_SCRN_MODE(td) == SCRN_MODE_COLOR)
	{
		CUR_SCRN_BG(td) = color_code(bg);
	}

	/*--------------------------------------------------------------------
	 * debug output
	 */
	xvt_log_output_seq(td, "bg:%s", color_name(bg));
}

/*------------------------------------------------------------------------
 * SGR		set graphic rendition (color/attributes)
 */
void xvt_vt100_CSI_SGR (TERMDATA *td)
{
	int i;

	for (i=0; i<=te->n_index; i++)
	{
		int n = te->n[i];

		switch (n)
		{
		/*----------------------------------------------------------------
		 * all mono attributes off
		 */
		case 0:
			xvt_vt100_SGR_AT(td, TRUE, 0);
			break;

		/*----------------------------------------------------------------
		 * mono attributes on
		 */
		case 1:		xvt_vt100_SGR_AT(td, TRUE,  BOLD);		break;
		case 4:		xvt_vt100_SGR_AT(td, TRUE,  UNDERLINE);	break;
		case 5:		xvt_vt100_SGR_AT(td, TRUE,  BLINK);		break;
		case 7:		xvt_vt100_SGR_AT(td, TRUE,  REVERSE);	break;

		/*----------------------------------------------------------------
		 * mono attributes off
		 */
		case 22:	xvt_vt100_SGR_AT(td, FALSE, BOLD);		break;
		case 24:	xvt_vt100_SGR_AT(td, FALSE, UNDERLINE);	break;
		case 25:	xvt_vt100_SGR_AT(td, FALSE, BLINK);		break;
		case 27:	xvt_vt100_SGR_AT(td, FALSE, REVERSE);	break;

		/*----------------------------------------------------------------
		 * fg colors 0-7
		 */
		case 30:
		case 31:
		case 32:
		case 33:
		case 34:
		case 35:
		case 36:
		case 37:	xvt_vt100_SGR_FG(td, n - 30);			break;

		/*----------------------------------------------------------------
		 * default fg color
		 */
		case 39:	xvt_vt100_SGR_FG(td, XVT_CLR_FG);		break;

		/*----------------------------------------------------------------
		 * bg colors 0-7
		 */
		case 40:
		case 41:
		case 42:
		case 43:
		case 44:
		case 45:
		case 46:
		case 47:	xvt_vt100_SGR_BG(td, n - 40);			break;

		/*----------------------------------------------------------------
		 * default bg color
		 */
		case 49:	xvt_vt100_SGR_BG(td, XVT_CLR_BG);		break;

		/*----------------------------------------------------------------
		 * unknown
		 */
		default:
			xvt_log_output_seq(td, "invalid SGR: %d", n);
			break;
		}
	}
}

/*------------------------------------------------------------------------
 * DECSTBM		set scrolling region
 */
void xvt_vt100_CSI_DECSTBM (TERMDATA *td)
{
	int p1 = te->n[0];
	int p2 = te->n[1];

	if (--p1 < 0)	p1 = 0;
	if (--p2 < 0)	p2 = 0;

	xvt_log_output_seq(td, "set-scroll-region: %d,%d", p1, p2);

	xvt_scrn_set_margins(CUR_SCRN(td), p1, p2);
}

/*------------------------------------------------------------------------
 * TERM			process terminal cmds
 */
void xvt_vt100_CSI_TERM (TERMDATA *td)
{
	int n = te->n[0];

	switch (n)
	{
	case 1:					/* un-iconify the window		*/
		xvt_log_output_seq(td, "iconify-off");
		xvt_term_set_icon_state(td, FALSE);
		break;

	case 2:					/* iconify the window			*/
		xvt_log_output_seq(td, "iconify-on");
		xvt_term_set_icon_state(td, TRUE);
		break;

	case 3:					/* set window pos to n[1], n[2]	*/
		if (te->n_index == 2)
		{
			xvt_log_output_seq(td, "set window position to (%d x %d)",
				te->n[1], te->n[2]);
			xvt_term_set_position(td, te->n[1], te->n[2]);
		}
		break;

	case 4:					/* set w/h to n[2], n[1] pixels	*/
		if (te->n_index == 2)
		{
			xvt_log_output_seq(td, "set window size to (%d x %d) pixels",
				te->n[2], te->n[1]);
			xvt_term_set_size(td, te->n[2], te->n[1]);
		}
		break;

	case 5:					/* raise window					*/
		xvt_log_output_seq(td, "raise window");
		xvt_term_raise(td);
		break;

	case 6:					/* lower window					*/
		xvt_log_output_seq(td, "lower window");
		xvt_term_lower(td);
		break;

	case 7:					/* refresh window				*/
		xvt_log_output_seq(td, "refresh window");
		xvt_term_repaint(td);
		break;

	case 8:					/* set w/h to n[2], n[1] chars	*/
		if (te->n_index == 2)
		{
			xvt_log_output_seq(td, "set window size to (%d x %d) chars",
				te->n[2], te->n[1]);
			xvt_term_resize(td, te->n[2], te->n[1]);
		}
		break;

	case XVT_CSI_TERM_CUR:	/* XVT: set cursor type to n[1]	*/
		n = te->n[1];
		xvt_log_output_seq(td, "set cursor type %d", n);
		xvt_term_set_x11_cursor(td, n);
		break;

	case 11:				/* report win state				*/
		xvt_log_output_seq(td, "report window state");
		xvt_term_write_report(td, "\033[%dt",
			td->tw->iconified ? 2 : 1);
		break;

	case 13:				/* report win position			*/
		{
			int x, y;

			xvt_log_output_seq(td, "report window position");
			xvt_term_get_position(td, &x, &y);
			xvt_term_write_report(td, "\033[3;%d;%dt", x, y);
		}
		break;

	case 14:				/* report win size (pixels)		*/
		{
			int w, h;

			xvt_log_output_seq(td, "report window size (pixels)");
			xvt_term_get_size(td, &w, &h);
			xvt_term_write_report(td, "\033[4;%d;%dt", w, h);
		}
		break;

	case 18:				/* report win size (chars)		*/
		xvt_log_output_seq(td, "report window size (chars)");
		xvt_term_write_report(td, "\033[8;%d;%dt",
			td->xd->cur_rows, td->xd->cur_cols);
		break;

	case 20:				/* report ico label				*/
		xvt_log_output_seq(td, "report icon label");
		xvt_term_write_report(td, "\033]L%s\033\\",
			td->ico_label);
		break;

	case 21:				/* report win label				*/
		xvt_log_output_seq(td, "report window label");
		xvt_term_write_report(td, "\033]l%s\033\\",
			td->org_label);
		break;

	default:
		if (n >= 24)
		{
							/* set win height to n rows		*/
		
			xvt_log_output_seq(td, "set window height to %d", n);
			xvt_term_resize(td, -1, n);
		}
		break;
	}
}

/*------------------------------------------------------------------------
 * PRT			process print cmds
 */
void xvt_vt100_CSI_PRT (TERMDATA *td)
{
	int n = te->n[0];

	switch (n)
	{
	case 0:
		xvt_log_output_seq(td, "print-screen");
		xvt_term_screen_print(td);
		break;

	case 4:
		/*----------------------------------------------------------------
		 * we should never get here ...
		 */
		xvt_log_output_seq(td, "stop print pass-thru");
		xvt_term_printer_close(td);
		break;

	case 5:
		xvt_log_output_seq(td, "start print pass-thru");
		xvt_term_printer_open(td);
		break;

	default:
		xvt_log_output_seq(td, "invalid PRT cmd %d", n);
		break;
	}
}

/*------------------------------------------------------------------------
 * DA			device attributes report
 */
void xvt_vt100_CSI_DA (TERMDATA *td)
{
	xvt_log_output_seq(td, "send device attribtes");
	xvt_term_write_report(td, "\033[?1;2c");
}

/*------------------------------------------------------------------------
 * DSR			device status report
 */
void xvt_vt100_CSI_DSR (TERMDATA *td)
{
	int n = te->n[0];

	switch (n)
	{
	case 5:						/* status report */
		xvt_log_output_seq(td, "status report");
		xvt_term_write_report(td, "\033[0n");
		break;

	case 6:						/* report cursor position */
		xvt_log_output_seq(td, "report cursor-position");
		xvt_term_write_report(td, "\033[%d;%dR",
			CUR_SCRN_Y(td), CUR_SCRN_X(td));
		break;

	case 7:						/* report display name */
		xvt_log_output_seq(td, "report display name");
		xvt_term_write_report(td, "%s\n", td->xd->display);
		break;

	case 8:						/* report version */
		xvt_log_output_seq(td, "report version");
		xvt_term_write_report(td, "%s\n", XVT_APPCLASS);
		break;

	default:
		xvt_log_output_seq(td, "invalid DSR cmd %d", n);
		break;
	}
}

/*------------------------------------------------------------------------
 * TBC			tab clear
 */
void xvt_vt100_CSI_TBC (TERMDATA *td)
{
	int n = te->n[0];

	switch (n)
	{
	case 0:			/* clear tab at cur col */
		xvt_log_output_seq(td, "clear tab at cur col");
		xvt_scrn_tab_set(CUR_SCRN(td), SCRN_TAB_CLR_CUR);
		break;

	case 3:			/* clear all tabs */
		xvt_log_output_seq(td, "clear all tabs");
		xvt_scrn_tab_set(CUR_SCRN(td), SCRN_TAB_CLR_ALL);
		break;

	default:
		xvt_log_output_seq(td, "invalid tab clear mode %d", n);
		break;
	}
}

/*------------------------------------------------------------------------
 * TBS			tab set
 */
void xvt_vt100_CSI_TBS (TERMDATA *td)
{
	int n = te->n[0];

	switch (n)
	{
	case 0:			/* set tab at cur col */
		xvt_log_output_seq(td, "set tab at cur col");
		xvt_scrn_tab_set(CUR_SCRN(td), SCRN_TAB_SET_CUR);
		break;

	case 2:			/* clear tab at cur col */
		xvt_log_output_seq(td, "clear tab at cur col");
		xvt_scrn_tab_set(CUR_SCRN(td), SCRN_TAB_CLR_CUR);
		break;

	case 5:			/* clear all tabs */
		xvt_log_output_seq(td, "clear all tabs");
		xvt_scrn_tab_set(CUR_SCRN(td), SCRN_TAB_CLR_ALL);
		break;

	default:
		xvt_log_output_seq(td, "invalid tab set mode %d", n);
		break;
	}
}

/*------------------------------------------------------------------------
 * CHT			tab move right
 */
void xvt_vt100_CSI_CHT (TERMDATA *td)
{
	int n = te->n[0];
	int x;

	if (n <= 0)
		n = 1;

	xvt_log_output_seq(td, "move %d tabstop%s right", n, PLURAL(n));

	for (x=CUR_SCRN_X(td) + 1; n>0 && x < CUR_SCRN_COLS(td); x++)
	{
		if (CUR_SCRN_TABSTOP(td, x))
			n--;
	}

	if (x < CUR_SCRN_COLS(td))
		CUR_SCRN_X(td) = x;
}

/*------------------------------------------------------------------------
 * CBT			tab move left
 */
void xvt_vt100_CSI_CBT (TERMDATA *td)
{
	int n = te->n[0];
	int x;

	if (n <= 0)
		n = 1;

	xvt_log_output_seq(td, "move %d tabstop%s right", n, PLURAL(n));

	for (x=CUR_SCRN_X(td) - 1; n>0 && x >= 0; x--)
	{
		if (CUR_SCRN_TABSTOP(td, x))
			n--;
	}

	if (x > 0)
		CUR_SCRN_X(td) = x;
}

/*------------------------------------------------------------------------
 * SU			scroll up n lines
 */
void xvt_vt100_CSI_SU (TERMDATA *td)
{
	int n = te->n[0];

	if (n <= 0)
		n = 1;

	xvt_log_output_seq(td, "scroll up %d line%s", n, PLURAL(n));
	xvt_curscrn_scroll(td, -n);
}

/*------------------------------------------------------------------------
 * SD			scroll down n lines
 */
void xvt_vt100_CSI_SD (TERMDATA *td)
{
	int n = te->n[0];

	if (n <= 0)
		n = 1;

	xvt_log_output_seq(td, "scroll down %d line%s", n, PLURAL(n));
	xvt_curscrn_scroll(td, n);
}

/*========================================================================
 * PRI mode commands
 *========================================================================
 */

/*------------------------------------------------------------------------
 * DECSET 	set private mode
 */
void xvt_vt100_PRI_DECSET (TERMDATA *td)
{
	int i;

	for (i=0; i<=te->n_index; i++)
	{
		int n = te->n[i];

		switch (n)
		{
		case 1:					/* application cursor keys on */
			xvt_log_output_seq(td, "application cursor keys on");
			xvt_vt100_DECCKM(td, TRUE);
			break;

		case 2:					/* set charsets G0-G3 as USASCII */
			xvt_log_output_seq(td, "set char sets G0-G3 as USASCII");
			td->g[CHAR_SET_G0] = GCS_US;
			td->g[CHAR_SET_G1] = GCS_US;
			td->g[CHAR_SET_G2] = GCS_US;
			td->g[CHAR_SET_G3] = GCS_US;
			break;

		case 3:					/* 132 column mode */
			xvt_log_output_seq(td, "set 132 column mode");
			if (td->xd->c132)
				xvt_term_resize(td, 132, -1);
			break;

		case 4:					/* smooth scroll */
			xvt_log_output_seq(td, "smooth scrolling: ignored");
			break;

		case 5:					/* reverse video on */
			xvt_log_output_seq(td, "reverse video on");
			xvt_term_set_rev_video(td, TRUE);
			break;

		case 6:				/* origin mode on */
			xvt_log_output_seq(td, "origin mode on");
			CUR_SCRN_ORG(td) = TRUE;
			break;

		case 7:				/* auto wrap-around on */
			td->xd->autowrap = TRUE;
			xvt_log_output_seq(td, "auto wrap-around on");
			break;

		case 8:				/* auto-repeat keys on */
			xvt_log_output_seq(td, "set auto-repeat keys on: ignored");
			break;

		case 25:			/* cursor on */
			if (! CUR_SCRN_CSV(td))
			{
				CUR_SCRN_CSV(td) = TRUE;
				xvt_term_set_txt_cursor(td, TRUE);
			}
			xvt_log_output_seq(td, "cursor on");
			break;

		case 38:				/* enter Tektronix mode */
			xvt_log_output_seq(td, "enter Tektronix mode: ignored");
			break;

		case 40:				/* 80-132 switch on */
			td->xd->c132 = TRUE;
			xvt_log_output_seq(td, "80-132 switch on");
			break;

		case 44:				/* margin bell on */
			td->xd->mbell = TRUE;
			xvt_log_output_seq(td, "margin bell on");
			break;

		case 45:				/* reverse wrap-around on */
			td->xd->revwrap = TRUE;
			xvt_log_output_seq(td, "reverse wrap-around on");
			break;

		case 47:                 /* alt-buf on */
			if (CUR_SCRN(td) == td->reg_scrn)
			{
				CUR_SCRN(td) = td->alt_scrn;
				xvt_term_repaint(td);
			}
			xvt_log_output_seq(td, "alt-buf on");
			break;

		case 66:				/* application keypad keys on */
			xvt_log_output_seq(td, "application keypad on");
			xvt_vt100_DECNKM(td, TRUE);
			break;

		case 1000:				/* mouse on */
			td->enable_mouse_events		= TRUE;
			xvt_log_output_seq(td, "mouse on");
			break;

		case 1010:				/* scroll to bottom on output	*/
			td->xd->sb_scrout = TRUE;
			xvt_log_output_seq(td, "scroll to bottom on output");
			break;

		case 1011:				/* scroll to bottom on keypress	*/
			td->xd->sb_scrkey = TRUE;
			xvt_log_output_seq(td, "scroll to bottom on keypress");
			break;

		/*----------------------------------------------------------------
		 * XVT-specific cmds
		 */
		case XVT_PRI_KBD_OFF:			/* XVT: kbd-map off */
		case XVT_PRI_KBD_ONE:			/* XVT: kbd-map = 1 */
		case XVT_PRI_KBD_TWO:			/* XVT: kbd-map = 2 */
			xvt_vt100_kbdmap(td, n - XVT_PRI_KBD_OFF);
			break;

		case XVT_PRI_LBLINFO:
			td->disp_info = TRUE;
			xvt_term_set_win_title(td, td->win_label);
			break;

		case XVT_PRI_LOG_EVT:
		case XVT_PRI_LOG_ACT:
		case XVT_PRI_LOG_OUT:
		case XVT_PRI_LOG_SCR:
		case XVT_PRI_LOG_INP:
		case XVT_PRI_LOG_KEY:
			n -= XVT_PRI_LOGGING;
			xvt_log_output_seq(td, "log %d on", n);
			xvt_log_set(td, n, TRUE);
			if (td->disp_info)
				xvt_term_set_win_title(td, td->win_label);
			break;

		default:
			xvt_log_output_seq(td, "invalid PRI set opt %d", n);
			break;
		}
	}
}

/*------------------------------------------------------------------------
 * DECRST	reset private mode
 */
void xvt_vt100_PRI_DECRST (TERMDATA *td)
{
	int i;

	for (i=0; i<=te->n_index; i++)
	{
		int n = te->n[i];

		switch (n)
		{
		case 1:					/* application cursor keys */
			xvt_log_output_seq(td, "application cursor keys off");
			xvt_vt100_DECCKM(td, FALSE);
			break;

		case 3:					/* 80 column mode */
			xvt_log_output_seq(td, "set 80 column mode");
			if (td->xd->c132)
				xvt_term_resize(td, 80, -1);
			break;

		case 4:					/* jump scroll */
			xvt_log_output_seq(td, "jump scrolling: ignored");
			break;

		case 5:					/* reverse video off */
			xvt_log_output_seq(td, "reverse video off");
			xvt_term_set_rev_video(td, FALSE);
			break;

		case 6:				/* origin mode off */
			xvt_log_output_seq(td, "origin mode off");
			CUR_SCRN_ORG(td) = FALSE;
			break;

		case 7:					/* auto wrap-around off */
			td->xd->autowrap = FALSE;
			xvt_log_output_seq(td, "auto wrap-around off");
			break;

		case 8:				/* auto-repeat keys off */
			xvt_log_output_seq(td, "set auto-repeat keys off: ignored");
			break;

		case 25:				/* cursor off */
			if (CUR_SCRN_CSV(td))
			{
				xvt_term_set_txt_cursor(td, FALSE);
				CUR_SCRN_CSV(td) = FALSE;
			}
			xvt_log_output_seq(td, "cursor off");
			break;

		case 40:				/* 80-132 switch off */
			td->xd->c132 = FALSE;
			xvt_log_output_seq(td, "80-132 switch off");
			break;

		case 44:				/* margin bell off */
			td->xd->mbell = FALSE;
			xvt_log_output_seq(td, "margin bell off");
			break;

		case 45:				/* reverse wrap-around off */
			td->xd->revwrap = FALSE;
			xvt_log_output_seq(td, "reverse wrap-around off");
			break;

		case 47:                 /* alt-buf off */
			if (CUR_SCRN(td) == td->alt_scrn)
			{
				CUR_SCRN(td) = td->reg_scrn;
				xvt_term_repaint(td);
			}
			xvt_log_output_seq(td, "alt-buf off");
			break;

		case 66:				/* application keypad keys off */
			xvt_log_output_seq(td, "application keypad off");
			xvt_vt100_DECNKM(td, FALSE);
			break;

		case 1000:			/* mouse off */
			td->enable_mouse_events		= FALSE;
			xvt_log_output_seq(td, "mouse off");
			break;

		case 1010:				/* no scroll to bottom on output	*/
			td->xd->sb_scrout = FALSE;
			xvt_log_output_seq(td, "no scroll to bottom on output");
			break;

		case 1011:				/* no scroll to bottom on keypress	*/
			td->xd->sb_scrkey = FALSE;
			xvt_log_output_seq(td, "no scroll to bottom on keypress");
			break;

		/*----------------------------------------------------------------
		 * XVT-specific cmds
		 */
		case XVT_PRI_KBD_OFF:			/* XVT: kbd-map off */
		case XVT_PRI_KBD_ONE:			/* XVT: kbd-map = 1 */
		case XVT_PRI_KBD_TWO:			/* XVT: kbd-map = 2 */
			xvt_vt100_kbdmap(td, n - XVT_PRI_KBD_OFF);
			break;

		case XVT_PRI_LBLINFO:
			td->disp_info = FALSE;
			xvt_term_set_win_title(td, td->win_label);
			break;

		case XVT_PRI_LOG_EVT:
		case XVT_PRI_LOG_ACT:
		case XVT_PRI_LOG_OUT:
		case XVT_PRI_LOG_SCR:
		case XVT_PRI_LOG_KEY:
			n -= XVT_PRI_LOGGING;
			xvt_log_output_seq(td, "log %d off", n);
			xvt_log_set(td, n, FALSE);
			if (td->disp_info)
				xvt_term_set_win_title(td, td->win_label);
			break;

		default:
			xvt_log_output_seq(td, "invalid PRI reset opt %d", n);
			break;
		}
	}
}

/*========================================================================
 * XVT mode commands
 *========================================================================
 */

/*------------------------------------------------------------------------
 * XVT_KBDMAP	set keyboard mapping
 */
void xvt_vt100_XVT_KBDMAP (TERMDATA *td)
{
	xvt_vt100_kbdmap(td, te->n[0]);
}

/*========================================================================
 * STP mode commands
 *========================================================================
 */

/*------------------------------------------------------------------------
 * STP_LABEL	STP labels
 */
void xvt_vt100_STP_LABEL (TERMDATA *td)
{
	int n = te->n[0];

	/*--------------------------------------------------------------------
	 * delete the trailing ^G
	 */
	te->data_chars[--te->data_cnt] = 0;

	/*--------------------------------------------------------------------
	 * now process the data
	 */
	switch (n)
	{
	case 0:
		xvt_log_output_seq(td, "xterm-label:%s", te->data_chars);
		strcpy(td->win_label, te->data_chars);
		strcpy(td->ico_label, te->data_chars);
		xvt_term_set_win_title(td, td->win_label);
		xvt_term_set_ico_title(td, td->ico_label);
		break;

	case 1:
		xvt_log_output_seq(td, "xterm-icon-label:%s", te->data_chars);
		strcpy(td->ico_label, te->data_chars);
		xvt_term_set_ico_title(td, td->ico_label);
		break;

	case 2:
		xvt_log_output_seq(td, "xterm-win-label:%s", te->data_chars);
		strcpy(td->win_label, te->data_chars);
		xvt_term_set_win_title(td, td->win_label);
		break;

	case 50:
		n = atoi(te->data_chars);
		xvt_log_output_seq(td, "xterm-font-no:%s", te->data_chars);
		xvt_term_set_font_no(td, n);
		break;

	default:
		xvt_log_output_seq(td, "unknown STP mode %d", n);
		break;
	}
}

/*========================================================================
 * DEV mode commands
 *========================================================================
 */
