/*------------------------------------------------------------------------
 * vt100 command table
 */
#include "xvtcommon.h"
#include "xvtvt100.h"

/*------------------------------------------------------------------------
 * TOP esc-seq cmds
 */
static const ESC_DATA	xvt_vt100_esctbl_top[] =
{
/*	  code	mode		type		rtn					desc				*/

	{ '(',	ESC_GCS,	0,			0,					0					},
	{ ')',	ESC_GCS,	1,			0,					0					},
	{ '*',	ESC_GCS,	2,			0,					0					},
	{ '+',	ESC_GCS,	3,			0,					0					},

	{ '[',	ESC_CSI,	-1,			0,					0					},
	{ ']',	ESC_STP,	-1,			0,					0					},
	{ '#',	ESC_DEC,	-1,			0,					0					},

	{ 'P',	ESC_DEV,	DEV_DCS,	0,					0					},
	{ '^',	ESC_DEV,	DEV_PM,		0,					0					},
	{ '_',	ESC_DEV,	DEV_APC,	0,					0					},

	{ '7',	-1,			-1,			xvt_vt100_TOP_DECSC,	0				},
	{ '8',	-1,			-1,			xvt_vt100_TOP_DECRC,	0				},
	{ '>',	-1,			-1,			xvt_vt100_TOP_DECPAM,	0				},
	{ '=',	-1,			-1,			xvt_vt100_TOP_DECPNM,	0				},
	{ 'c',	-1,			-1,			xvt_vt100_TOP_RIS,		0				},
	{ 'n',	-1,			-1,			xvt_vt100_TOP_LS2,		0				},
	{ 'o',	-1,			-1,			xvt_vt100_TOP_LS3,		0				},
	{ '\\',	-1,			-1,			xvt_vt100_TOP_DEVEND,	0				},
	{ 'Z',	-1,			-1,			xvt_vt100_TOP_DA,		0				},
	{ 'F',	-1,			-1,			xvt_vt100_TOP_CLL,		0				},
	{ 'H',	-1,			-1,			xvt_vt100_TOP_HTS,		0				},
	{ 'N',	-1,			-1,			xvt_vt100_TOP_SS2,		0				},
	{ 'O',	-1,			-1,			xvt_vt100_TOP_SS3,		0				},
	{ 'D',	-1,			-1,			xvt_vt100_TOP_IND,		0				},
	{ 'M',	-1,			-1,			xvt_vt100_TOP_RI,		0				},
	{ 'E',	-1,			-1,			xvt_vt100_TOP_NEL,		0				},
	{ 'l',	-1,			-1,			xvt_vt100_TOP_MEML,		0				},
	{ 'm',	-1,			-1,			xvt_vt100_TOP_MEMU,		0				},

	{ '6',	-1,			-1,			0,					"back-index"		},
	{ '9',	-1,			-1,			0,					"fwd-index"			},
	{ '|',	-1,			-1,			0,					"invoke-G3-as-GR"	},
	{ '}',	-1,			-1,			0,					"invoke-G2-as-GR"	},
	{ '~',	-1,			-1,			0,					"invoke-G1-as-GR"	},

	{ -1, 0, 0, 0, 0 }
};

/*------------------------------------------------------------------------
 * DEC esc-seq cmds
 */
static const ESC_DATA	xvt_vt100_esctbl_dec[] =
{
/*	  code	mode		type		rtn					desc				*/

	{ '8',	-1,			-1,			xvt_vt100_DEC_DECALN,	0				},

	{ '3',	-1,			-1,			0,					"dbl-height, top"	},
	{ '4',	-1,			-1,			0,					"dbl-height, bot"	},
	{ '5',	-1,			-1,			0,					"dbl-width"			},
	{ '6',	-1,			-1,			0,					"dbl-width"			},

	{ -1, 0, 0, 0, 0 }
};

/*------------------------------------------------------------------------
 * GCS esc-seq cmds
 */
static const ESC_DATA	xvt_vt100_esctbl_gcs[] =
{
/*	  code	mode		type		rtn					desc				*/

	{ 0,	-1,			-1,			xvt_vt100_GCS_CODE,		0				},

	{ -1, 0, 0, 0, 0 }
};

/*------------------------------------------------------------------------
 * CSI esc-seq cmds
 */
static const ESC_DATA	xvt_vt100_esctbl_csi[] =
{
/*	  code	mode		type		rtn					desc				*/

	{ '?',	ESC_PRI,	-1,			0,					0					},

	{ '`',	-1,			-1,			xvt_vt100_CSI_HPA,	0					},
	{ 'G',	-1,			-1,			xvt_vt100_CSI_HPA,	0					},
	{ 'd',	-1,			-1,			xvt_vt100_CSI_VPA,	0					},
	{ 'f',	-1,			-1,			xvt_vt100_CSI_CUP,	0					},
	{ 'H',	-1,			-1,			xvt_vt100_CSI_CUP,	0					},
	{ 'J',	-1,			-1,			xvt_vt100_CSI_ED,	0					},
	{ 'K',	-1,			-1,			xvt_vt100_CSI_EL,	0					},
	{ 'e',	-1,			-1,			xvt_vt100_CSI_CUU,	0					},
	{ 'A',	-1,			-1,			xvt_vt100_CSI_CUU,	0					},
	{ 'B',	-1,			-1,			xvt_vt100_CSI_CUD,	0					},
	{ 'D',	-1,			-1,			xvt_vt100_CSI_CUB,	0					},
	{ 'a',	-1,			-1,			xvt_vt100_CSI_CUF,	0					},
	{ 'C',	-1,			-1,			xvt_vt100_CSI_CUF,	0					},
	{ '@',	-1,			-1,			xvt_vt100_CSI_ICH,	0					},
	{ 'P',	-1,			-1,			xvt_vt100_CSI_DCH,	0					},
	{ 'L',	-1,			-1,			xvt_vt100_CSI_AL,	0					},
	{ 'M',	-1,			-1,			xvt_vt100_CSI_DL,	0					},
	{ 'h',	-1,			-1,			xvt_vt100_CSI_SM,	0					},
	{ 'l',	-1,			-1,			xvt_vt100_CSI_RM,	0					},
	{ 'r',	-1,			-1,			xvt_vt100_CSI_DECSTBM,	0				},
	{ 'm',	-1,			-1,			xvt_vt100_CSI_SGR,	0					},
	{ 't',	-1,			-1,			xvt_vt100_CSI_TERM,	0					},
	{ 'i',	-1,			-1,			xvt_vt100_CSI_PRT,	0					},
	{ 'c',	-1,			-1,			xvt_vt100_CSI_DA,	0					},
	{ 'n',	-1,			-1,			xvt_vt100_CSI_DSR,	0					},
	{ 'E',	-1,			-1,			xvt_vt100_CSI_CNL,	0					},
	{ 'F',	-1,			-1,			xvt_vt100_CSI_CPL,	0					},
	{ '`',	-1,			-1,			xvt_vt100_CSI_HPA,	0					},
	{ 'G',	-1,			-1,			xvt_vt100_CSI_HPA,	0					},
	{ 'd',	-1,			-1,			xvt_vt100_CSI_VPA,	0					},
	{ 'X',	-1,			-1,			xvt_vt100_CSI_ECH,	0					},
	{ 'g',	-1,			-1,			xvt_vt100_CSI_TBC,	0					},
	{ 'x',	-1,			-1,			xvt_vt100_CSI_TBS,	0					},
	{ 'I',	-1,			-1,			xvt_vt100_CSI_CHT,	0					},
	{ 'Z',	-1,			-1,			xvt_vt100_CSI_CBT,	0					},
	{ '^',	-1,			-1,			xvt_vt100_CSI_SU,	0					},
	{ 'T',	-1,			-1,			xvt_vt100_CSI_SU,	0					},
	{ 'S',	-1,			-1,			xvt_vt100_CSI_SD,	0					},

	{ -1, 0, 0, 0, 0 }
};

/*------------------------------------------------------------------------
 * PRI esc-seq cmds
 */
static const ESC_DATA	xvt_vt100_esctbl_pri[] =
{
/*	  code	mode		type		rtn					desc				*/

	{ 'l',	-1,			-1,			xvt_vt100_PRI_DECRST,	0				},
	{ 'h',	-1,			-1,			xvt_vt100_PRI_DECSET,	0				},

	{ 'r',	-1,			-1,			0,					"rest-pri-values"	},
	{ 's',	-1,			-1,			0,					"save-pri-values"	},
	{ 't',	-1,			-1,			0,					"toggle-pri-values"	},

	{ -1, 0, 0, 0, 0 }
};

/*------------------------------------------------------------------------
 * STP esc-seq cmds
 */
static const ESC_DATA	xvt_vt100_esctbl_stp[] =
{
/*	  code	mode		type		rtn					desc				*/

	{ 0x07,	-1,			-1,			xvt_vt100_STP_LABEL,	0				},
	{ 0,	-1,			-1,			0,					0					},

	{ -1, 0, 0, 0, 0 }
};

/*------------------------------------------------------------------------
 * DEV esc-seq cmds
 */
static const ESC_DATA	xvt_vt100_esctbl_dev[] =
{
/*	  code	mode		type		rtn					desc				*/

	{ 0,	-1,			-1,			0,					0					},

	{ -1, 0, 0, 0, 0 }
};

/*------------------------------------------------------------------------
 * escape cmd table
 */
static const ESC_CMD	xvt_vt100_esctbl[] =
{
/*	  name		cmd-tbl					data			params	*/

	{ "TOP",	xvt_vt100_esctbl_top,	DATA_NONE,		0		},
	{ "DEC",	xvt_vt100_esctbl_dec,	DATA_NONE,		0		},
	{ "GCS",	xvt_vt100_esctbl_gcs,	DATA_NONE,		0		},
	{ "CSI",	xvt_vt100_esctbl_csi,	DATA_NONE,		6		},
	{ "PRI",	xvt_vt100_esctbl_pri,	DATA_NONE,		2		},
	{ "STP",	xvt_vt100_esctbl_stp,	DATA_CLEAN,		1		},
	{ "DEV",	xvt_vt100_esctbl_dev,	DATA_ESCAPE,	0		}
};

/*------------------------------------------------------------------------
 * control-char cmd table
 */
static const CTL_CMD	xvt_vt100_ctltbl[] =
{
	/*--------------------------------------------------------------------
	 * these are pretty standard
	 */
	{ 0x07,	xvt_vt100_CTL_BL	},
	{ 0x08,	xvt_vt100_CTL_BS	},
	{ 0x09,	xvt_vt100_CTL_HT	},
	{ 0x0a,	xvt_vt100_CTL_LF	},
	{ 0x0b,	xvt_vt100_CTL_LF	},
	{ 0x0c,	xvt_vt100_CTL_FF	},
	{ 0x0d,	xvt_vt100_CTL_CR	},

	/*--------------------------------------------------------------------
	 * these are vt100-specific
	 */
	{ 0x05,	xvt_vt100_CTL_ENQ	},
	{ 0x0e,	xvt_vt100_CTL_SO	},
	{ 0x0f,	xvt_vt100_CTL_SI	},

	{ -1, 0 }
};

/*------------------------------------------------------------------------
 * vt100 cmd table
 */
const CMD_TBL	xvt_vt100_cmdtbl =
{
	"vt100",
	xvt_vt100_init,
	xvt_vt100_char_output,
	xvt_vt100_char_print,
	xvt_vt100_ctltbl,
	xvt_vt100_esctbl
};
