/*------------------------------------------------------------------------
 * xvt X routines
 */
#include "xvtcommon.h"
#include "xvtx.h"

/*------------------------------------------------------------------------
 * Color arrays
 *
 * These arrays contain a default RGB value in case we can't parse the
 * color name, and a "name" describing the color entry.
 */
#define ALL3		( DoRed | DoGreen | DoBlue )
#define PIXL(n)		( (n) << 8 )
#define CMAP(r,g,b)	{ 0, PIXL(r), PIXL(g), PIXL(b), ALL3, 0 }

struct x_color
{
	XColor			clr_xcolor;		/* default XColor entry	*/
	const char *	clr_name;		/* color name			*/
};
typedef struct x_color	X_COLOR;

static const X_COLOR xvt_w_colors[] =
{
	/*--------------------------------------------------------------------
	 * screen colors
	 */
	{ CMAP(0x00, 0x00, 0x00),	"black"			},
	{ CMAP(0xcd, 0x00, 0x00),	"red"			},
	{ CMAP(0x00, 0xcd, 0x00),	"green"			},
	{ CMAP(0xcd, 0xcd, 0x00),	"yellow"		},
	{ CMAP(0x00, 0x00, 0xcd),	"blue"			},
	{ CMAP(0xcd, 0x00, 0xcd),	"magenta"		},
	{ CMAP(0x00, 0xcd, 0xcd),	"cyan"			},
	{ CMAP(0xfa, 0xeb, 0xd7),	"white"			},
 
	{ CMAP(0x40, 0x40, 0x40),	"grey"			},
	{ CMAP(0xff, 0x00, 0x00),	"lt-red"		},
	{ CMAP(0x00, 0xff, 0x00),	"lt-green"		},
	{ CMAP(0xff, 0xff, 0x00),	"lt-yellow"		},
	{ CMAP(0x00, 0x00, 0xff),	"lt-blue"		},
	{ CMAP(0xff, 0x00, 0xff),	"lt-magenta"	},
	{ CMAP(0x00, 0xff, 0xff),	"lt-cyan"		},
	{ CMAP(0xff, 0xff, 0xff),	"lt-white"		},

	/*--------------------------------------------------------------------
	 * other colors
	 */
	{ CMAP(0x00, 0x00, 0x00),	"foreground"	},
	{ CMAP(0xff, 0xff, 0xff),	"background"	},
	{ CMAP(0x00, 0x00, 0x00),	"pointer"		},
	{ CMAP(0x00, 0x00, 0x00),	"border"		},
	{ CMAP(0x00, 0x00, 0x00),	"cursor"		},
	{ CMAP(0xff, 0xff, 0xff),	"scrollbar"		},
	{ CMAP(0x00, 0x00, 0x00),	"SB-trough"		},

	/*--------------------------------------------------------------------
	 * calculated colors
	 */
	{ CMAP(0x00, 0x00, 0x00),	"SB-top-shadow"	},
	{ CMAP(0x00, 0x00, 0x00),	"SB-bot-shadow"	}
};

/*------------------------------------------------------------------------
 * cursor types table
 */
struct pointer_types
{
	int				ptr_code;		/* type code	*/
	const char *	ptr_name;		/* pointer name	*/
};
typedef struct pointer_types CURSOR_TYPES;

static const CURSOR_TYPES xvt_w_pointers[] =
{
	{ XC_X_cursor,				"x-cursor"				},
	{ XC_arrow,					"arrow"					},
	{ XC_based_arrow_down,		"based-arrow-down"		},
	{ XC_based_arrow_up,		"based-arrow-up"		},
	{ XC_boat,					"boat"					},
	{ XC_bogosity,				"bogosity"				},
	{ XC_bottom_left_corner,	"bottom-left-corner"	},
	{ XC_bottom_right_corner,	"bottom-right-corner"	},
	{ XC_bottom_side,			"bottom-side"			},
	{ XC_bottom_tee,			"bottom-tee"			},
	{ XC_box_spiral,			"box-spiral"			},
	{ XC_center_ptr,			"center-ptr"			},
	{ XC_circle,				"circle"				},
	{ XC_clock,					"clock"					},
	{ XC_coffee_mug,			"coffee-mug"			},
	{ XC_cross,					"cross"					},
	{ XC_cross_reverse,			"cross-reverse"			},
	{ XC_crosshair,				"crosshair"				},
	{ XC_diamond_cross,			"diamond-cross"			},
	{ XC_dot,					"dot"					},
	{ XC_dotbox,				"dotbox"				},
	{ XC_double_arrow,			"double-arrow"			},
	{ XC_draft_large,			"draft-large"			},
	{ XC_draft_small,			"draft-small"			},
	{ XC_draped_box,			"draped-box"			},
	{ XC_exchange,				"exchange"				},
	{ XC_fleur,					"fleur"					},
	{ XC_gobbler,				"gobbler"				},
	{ XC_gumby,					"gumby"					},
	{ XC_hand1,					"hand1"					},
	{ XC_hand2,					"hand2"					},
	{ XC_heart,					"heart"					},
	{ XC_icon,					"icon"					},
	{ XC_iron_cross,			"iron-cross"			},
	{ XC_left_ptr,				"left-ptr"				},
	{ XC_left_side,				"left-side"				},
	{ XC_left_tee,				"left-tee"				},
	{ XC_leftbutton,			"left-button"			},
	{ XC_ll_angle,				"ll-angle"				},
	{ XC_lr_angle,				"lr-angle"				},
	{ XC_man,					"man"					},
	{ XC_middlebutton,			"middle-button"			},
	{ XC_mouse,					"mouse"					},
	{ XC_pencil,				"pencil"				},
	{ XC_pirate,				"pirate"				},
	{ XC_plus,					"plus"					},
	{ XC_question_arrow,		"question-arrow"		},
	{ XC_right_ptr,				"right-ptr"				},
	{ XC_right_side,			"right-side"			},
	{ XC_right_tee,				"right-tee"				},
	{ XC_rightbutton,			"right-button"			},
	{ XC_rtl_logo,				"rtl-logo"				},
	{ XC_sailboat,				"sailboat"				},
	{ XC_sb_down_arrow,			"sb-down-arrow"			},
	{ XC_sb_h_double_arrow,		"sb-h-double-arrow"		},
	{ XC_sb_left_arrow,			"sb-left-arrow"			},
	{ XC_sb_right_arrow,		"sb-right-arrow"		},
	{ XC_sb_up_arrow,			"sb-up-arrow"			},
	{ XC_sb_v_double_arrow,		"sb-v-double-arrow"		},
	{ XC_shuttle,				"shuttle"				},
	{ XC_sizing,				"sizing"				},
	{ XC_spider,				"spider"				},
	{ XC_spraycan,				"spraycan"				},
	{ XC_star,					"star"					},
	{ XC_target,				"target"				},
	{ XC_tcross,				"tcross"				},
	{ XC_top_left_arrow,		"top-left-arrow"		},
	{ XC_top_left_corner,		"top-left-corner"		},
	{ XC_top_right_corner,		"top-right-corner"		},
	{ XC_top_side,				"top-side"				},
	{ XC_top_tee,				"top-tee"				},
	{ XC_trek,					"trek"					},
	{ XC_ul_angle,				"ul-angle"				},
	{ XC_umbrella,				"umbrella"				},
	{ XC_ur_angle,				"ur-angle"				},
	{ XC_watch,					"watch"					},
	{ XC_xterm,					"xterm"					},

	{ -1, 0 }
};

#define NUM_CURSORS	((sizeof(xvt_w_pointers) / sizeof(*xvt_w_pointers)) - 1)

/*------------------------------------------------------------------------
 * X error routines
 */
static int xvt_w_error_rtn_dummy (Display *display, XErrorEvent *ev)
{
	return (0);
}

static int xvt_w_error_rtn (Display *display, XErrorEvent *ev)
{
	fprintf(stderr, "XError: Request: %3d  Error: %2d\n",
		ev->request_code, ev->error_code);

	return (0);
}

/*------------------------------------------------------------------------
 * get connection number
 */
int xvt_w_get_xfd (const TERMWIN *tw)
{
	if (tw == 0)
		return (-1);

	return (tw->X->xfd);
}

/*------------------------------------------------------------------------
 * get window ID
 */
unsigned int xvt_w_get_windowid	(const TERMWIN *tw)
{
	if (tw == 0)
		return (0);

	return ((unsigned int)tw->X->win_parent);
}

/*------------------------------------------------------------------------
 * setup the display
 */
static TERMWIN * xvt_w_setup_display (XVT_DATA *xd)
{
	X_DATA *		xp;
	TERMWIN *		tw;
	const char *	display;

	/*--------------------------------------------------------------------
	 * Check if in a window
	 */
	display = xd->display;
	if (display == 0 || *display == 0)
	{
		sprintf(xd->error_buf, "Not in window environment");
		xd->error_num = XVT_ERR_NOTINWIN;
		return (0);
	}

	/*--------------------------------------------------------------------
	 * initialize x-info struct
	 */
	xp = (X_DATA *)xvt_mem_alloc(sizeof(*xp));
	if (xp == 0)
	{
		sprintf(xd->error_buf, "No memory");
		xd->error_num = XVT_ERR_NOMEM;
		return (0);
	}

	memset(xp,  0, sizeof(*xp));

	/*--------------------------------------------------------------------
	 * initialize win-info struct
	 */
	tw = (TERMWIN *)xvt_mem_alloc(sizeof(*tw));
	if (tw == 0)
	{
		xvt_mem_free(xp);
		sprintf(xd->error_buf, "No memory");
		xd->error_num = XVT_ERR_NOMEM;
		return (0);
	}

	memset(tw,  0, sizeof(*tw));
	tw->X	= xp;

	/*--------------------------------------------------------------------
	 * set defaults in struct
	 */
	tw->reverse		= FALSE;

	tw->border		= xd->border;
	tw->margin		= xd->margin;
	tw->extra		= (tw->border + tw->margin);

	tw->sb_enable	= xd->sb_enable;
	tw->sb_right	= xd->sb_right;
	tw->sb_shadow	= xd->sb_shadow;

	tw->cr_color	= -1;		/* this may be changed later */

	/*--------------------------------------------------------------------
	 * open the display
	 */
	tw->X->display = XOpenDisplay(display);
	if (tw->X->display == None)
	{
		xvt_w_free(tw);

		sprintf(xd->error_buf, "Cannot open display %s", display);
		xd->error_num = XVT_ERR_CANTOPENDISP;
		return (0);
	}

	tw->X->win_root	= DefaultRootWindow(tw->X->display);
	tw->X->screen	= DefaultScreen(tw->X->display);
	tw->X->xfd		= XConnectionNumber(tw->X->display);

	/*--------------------------------------------------------------------
	 * install error routine
	 */
	if (xd->xerrors)
		XSetErrorHandler(xvt_w_error_rtn);
	else
		XSetErrorHandler(xvt_w_error_rtn_dummy);

	/*--------------------------------------------------------------------
	 * get WM delete atom
	 */
	tw->X->wm_delete = XInternAtom(tw->X->display, "WM_DELETE_WINDOW", False);

	return (tw);
}

/*------------------------------------------------------------------------
 * close the display
 */
static void xvt_w_close_display (TERMWIN *tw)
{
	if (tw != 0)
	{
		if (tw->X->display != 0)
		{
			XCloseDisplay(tw->X->display);
			tw->X->display = 0;
		}

		xvt_w_free(tw);
	}
}

/*------------------------------------------------------------------------
 * close the cursor
 */
static void xvt_w_close_cursor (TERMWIN *tw)
{
	if (tw->X->cursor != 0)
	{
		XFreeCursor(tw->X->display, tw->X->cursor);
		tw->X->cursor = 0;
	}
}

/*------------------------------------------------------------------------
 * check if two fonts are the same size
 */
static int xvt_w_same_font (XFontStruct *f1, XFontStruct *f2)
{
	return (f1->max_bounds.width   == f2->max_bounds.width  &&
			f1->max_bounds.ascent  == f2->max_bounds.ascent &&
			f1->max_bounds.descent == f2->max_bounds.descent);
}

/*------------------------------------------------------------------------
 * load a font
 */
static int xvt_w_load_font (TERMWIN *tw, FONT_INFO *fi,
	const char *rname, const char *bname)
{
	/*--------------------------------------------------------------------
	 * check font names
	 */
	if (rname == 0 || *rname == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * cache names in font struct
	 */
	fi->reg_fname = rname;
	fi->bld_fname = bname;

	/*--------------------------------------------------------------------
	 * try to load the reg font
	 */
	fi->reg_font = XLoadQueryFont(tw->X->display, rname);
	if (fi->reg_font == 0)
	{
		fi->reg_fname = 0;
		fi->bld_fname = 0;
		return (-1);
	}

	/*--------------------------------------------------------------------
	 * cache reg font info
	 */
	fi->cell_xhome	= 0;
	fi->cell_yhome	= fi->reg_font->max_bounds.ascent;

	fi->cell_width	= fi->reg_font->max_bounds.width;

	fi->cell_height	= fi->reg_font->max_bounds.ascent +
					  fi->reg_font->max_bounds.descent;

	/*--------------------------------------------------------------------
	 * try to load the bold font
	 */
	if (bname != 0 && *bname != 0)
		fi->bld_font = XLoadQueryFont(tw->X->display, bname);
	else
		fi->bld_font = 0;

	/*--------------------------------------------------------------------
	 * if bold-font load succeeded, check if same size as reg
	 */
	if (fi->bld_font != 0)
	{
		if (! xvt_w_same_font(fi->reg_font, fi->bld_font))
		{
			XUnloadFont(tw->X->display, fi->bld_font->fid);
			fi->bld_font	= 0;
			fi->bld_fname	= 0;
		}
	}

	return (0);
}

/*------------------------------------------------------------------------
 * load all fonts to be used
 */
static int xvt_w_setup_fonts (TERMWIN *tw, XVT_DATA *xd)
{
	FONT_INFO *		fi	= tw->X->fonts;
	FONT_INFO *		fd;
	int				i;

	/*--------------------------------------------------------------------
	 * first try loading any user fonts
	 */
	for (i=0; i<XVT_NUM_FONTS; i++)
		xvt_w_load_font(tw, fi + i, xd->reg_fname[i], xd->bld_fname[i]);

	/*--------------------------------------------------------------------
	 * now check if default font got loaded OK
	 */
	fd = fi + xd->fontno;
	if (fd->reg_font == 0)
	{
		if (fd->reg_fname == 0)
		{
			sprintf(xd->error_buf, "No font specified for font %d",
				xd->fontno);
		}
		else
		{
			sprintf(xd->error_buf, "Cannot load font %d (%s)",
				xd->fontno, fd->reg_fname);
		}
		xd->error_num = XVT_ERR_CANTLOADFONTS;
		return (XVT_ERR_CANTLOADFONTS);
	}

	/*--------------------------------------------------------------------
	 * set initial font number
	 */
	tw->cur_font_no = xd->fontno;

	return (XVT_OK);
}

/*------------------------------------------------------------------------
 * free any font resources which were loaded
 */
static void xvt_w_close_fonts (TERMWIN *tw)
{
	int i;

	for (i=0; i<XVT_NUM_FONTS; i++)
	{
		FONT_INFO *	fi = tw->X->fonts + i;

		if (fi->reg_font != 0)
		{
			XUnloadFont(tw->X->display, fi->reg_font->fid);
			fi->reg_font = 0;
		}

		if (fi->bld_font != 0)
		{
			XUnloadFont(tw->X->display, fi->bld_font->fid);
			fi->bld_font = 0;
		}
	}
}

/*------------------------------------------------------------------------
 * setup mono colors
 */
static void xvt_w_cms_setup_mono (TERMWIN *tw, XVT_DATA *xd)
{
	Pixel	bp	= XBlackPixel(tw->X->display, tw->X->screen);
	Pixel	wp	= XWhitePixel(tw->X->display, tw->X->screen);

	/*--------------------------------------------------------------------
	 * if mono, all colors are either black or white
	 */
	COLOR_SET(tw, XVT_CLR_BLACK,     bp, FALSE);
	COLOR_SET(tw, XVT_CLR_RED,       bp, FALSE);
	COLOR_SET(tw, XVT_CLR_GREEN,     bp, FALSE);
	COLOR_SET(tw, XVT_CLR_YELLOW,    bp, FALSE);
	COLOR_SET(tw, XVT_CLR_BLUE,      bp, FALSE);
	COLOR_SET(tw, XVT_CLR_MAGENTA,   bp, FALSE);
	COLOR_SET(tw, XVT_CLR_CYAN,      bp, FALSE);
	COLOR_SET(tw, XVT_CLR_WHITE,     wp, FALSE);

	COLOR_SET(tw, XVT_CLR_GREY,      bp, FALSE);
	COLOR_SET(tw, XVT_CLR_LTRED,     bp, FALSE);
	COLOR_SET(tw, XVT_CLR_LTGREEN,   bp, FALSE);
	COLOR_SET(tw, XVT_CLR_LTYELLOW,  bp, FALSE);
	COLOR_SET(tw, XVT_CLR_LTBLUE,    bp, FALSE);
	COLOR_SET(tw, XVT_CLR_LTMAGENTA, bp, FALSE);
	COLOR_SET(tw, XVT_CLR_LTCYAN,    bp, FALSE);
	COLOR_SET(tw, XVT_CLR_LTWHITE,   wp, FALSE);

	COLOR_SET(tw, XVT_CLR_FG,        bp, FALSE);
	COLOR_SET(tw, XVT_CLR_BG,        wp, FALSE);
	COLOR_SET(tw, XVT_CLR_PR,        bp, FALSE);
	COLOR_SET(tw, XVT_CLR_BD,        bp, FALSE);
	COLOR_SET(tw, XVT_CLR_CR,        None, FALSE);
	COLOR_SET(tw, XVT_CLR_SB,        bp, FALSE);
	COLOR_SET(tw, XVT_CLR_ST,        bp, FALSE);
	COLOR_SET(tw, XVT_CLR_SB_TS,     wp, FALSE);
	COLOR_SET(tw, XVT_CLR_SB_BS,     wp, FALSE);

	/*--------------------------------------------------------------------
	 * If reverse requested, switch FG & BG colors
	 */
	if (xd->reverse)
		xvt_w_set_rev_video(tw, TRUE);

	/*--------------------------------------------------------------------
	 * setup extended TERM name & definitions
	 */
	strcpy(xd->term_type, XVT_TERM_MONO);
	xd->term_defs = xvt_termdefs_m;
}

/*------------------------------------------------------------------------
 * parse a color
 */
static void xvt_w_parse_color (TERMWIN *tw, const char *color, XColor *xc,
	int n)
{
	char			color_buf[128];
	char *			b	= color_buf;
	const char *	s;

	/*--------------------------------------------------------------------
	 * remove spaces, "-", & "_" from name
	 */
	for (s=color; *s; s++)
	{
		if (isspace(*s) || *s == '-' || *s == '_')
			continue;
		*b++ = *s;
	}
	*b = 0;

	/*--------------------------------------------------------------------
	 * try to parse the color & use default if can't
	 */
	if (XParseColor(tw->X->display, tw->X->cmap, color_buf, xc) == 0)
		memcpy(xc, &xvt_w_colors[n].clr_xcolor, sizeof(*xc));
}

/*------------------------------------------------------------------------
 * allocate a color
 */
static void xvt_w_set_clr (TERMWIN *tw, const char *color, int n,
	Pixel pix_dflt)
{
	if (color == 0 || *color == 0 || xvt_strccmp(color, "none") == 0)
	{
		COLOR_SET(tw, n, pix_dflt, FALSE);
	}
	else
	{
		XColor	xcolor;

		xvt_w_parse_color(tw, color, &xcolor, n);
		if (XAllocColor(tw->X->display, tw->X->cmap, &xcolor) == 0)
			COLOR_SET(tw, n, pix_dflt, FALSE);
		else
			COLOR_SET(tw, n, xcolor.pixel, TRUE);
	}
}

/*------------------------------------------------------------------------
 * calculate shadow colors
 *
 * These calculations were lifted from the rxvt source
 * (which took them from the fvwm source).
 */
static Pixel xvt_w_get_top_shadow (TERMWIN *tw, int color)
{
	XColor	xcolor;

	/*--------------------------------------------------------------------
	 * Bottom shadow:	1/2 of the color
	 */
	xcolor.pixel = COLOR_PIX(tw, color);
	XQueryColor(tw->X->display, tw->X->cmap, &xcolor);

	xcolor.red		= (xcolor.red	/ 2);
	xcolor.green	= (xcolor.green	/ 2);
	xcolor.blue		= (xcolor.blue	/ 2);

	if (XAllocColor(tw->X->display, tw->X->cmap, &xcolor) == 0)
		xcolor.pixel = COLOR_PIX(tw, color);

	return (xcolor.pixel);
}

static Pixel xvt_w_get_bot_shadow (TERMWIN *tw, int color)
{
	XColor	xcolor;
	XColor	xwhite;
	unsigned short	r, g, b;

	/*--------------------------------------------------------------------
	 * Top shadow: magic calculation
	 */
	xcolor.pixel = COLOR_PIX(tw, color);
	XQueryColor(tw->X->display, tw->X->cmap, &xcolor);

	xwhite.pixel = COLOR_PIX(tw, XVT_CLR_LTWHITE);
	XQueryColor(tw->X->display, tw->X->cmap, &xwhite);

	r = (xwhite.red   / 5);
	g = (xwhite.green / 5);
	b = (xwhite.blue  / 5);
	if (xcolor.red   < r)	xcolor.red   = r;
	if (xcolor.green < g)	xcolor.green = g;
	if (xcolor.blue  < b)	xcolor.blue  = b;

	xcolor.red   = ((xcolor.red   * 7) / 5);
	xcolor.green = ((xcolor.green * 7) / 5);
	xcolor.blue  = ((xcolor.blue  * 7) / 5);
	r = (xwhite.red);
	g = (xwhite.green);
	b = (xwhite.blue);
	if (xcolor.red   > r)	xcolor.red   = r;
	if (xcolor.green > g)	xcolor.green = g;
	if (xcolor.blue  > b)	xcolor.blue  = b;

	if (XAllocColor(tw->X->display, tw->X->cmap, &xcolor) == 0)
		xcolor.pixel = COLOR_PIX(tw, color);

	return (xcolor.pixel);
}

/*------------------------------------------------------------------------
 * setup scrollbar shadow colors
 */
static void xvt_w_set_sb_shadows (TERMWIN *tw)
{
	COLOR_SET(tw, XVT_CLR_SB_TS, xvt_w_get_top_shadow(tw, XVT_CLR_SB), TRUE);
	COLOR_SET(tw, XVT_CLR_SB_BS, xvt_w_get_bot_shadow(tw, XVT_CLR_SB), TRUE);
}

/*------------------------------------------------------------------------
 * set reverse video
 */
void xvt_w_set_rev_video (TERMWIN *tw, int reverse)
{
	if (reverse != tw->reverse)
	{
		Pixel	p;
		int		g;

		p							    = tw->X->clr_tbl[XVT_CLR_FG].pix;
		tw->X->clr_tbl[XVT_CLR_FG].pix	= tw->X->clr_tbl[XVT_CLR_BG].pix;
		tw->X->clr_tbl[XVT_CLR_BG].pix	= p;

		g							    = tw->X->clr_tbl[XVT_CLR_FG].got;
		tw->X->clr_tbl[XVT_CLR_FG].got	= tw->X->clr_tbl[XVT_CLR_BG].got;
		tw->X->clr_tbl[XVT_CLR_BG].got	= g;

		tw->reverse = reverse;
	}

}

/*------------------------------------------------------------------------
 * setup color colors
 */
static void xvt_w_cms_setup_color (TERMWIN *tw, XVT_DATA *xd)
{
	Pixel	bp	= XBlackPixel(tw->X->display, tw->X->screen);
	Pixel	wp	= XWhitePixel(tw->X->display, tw->X->screen);

	/*--------------------------------------------------------------------
	 * Setup all colors from user-spec or default
	 */
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_BLACK    ], XVT_CLR_BLACK,     bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_RED      ], XVT_CLR_RED,       bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_GREEN    ], XVT_CLR_GREEN,     bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_YELLOW   ], XVT_CLR_YELLOW,    bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_BLUE     ], XVT_CLR_BLUE,      bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_MAGENTA  ], XVT_CLR_MAGENTA,   bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_CYAN     ], XVT_CLR_CYAN,      bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_WHITE    ], XVT_CLR_WHITE,     wp);

	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_GREY     ], XVT_CLR_GREY,      bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_LTRED    ], XVT_CLR_LTRED,     bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_LTGREEN  ], XVT_CLR_LTGREEN,   bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_LTYELLOW ], XVT_CLR_LTYELLOW,  bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_LTBLUE   ], XVT_CLR_LTBLUE,    bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_LTMAGENTA], XVT_CLR_LTMAGENTA, bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_LTCYAN   ], XVT_CLR_LTCYAN,    bp);
	xvt_w_set_clr(tw, xd->scr_color[XVT_CLR_LTWHITE  ], XVT_CLR_LTWHITE,   wp);

	xvt_w_set_clr(tw, xd->fg_color,                     XVT_CLR_FG,        bp);
	xvt_w_set_clr(tw, xd->bg_color,                     XVT_CLR_BG,        wp);
	xvt_w_set_clr(tw, xd->pr_color,                     XVT_CLR_PR,        bp);
	xvt_w_set_clr(tw, xd->bd_color,                     XVT_CLR_BD,        bp);
	xvt_w_set_clr(tw, xd->cr_color,                     XVT_CLR_CR,        bp);
	xvt_w_set_clr(tw, xd->sb_color,                     XVT_CLR_SB,        bp);
	xvt_w_set_clr(tw, xd->st_color,                     XVT_CLR_ST,        wp);

	/*--------------------------------------------------------------------
	 * Setup scrollbar shadow colors
	 */
	xvt_w_set_sb_shadows(tw);

	/*--------------------------------------------------------------------
	 * If a text cursor color was specified, cache it.
	 *
	 * Remember, that if no text cursor color is specified, then
	 * the text cursor is not a color but the inverse of the color
	 * in that cell.
	 */
	if (xd->cr_color != 0 && *xd->cr_color != 0 &&
	    xvt_strccmp(xd->cr_color, "none") != 0)
	{
		tw->cr_color = XVT_CLR_CR;
	}

	/*--------------------------------------------------------------------
	 * If reverse requested, switch FG & BG colors
	 */
	if (xd->reverse)
		xvt_w_set_rev_video(tw, TRUE);

	/*--------------------------------------------------------------------
	 * setup extended TERM name & definitions
	 */
	strcpy(xd->term_type, XVT_TERM_COLOR);
	xd->term_defs = xvt_termdefs_c;
}

/*------------------------------------------------------------------------
 * setup initial colors
 */
static void xvt_w_setup_colors (TERMWIN *tw, XVT_DATA *xd)
{
	/*--------------------------------------------------------------------
	 * Check if in color or mono
	 */
	tw->can_be_color = (DisplayPlanes(tw->X->display, tw->X->screen) > 1);

	/*--------------------------------------------------------------------
	 * allow user-override to mono
	 */
	if (xd->monodisp)
		tw->can_be_color = FALSE;

	/*--------------------------------------------------------------------
	 * now setup colors
	 */
	tw->X->cmap = DefaultColormap(tw->X->display, tw->X->screen);

	if (tw->can_be_color)
		xvt_w_cms_setup_color(tw, xd);
	else
		xvt_w_cms_setup_mono (tw, xd);

	/*--------------------------------------------------------------------
	 * we always startup in mono mode
	 */
	xvt_w_set_scrn_mode(tw, SCRN_MODE_MONO);
}

/*------------------------------------------------------------------------
 * free any color resources created
 */
static void xvt_w_close_colors (TERMWIN *tw)
{
	if (tw->can_be_color)
	{
#if 0
		/*----------------------------------------------------------------
		 * For some reason, when we explicitly free the colors we allocated,
		 * we get an X_FreeColors protocol error from XCloseDisplay().
		 */
		Pixel	pixels[CMS_COLORSIZE];
		int		i;
		int		n;

		for (n=0, i=0; i<CMS_COLORSIZE; i++)
		{
			if (tw->X->clr_tbl[i].got)
				pixels[n++] = tw->X->clr_tbl[i].got;
		}

		if (n > 0)
			XFreeColors(tw->X->display, tw->X->cmap, pixels, n, 0);
#endif
	}
}

/*------------------------------------------------------------------------
 * adjust the size struct
 */
static void xvt_w_setup_size (TERMWIN *tw)
{
	/*--------------------------------------------------------------------
	 * calculate scrollbar size
	 */
	if (tw->sb_enable)
	{
		tw->sb_size	= XVT_SB_WIDTH;
		if (tw->sb_shadow)
			tw->sb_size += (2 * XVT_SB_SHADOW);
	}
	else
	{
		tw->sb_size	= 0;
	}

	/*--------------------------------------------------------------------
	 * parent window stuff
	 */
	tw->X->s_hints.min_width		= COL_TO_W(tw, XVT_MIN_COLS) +
										(2 * tw->extra) + tw->sb_size;
	tw->X->s_hints.min_height		= ROW_TO_H(tw, XVT_MIN_ROWS) +
										(2 * tw->extra);

	tw->X->s_hints.width_inc		= COL_TO_W(tw, 1);
	tw->X->s_hints.height_inc		= ROW_TO_H(tw, 1);

	tw->X->s_hints.width			= COL_TO_W(tw, tw->win_cols) +
										(2 * tw->extra) + tw->sb_size;
	tw->X->s_hints.height			= ROW_TO_H(tw, tw->win_rows) +
										(2 * tw->extra);

	/*--------------------------------------------------------------------
	 * vt window stuff
	 */
	if (tw->sb_enable)
	{
		tw->X->vt_x = (tw->sb_right ? 0 : tw->sb_size);
		tw->X->vt_y = 0;
	}
	else
	{
		tw->X->vt_x = 0;
		tw->X->vt_y = 0;
	}
	tw->X->vt_width		= tw->X->s_hints.width - tw->sb_size;
	tw->X->vt_height	= tw->X->s_hints.height;

	/*--------------------------------------------------------------------
	 * sb window stuff
	 */
	if (tw->sb_enable)
	{
		tw->X->sb_x			= (tw->sb_right ? tw->X->vt_width : 0);
		tw->X->sb_y			= 0;
		tw->X->sb_width		= tw->sb_size;
		tw->X->sb_height	= tw->X->s_hints.height;
	}
	else
	{
		tw->X->sb_x			= 0;
		tw->X->sb_y			= 0;
		tw->X->sb_width		= 0;
		tw->X->sb_height	= 0;
	}
}

/*------------------------------------------------------------------------
 * setup the windows
 */
static int xvt_w_setup_windows (TERMWIN *tw, XVT_DATA *xd)
{
	unsigned int	mask;
	unsigned int	flags	= PMinSize | PResizeInc | PWinGravity;
	int				x_pos	= 0;
	int				y_pos	= 0;
	unsigned int	gravity	= NorthWestGravity;

	/*--------------------------------------------------------------------
	 * check # rows & cols
	 */
	if (xd->cur_cols < XVT_MIN_COLS)
		xd->cur_cols = XVT_MIN_COLS;

	if (xd->cur_rows < XVT_MIN_ROWS)
		xd->cur_rows = XVT_MIN_ROWS;

	/*--------------------------------------------------------------------
	 * check if geometry specified
	 */
	if (xd->geometry != 0 && *xd->geometry != 0)
	{
		int				x;
		int				y;
		unsigned int	width;
		unsigned int	height;
		int				rc;

		rc = XParseGeometry(xd->geometry, &x, &y, &width, &height);

		/*----------------------------------------------------------------
		 * check if width specified
		 */
		if (rc & WidthValue)
		{
			flags |= USSize;
			xd->cur_cols = width;

			if (xd->cur_cols < XVT_MIN_COLS)
				xd->cur_cols = XVT_MIN_COLS;
		}

		/*----------------------------------------------------------------
		 * check if height specified
		 */
		if (rc & HeightValue)
		{
			flags |= USSize;
			xd->cur_rows = height;

			if (xd->cur_rows < XVT_MIN_ROWS)
				xd->cur_rows = XVT_MIN_ROWS;
		}

		/*----------------------------------------------------------------
		 * check if x position specified
		 */
		if (rc & XValue)
		{
			flags |= USPosition;
			x_pos  = x;

			if (rc & XNegative)
			{
				gravity = NorthEastGravity;

				x_pos += DisplayWidth(tw->X->display, tw->X->screen)
					- (COL_TO_W(tw, xd->cur_cols) + (2 * tw->extra));
			}
		}

		/*----------------------------------------------------------------
		 * check if y position specified
		 */
		if (rc & YValue)
		{
			flags |= USPosition;
			y_pos  = y;

			if (rc & YNegative)
			{
				gravity = (gravity == NorthEastGravity ?
					SouthEastGravity : SouthWestGravity);

				y_pos += DisplayHeight(tw->X->display, tw->X->screen)
					- (ROW_TO_H(tw, xd->cur_rows) + (2 * tw->extra));
			}
		}
	}

	tw->win_cols	= xd->cur_cols;
	tw->win_rows	= xd->cur_rows;

	/*--------------------------------------------------------------------
	 * setup size hints
	 */
	tw->X->s_hints.flags			= flags;
	tw->X->s_hints.x				= x_pos;
	tw->X->s_hints.y				= y_pos;
	tw->X->s_hints.win_gravity		= gravity;

	xvt_w_setup_size(tw);

	/*--------------------------------------------------------------------
	 * setup window hints
	 */
	tw->X->w_hints.flags			= InputHint | StateHint;
	tw->X->w_hints.input			= True;
	tw->X->w_hints.initial_state	= xd->iconic ? IconicState : NormalState;

	tw->iconified					= xd->iconic;

	/*--------------------------------------------------------------------
	 * load icon file if specified
	 */
	if (xd->ico_file != 0 && *xd->ico_file != 0)
	{
		char	icon_path[256];

		xvt_path_resolve(icon_path, xd->ico_file);
		if (*icon_path != 0)
		{
			unsigned int	w, h;
			int				hx, hy;
			int				rc;

			rc = XReadBitmapFile(tw->X->display, tw->X->win_root,
				icon_path, &w, &h, &tw->X->icon, &hx, &hy);
			if (rc == BitmapSuccess)
			{
				tw->X->w_hints.flags		|= IconPixmapHint;
				tw->X->w_hints.icon_pixmap	 = tw->X->icon;
			}
		}
	}

	/*--------------------------------------------------------------------
	 * check if any icon-geometry was specified
	 *
	 * Note that we assume an icon size of 64x64
	 */
	if (xd->icongeom != 0 && *xd->icongeom != 0)
	{
		int				x;
		int				y;
		unsigned int	width;
		unsigned int	height;
		int				rc;

		rc = XParseGeometry(xd->icongeom, &x, &y, &width, &height);

		/*----------------------------------------------------------------
		 * check if x position specified
		 */
		if (rc & XValue)
		{
			tw->X->w_hints.flags |= IconPositionHint;
			tw->X->w_hints.icon_x = x;

			if (rc & XNegative)
			{
				tw->X->w_hints.icon_x =
					DisplayWidth(tw->X->display, tw->X->screen) - 64;
			}
		}

		/*----------------------------------------------------------------
		 * check if y position specified
		 */
		if (rc & YValue)
		{
			tw->X->w_hints.flags |= IconPositionHint;
			tw->X->w_hints.icon_y = y;

			if (rc & YNegative)
			{
				tw->X->w_hints.icon_y =
					DisplayHeight(tw->X->display, tw->X->screen) - 64;
			}
		}
	}

	/*--------------------------------------------------------------------
	 * setup class hints
	 */
	{
		const char *	resname;

		if (xd->resname != 0 && *xd->resname != 0)
			resname = xd->resname;
		else if (xd->appname != 0 && *xd->appname != 0)
			resname = xd->appname;
		else
			resname = XVT_NAME;

		tw->X->c_hints.res_name		= (char *)resname;
		tw->X->c_hints.res_class	= XVT_APPCLASS;
	}

	/*--------------------------------------------------------------------
	 * create parent window
	 */
	tw->X->win_parent = XCreateSimpleWindow(tw->X->display, tw->X->win_root,
		tw->X->s_hints.x, tw->X->s_hints.y,
		tw->X->s_hints.width, tw->X->s_hints.height, 0,
		COLOR_PIX(tw, XVT_CLR_BD), COLOR_PIX(tw, XVT_CLR_BG));

	if (tw->X->win_parent == None)
	{
		sprintf(xd->error_buf, "Cannot create parent window");
		xd->error_num = XVT_ERR_CANTMAKEWIN;
		return (XVT_ERR_CANTMAKEWIN);
	}

	/*--------------------------------------------------------------------
	 * set window properties
	 */
	XSetWMProperties(tw->X->display, tw->X->win_parent,
		0, 0, 0, 0,
		&tw->X->s_hints, &tw->X->w_hints, &tw->X->c_hints);

	XSetWMProtocols(tw->X->display, tw->X->win_parent, &tw->X->wm_delete, 1);

	/*--------------------------------------------------------------------
	 * create cursor to use
	 */
	{
		int	shape = -1;

		if (xd->pointer != 0)
			shape = xvt_w_pointer_value(xd->pointer);

		if (shape != -1)
			xvt_w_set_cursor(tw, shape);
	}

	/*--------------------------------------------------------------------
	 * now create vt window
	 */
	tw->X->win_vt = XCreateSimpleWindow(tw->X->display, tw->X->win_parent,
		tw->X->vt_x, tw->X->vt_y, 
		tw->X->vt_width, tw->X->vt_height, 0,
		COLOR_PIX(tw, XVT_CLR_BD), COLOR_PIX(tw, XVT_CLR_BG));

	if (tw->X->win_vt == None)
	{
		sprintf(xd->error_buf, "Cannot create vt window");
		xd->error_num = XVT_ERR_CANTMAKEWIN;
		return (XVT_ERR_CANTMAKEWIN);
	}

	/*--------------------------------------------------------------------
	 * now create sb window if requested
	 */
	if (tw->sb_enable)
	{
		tw->X->win_sb = XCreateSimpleWindow(tw->X->display, tw->X->win_parent,
			tw->X->sb_x, tw->X->sb_y, 
			tw->X->sb_width, tw->X->sb_height, 0,
			COLOR_PIX(tw, XVT_CLR_SB), COLOR_PIX(tw, XVT_CLR_ST));

		if (tw->X->win_sb == None)
		{
			sprintf(xd->error_buf, "Cannot create scrollbar window");
			xd->error_num = XVT_ERR_CANTMAKEWIN;
			return (XVT_ERR_CANTMAKEWIN);
		}
	}

	/*--------------------------------------------------------------------
	 * set selection masks
	 */
	mask = 0
			| FocusChangeMask
			| StructureNotifyMask
			;

	XSelectInput(tw->X->display, tw->X->win_parent, mask);

	mask = 0
			| KeyPressMask
			| KeyReleaseMask
			| KeymapStateMask
			| ExposureMask
			| FocusChangeMask
			;

	if (! xd->ignptr)
	{
		mask |= 0
				| ButtonPressMask
				| ButtonReleaseMask
				| PointerMotionMask
				;
	}

	XSelectInput(tw->X->display, tw->X->win_vt, mask);

	if (tw->sb_enable)
	{
		mask = 0
				| ButtonPressMask
				| ButtonReleaseMask
				| PointerMotionMask
				;

		XSelectInput(tw->X->display, tw->X->win_sb, mask);
	}

	/*--------------------------------------------------------------------
	 * set window to be displayed
	 */
	if (tw->sb_enable)
		XMapWindow(tw->X->display, tw->X->win_sb);
	XMapWindow(tw->X->display, tw->X->win_vt);
	XMapWindow(tw->X->display, tw->X->win_parent);

	return (0);
}

/*------------------------------------------------------------------------
 * resize the windows
 */
static void xvt_w_resize_win (TERMWIN *tw)
{
	xvt_w_setup_size(tw);

	XSetWMNormalHints(tw->X->display, tw->X->win_parent,
		&tw->X->s_hints);
#if 1
	XResizeWindow(tw->X->display, tw->X->win_parent,
		tw->X->s_hints.width, tw->X->s_hints.height);
#endif

	XMoveResizeWindow(tw->X->display, tw->X->win_vt,
		tw->X->vt_x, tw->X->vt_y,
		tw->X->vt_width, tw->X->vt_height);

	if (tw->sb_enable)
	{
		XMoveResizeWindow(tw->X->display, tw->X->win_sb,
			tw->X->sb_x, tw->X->sb_y,
			tw->X->sb_width, tw->X->sb_height);
	}

	XSync(tw->X->display, FALSE);
}

/*------------------------------------------------------------------------
 * user call to resize the windows
 */
void xvt_w_resize (TERMWIN *tw, int cols, int rows)
{
	if (cols >= XVT_MIN_COLS)
		tw->win_cols = cols;

	if (rows >= XVT_MIN_ROWS)
		tw->win_rows = rows;

	xvt_w_resize_win(tw);
}

/*------------------------------------------------------------------------
 * raise the window
 */
void xvt_w_raise (TERMWIN *tw)
{
	XRaiseWindow(tw->X->display, tw->X->win_parent);
}

/*------------------------------------------------------------------------
 * lower the window
 */
void xvt_w_lower (TERMWIN *tw)
{
	XLowerWindow(tw->X->display, tw->X->win_parent);
}

/*------------------------------------------------------------------------
 * get window size
 */
void xvt_w_get_size (TERMWIN *tw, int *w, int *h)
{
	XWindowAttributes	attrs;

	XGetWindowAttributes(tw->X->display, tw->X->win_parent, &attrs);
	*w = attrs.width;
	*h = attrs.height;
}

/*------------------------------------------------------------------------
 * set window size
 */
void xvt_w_set_size (TERMWIN *tw, int w, int h)
{
	int cols = W_TO_COL(tw, w);
	int rows = H_TO_ROW(tw, h);

	xvt_w_resize(tw, cols, rows);
}

/*------------------------------------------------------------------------
 * get window position
 */
void xvt_w_get_position (TERMWIN *tw, int *x, int *y)
{
	XWindowAttributes	attrs;
	Window				dummy;

	XGetWindowAttributes(tw->X->display, tw->X->win_parent, &attrs);
	XTranslateCoordinates(tw->X->display, tw->X->win_parent, tw->X->win_root,
		-attrs.border_width, -attrs.border_width, x, y, &dummy);
}

/*------------------------------------------------------------------------
 * set window position
 */
void xvt_w_set_position (TERMWIN *tw, int x, int y)
{
	XMoveWindow(tw->X->display, tw->X->win_parent, x, y);
}

/*------------------------------------------------------------------------
 * close the windows
 */
static void xvt_w_close_windows (TERMWIN *tw)
{
	if (tw->X->icon != None)
	{
		XFreePixmap(tw->X->display, tw->X->icon);
		tw->X->icon = None;
	}
}

/*------------------------------------------------------------------------
 * setup the GC
 */
static void xvt_w_setup_gc (TERMWIN *tw)
{
	XGCValues	gc_val;
	int			valuemask;
	char		fill_stipple_bits[1];

	/*--------------------------------------------------------------------
	 * create the stipple to use
	 */
	fill_stipple_bits[0] = 0;
	tw->X->fill_stipple = XCreateBitmapFromData(tw->X->display, tw->X->win_vt,
		fill_stipple_bits, 8, 1);

	/*--------------------------------------------------------------------
	 * create the vt GC
	 */
	memset(&gc_val, 0, sizeof(gc_val));

	gc_val.function				= GXcopy;
	gc_val.fill_style			= FillOpaqueStippled;
	gc_val.stipple				= tw->X->fill_stipple;
	gc_val.graphics_exposures	= FALSE;
	gc_val.foreground			= COLOR_PIX(tw, XVT_CLR_FG);
	gc_val.background			= COLOR_PIX(tw, XVT_CLR_BG);

	valuemask					= 0
									| GCFunction
									| GCFillStyle
									| GCStipple
									| GCForeground
									| GCBackground
									| GCGraphicsExposures
									;

	tw->X->gc_vt = XCreateGC(tw->X->display, tw->X->win_vt, valuemask,
		&gc_val);

	/*--------------------------------------------------------------------
	 * create the sb GCs
	 */
	if (tw->sb_enable)
	{
		memset(&gc_val, 0, sizeof(gc_val));

		gc_val.foreground = COLOR_PIX(tw, XVT_CLR_SB);
		tw->X->gc_sb    = XCreateGC(tw->X->display, tw->X->win_sb,
			GCForeground, &gc_val);

		gc_val.foreground = COLOR_PIX(tw, XVT_CLR_SB_TS);
		tw->X->gc_sb_ts = XCreateGC(tw->X->display, tw->X->win_sb,
			GCForeground, &gc_val);

		gc_val.foreground = COLOR_PIX(tw, XVT_CLR_SB_BS);
		tw->X->gc_sb_bs = XCreateGC(tw->X->display, tw->X->win_sb,
			GCForeground, &gc_val);

		if (tw->sb_shadow)
		{
			XSetWindowBackground(tw->X->display, tw->X->win_sb,
				COLOR_PIX(tw, XVT_CLR_ST));
			XClearWindow(tw->X->display, tw->X->win_sb);
		}
	}
}

/*------------------------------------------------------------------------
 * free the GC
 */
static void xvt_w_close_gc (TERMWIN *tw)
{
	if (tw->X->gc_vt != None)
	{
		XFreeGC(tw->X->display, tw->X->gc_vt);
		tw->X->gc_vt = None;
	}

	if (tw->X->gc_sb != None)
	{
		XFreeGC(tw->X->display, tw->X->gc_sb);
		tw->X->gc_sb = None;
	}

	if (tw->X->gc_sb_ts != None)
	{
		XFreeGC(tw->X->display, tw->X->gc_sb_ts);
		tw->X->gc_sb_ts = None;
	}

	if (tw->X->gc_sb_bs != None)
	{
		XFreeGC(tw->X->display, tw->X->gc_sb_bs);
		tw->X->gc_sb_bs = None;
	}

	if (tw->X->fill_stipple != None)
	{
		XFreePixmap(tw->X->display, tw->X->fill_stipple);
		tw->X->fill_stipple = None;
	}
}

/*------------------------------------------------------------------------
 * change font number
 *
 * direction: <0 DOWN, >0 UP
 */
void xvt_w_chg_font_no (TERMWIN *tw, int direction)
{
	FONT_INFO *	fi;
	int			i;

	if (direction == 0)
	{
		return;
	}
	else if (direction < 0)
	{
		for (i=tw->cur_font_no-1; i>=0; i--)
		{
			fi = tw->X->fonts + i;
			if (fi->reg_font != 0)
				break;
		}
	}
	else /* (direction > 0) */
	{
		for (i=tw->cur_font_no+1; i<XVT_NUM_FONTS; i++)
		{
			fi = tw->X->fonts + i;
			if (fi->reg_font != 0)
				break;
		}
	}

	if (i >= 0 && i < XVT_NUM_FONTS)
		xvt_w_set_font_no(tw, i);
}

/*------------------------------------------------------------------------
 * setup current font number
 */
void xvt_w_set_font_no (TERMWIN *tw, int font_no)
{
	if (font_no != tw->cur_font_no)
	{
		if (font_no >= 0 && font_no < XVT_NUM_FONTS)
		{
			FONT_INFO *	fi = tw->X->fonts + font_no;

			if (fi->reg_font != 0)
			{
				int font_type = tw->cur_font_type;

				tw->cur_font_no = font_no;
				tw->cur_font_type = FONT_NONE;
				xvt_w_set_font_type(tw, font_type);
				xvt_w_resize_win(tw);
			}
		}
	}
}

/*------------------------------------------------------------------------
 * setup current font type
 */
void xvt_w_set_font_type (TERMWIN *tw, int font_type)
{
	if (font_type != tw->cur_font_type)
	{
		FONT_INFO *	fi = tw->X->fonts + tw->cur_font_no;
		Font		fid;

		switch (font_type)
		{
		case FONT_BOLD:
			if (fi->bld_font != 0)
			{
				fid = fi->bld_font->fid;
				break;
			}
			/*FALLTHROUGH*/

		case FONT_NORM:
			fid = fi->reg_font->fid;
			break;

		default:
			return;
		}

		XSetFont(tw->X->display, tw->X->gc_vt, fid);

		tw->cur_font_type = font_type;
	}
}

/*------------------------------------------------------------------------
 * setup current screen display as color or mono
 */
int xvt_w_set_scrn_mode (TERMWIN *tw, int mode)
{
	if (mode == SCRN_MODE_COLOR)
		mode = (tw->can_be_color ? SCRN_MODE_COLOR : SCRN_MODE_MONO);

	tw->cur_fg = XVT_CLR_FG;
	tw->cur_bg = XVT_CLR_BG;

	return (mode);
}

/*------------------------------------------------------------------------
 * specify X cursor
 */
void xvt_w_set_cursor (TERMWIN *tw, int shape)
{
	if (shape >= 0 && shape < XC_num_glyphs)
	{
		Cursor cursor;

		shape &= ~1;
		cursor = XCreateFontCursor(tw->X->display, shape);
		if (cursor != 0)
		{
			if (tw->can_be_color)
			{
				XColor	cursor_fg;
				XColor	cursor_bg;

				cursor_fg.pixel = COLOR_PIX(tw, XVT_CLR_PR);
				cursor_bg.pixel = COLOR_PIX(tw, XVT_CLR_BG);

				XQueryColor(tw->X->display, tw->X->cmap, &cursor_fg);
				XQueryColor(tw->X->display, tw->X->cmap, &cursor_bg);

				XRecolorCursor(tw->X->display, cursor, &cursor_fg, &cursor_bg);
			}

			xvt_w_close_cursor(tw);
			tw->X->cursor = cursor;
			XDefineCursor(tw->X->display, tw->X->win_parent, tw->X->cursor);
		}
	}
}

/*------------------------------------------------------------------------
 * setup everything
 */
TERMWIN * xvt_w_setup_all (XVT_DATA *xd)
{
	TERMWIN *	tw;
	int			rc;

	/*--------------------------------------------------------------------
	 * open the display
	 */
	tw = xvt_w_setup_display(xd);
	if (tw == 0)
	{
		return (0);
	}

	/*--------------------------------------------------------------------
	 * setup fonts to use
	 */
	rc = xvt_w_setup_fonts(tw, xd);
	if (rc)
	{
		xvt_w_close_all(tw);
		return (0);
	}

	/*--------------------------------------------------------------------
	 * setup colors
	 */
	xvt_w_setup_colors(tw, xd);

	/*--------------------------------------------------------------------
	 * create the window
	 */
	rc = xvt_w_setup_windows(tw, xd);
	if (rc)
	{
		xvt_w_close_all(tw);
		return (0);
	}

	/*--------------------------------------------------------------------
	 * create Graphics Context
	 */
	xvt_w_setup_gc(tw);

	return (tw);
}

/*------------------------------------------------------------------------
 * free our TERMWIN struct
 */
void xvt_w_free (TERMWIN *tw)
{
	if (tw != 0)
	{
		xvt_mem_free(tw->X);
		xvt_mem_free(tw);
	}
}

/*------------------------------------------------------------------------
 * close out everything
 */
void xvt_w_close_all (TERMWIN *tw)
{
	if (tw != 0)
	{
		xvt_w_close_cursor(tw);
		xvt_w_close_windows(tw);
		xvt_w_close_gc(tw);
		xvt_w_close_colors(tw);
		xvt_w_close_fonts(tw);
		xvt_w_close_display(tw);
	}
}

/*------------------------------------------------------------------------
 * close out connection
 */
void xvt_w_close_con (TERMWIN *tw)
{
	if (tw != 0)
	{
		close(tw->X->xfd);
	}
}

/*------------------------------------------------------------------------
 * ring the bell
 */
void xvt_w_bell (TERMWIN *tw)
{
	XBell(tw->X->display, 0);
}

/*------------------------------------------------------------------------
 * iconify the window
 */
void xvt_w_set_icon_state (TERMWIN *tw, int state)
{
	if (state)
	{
		XIconifyWindow(tw->X->display, tw->X->win_parent, tw->X->screen);
	}
	else
	{
		XMapWindow(tw->X->display, tw->X->win_parent);
	}

	tw->iconified = state;
}

/*------------------------------------------------------------------------
 * set window title
 */
void xvt_w_set_win_title (TERMWIN *tw, const char *str)
{
	char *	title;

	if (XFetchName(tw->X->display, tw->X->win_parent, &title))
		title = 0;

	if (title == 0 || strcmp(title, str) != 0)
		XStoreName(tw->X->display, tw->X->win_parent, str);

	if (title != 0)
		XFree(title);
}

/*------------------------------------------------------------------------
 * set icon title
 */
void xvt_w_set_ico_title (TERMWIN *tw, const char *str)
{
	char *	title;

	if (XGetIconName(tw->X->display, tw->X->win_parent, &title))
		title = 0;

	if (title == 0 || strcmp(title, str) != 0)
		XSetIconName(tw->X->display, tw->X->win_parent, str);

	if (title != 0)
		XFree(title);
}

/*------------------------------------------------------------------------
 * set the icon for a window
 */
void xvt_w_set_icon (TERMWIN *tw, int width, int height, unsigned char *bits)
{
	/*--------------------------------------------------------------------
	 * sanity checks
	 */
	if (width <= 0 || height <= 0 || bits == 0)
		return;

	/*--------------------------------------------------------------------
	 * free any existing icon pixmap
	 */
	if (tw->X->icon != 0)
	{
		XFreePixmap(tw->X->display, tw->X->icon);

		tw->X->w_hints.flags		&= ~IconPixmapHint;
		tw->X->w_hints.icon_pixmap	 = 0;
	}

	/*--------------------------------------------------------------------
	 * create new icon pixmap
	 */
	tw->X->icon = XCreateBitmapFromData(tw->X->display, tw->X->win_root,
		(char *)bits, width, height);
	if (tw->X->icon != 0)
	{
		tw->X->w_hints.flags		|= IconPixmapHint;
		tw->X->w_hints.icon_pixmap	 = tw->X->icon;
	}

	/*--------------------------------------------------------------------
	 * set icon into window properties
	 */
	XSetWMProperties(tw->X->display, tw->X->win_parent,
		0, 0, 0, 0, &tw->X->s_hints, &tw->X->w_hints, 0);
}

/*------------------------------------------------------------------------
 * set foreground color
 */
void xvt_w_set_fg	(TERMWIN *tw, int fg, int force)
{
	if (force || fg != tw->cur_fg)
	{
		tw->cur_fg = fg;
		XSetForeground(tw->X->display, tw->X->gc_vt,
			COLOR_PIX(tw, tw->cur_fg));
	}
}

/*------------------------------------------------------------------------
 * set background color
 */
void xvt_w_set_bg	(TERMWIN *tw, int bg, int force)
{
	if (force || bg != tw->cur_bg)
	{
		tw->cur_bg = bg;
		XSetBackground(tw->X->display, tw->X->gc_vt,
			COLOR_PIX(tw, tw->cur_bg));
	}
}

/*------------------------------------------------------------------------
 * draw the window border & clean margin area
 */
static void xvt_w_draw_border (TERMWIN *tw)
{
	int			x1;
	int			y1;
	int			x2;
	int			y2;
	XRectangle	r[4];

	/*--------------------------------------------------------------------
	 * do the border
	 */
	if (tw->border > 0)
	{
		x1	= 0;
		y1	= 0;
		x2	= tw->X->vt_width  - tw->border;
		y2	= tw->X->vt_height - tw->border;

		/* left border */

		r[0].x		= x1;
		r[0].y		= y1;
		r[0].width	= tw->border;
		r[0].height	= tw->X->vt_height;

		/* top border */

		r[1].x		= x1;
		r[1].y		= y1;
		r[1].width	= tw->X->vt_width;
		r[1].height	= tw->border;

		/* right border */

		r[2].x		= x2;
		r[2].y		= y1;
		r[2].width	= tw->border;
		r[2].height	= tw->X->vt_height;

		/* bottom border */

		r[3].x		= x1;
		r[3].y		= y2;
		r[3].width	= tw->X->vt_width;
		r[3].height	= tw->border;

		XSetForeground(tw->X->display, tw->X->gc_vt,
			COLOR_PIX(tw, XVT_CLR_BD));
		XSetBackground(tw->X->display, tw->X->gc_vt,
			COLOR_PIX(tw, XVT_CLR_BD));

		XFillRectangles(tw->X->display, tw->X->win_vt, tw->X->gc_vt, r, 4);

		XSetForeground(tw->X->display, tw->X->gc_vt,
			COLOR_PIX(tw, tw->cur_fg));
		XSetBackground(tw->X->display, tw->X->gc_vt,
			COLOR_PIX(tw, tw->cur_bg));
	}

	/*--------------------------------------------------------------------
	 * do the margin
	 */
	if (tw->margin > 0)
	{
		x1	= tw->border;
		y1	= tw->border;
		x2	= tw->X->vt_width  - tw->extra;
		y2	= tw->X->vt_height - tw->extra;

		/* left side */

		r[0].x		= x1;
		r[0].y		= y1;
		r[0].width	= tw->margin;
		r[0].height	= tw->X->vt_height;

		/* top side */

		r[1].x		= x1;
		r[1].y		= y1;
		r[1].width	= tw->X->vt_width;
		r[1].height	= tw->margin;

		/* right side */

		r[2].x		= x2;
		r[2].y		= y1;
		r[2].width	= tw->margin;
		r[2].height	= tw->X->vt_height;

		/* bottom side */

		r[3].x		= x1;
		r[3].y		= y2;
		r[3].width	= tw->X->vt_width;
		r[3].height	= tw->margin;

		XSetForeground(tw->X->display, tw->X->gc_vt,
			COLOR_PIX(tw, XVT_CLR_BG));
		XSetBackground(tw->X->display, tw->X->gc_vt,
			COLOR_PIX(tw, XVT_CLR_BG));

		XFillRectangles(tw->X->display, tw->X->win_vt, tw->X->gc_vt, r, 4);

		XSetForeground(tw->X->display, tw->X->gc_vt,
			COLOR_PIX(tw, tw->cur_fg));
		XSetBackground(tw->X->display, tw->X->gc_vt,
			COLOR_PIX(tw, tw->cur_bg));
	}
}

/*------------------------------------------------------------------------
 * decode keyboard state
 */
static int xvt_w_get_kbd_state (TERMWIN *tw, char *keys)
{
	int	kbd_state	= KEY_MASK_NONE;
	int i;

	for (i=0; i<32; i++)
	{
		if (keys[i] != 0)
		{
			int c = (unsigned char)(keys[i]);
			int j;

			for (j=0; j<8; j++)
			{
				int b = c & (0x01 << j);

				if (b != 0)
				{
					int		m = (i * 8) + j;
					KeySym	k = XKeycodeToKeysym(tw->X->display, m, 0);

					switch (k)
					{
					case XK_Shift_L:
					case XK_Shift_R:	kbd_state |= KEY_MASK_SHFT;	break;

					case XK_Control_L:
					case XK_Control_R:	kbd_state |= KEY_MASK_CTRL;	break;

					case XK_Meta_L:
					case XK_Meta_R:
					case XK_Alt_L:
					case XK_Alt_R:		kbd_state |= KEY_MASK_META;	break;

					case XK_Num_Lock:	kbd_state |= KEY_MASK_NUML;	break;
					}
				}
			}
		}
	}

	return (kbd_state);
}

/*------------------------------------------------------------------------
 * get current state of keyboard prefix keys
 */
int xvt_w_query_kbd (TERMWIN *tw)
{
	char	keys[32];
	int		kbd_state;

	/*--------------------------------------------------------------------
	 * get key states
	 */
	XQueryKeymap(tw->X->display, keys);

	/*--------------------------------------------------------------------
	 * now decode it
	 */
	kbd_state = xvt_w_get_kbd_state(tw, keys);

	return (kbd_state);
}

/*------------------------------------------------------------------------
 * fill a rectangle
 */
void xvt_w_rect_fill (TERMWIN *tw, int x, int y, int w, int h)
{
	XFillRectangle(tw->X->display, tw->X->win_vt, tw->X->gc_vt,
		COL_TO_X(tw, x), ROW_TO_Y(tw, y),
		COL_TO_W(tw, w), ROW_TO_H(tw, h));
}

/*------------------------------------------------------------------------
 * copy a rectangle
 */
void xvt_w_rect_copy (TERMWIN *tw, int dx, int dy,
	int dw, int dh, int sx, int sy)
{
    XCopyArea(tw->X->display, tw->X->win_vt, tw->X->win_vt, tw->X->gc_vt,
        COL_TO_X(tw, sx), ROW_TO_Y(tw, sy),
        COL_TO_W(tw, dw), ROW_TO_H(tw, dh),
        COL_TO_X(tw, dx), ROW_TO_Y(tw, dy));
}

/*------------------------------------------------------------------------
 * draw a rectangle around a character
 */
void xvt_w_char_rect (TERMWIN *tw, int x, int y)
{
	XSetFillStyle (tw->X->display, tw->X->gc_vt, FillSolid);
	XDrawRectangle(tw->X->display, tw->X->win_vt, tw->X->gc_vt,
		COL_TO_X(tw, x),     ROW_TO_Y(tw, y),
		COL_TO_W(tw, 1) - 1, ROW_TO_H(tw, 1) - 1);
	XSetFillStyle (tw->X->display, tw->X->gc_vt, FillOpaqueStippled);
}

/*------------------------------------------------------------------------
 * draw a line under a char
 */
void xvt_w_char_line (TERMWIN *tw, int x, int y)
{
	XSetFillStyle(tw->X->display, tw->X->gc_vt, FillSolid);

	x = COL_TO_X(tw, x) + CHAR_X_OFFSET(tw);
	y = ROW_TO_Y(tw, y) + CHAR_Y_OFFSET(tw) + 1;

	XDrawLine(tw->X->display, tw->X->win_vt, tw->X->gc_vt,
		x, y, x + COL_TO_W(tw, 1), y);

	XSetFillStyle(tw->X->display, tw->X->gc_vt, FillOpaqueStippled);
}

/*------------------------------------------------------------------------
 * draw a char
 */
void xvt_w_char_disp (TERMWIN *tw, int x, int y, int c)
{
	char text[2];

	text[0] = (char)c;
	XDrawImageString(tw->X->display, tw->X->win_vt, tw->X->gc_vt,
		COL_TO_X(tw, x) + CHAR_X_OFFSET(tw),
		ROW_TO_Y(tw, y) + CHAR_Y_OFFSET(tw),
		text, 1);
}

/*------------------------------------------------------------------------
 * draw a char in bold
 */
void xvt_w_char_bold (TERMWIN *tw, int x, int y, int c)
{
	if (tw->X->fonts[tw->cur_font_no].bld_font != 0)
	{
		/*----------------------------------------------------------------
		 * draw char using bold font
		 */
		xvt_w_set_font_type(tw, FONT_BOLD);
		xvt_w_char_disp(tw, x, y, c);
		xvt_w_set_font_type(tw, FONT_NORM);
	}
	else
	{
		/*----------------------------------------------------------------
		 * draw char, then copy the char over 1 pixel
		 */
		xvt_w_char_disp(tw, x, y, c);

		XSetFunction(tw->X->display, tw->X->gc_vt, GXor);
		XCopyArea(tw->X->display, tw->X->win_vt, tw->X->win_vt, tw->X->gc_vt,
			COL_TO_X(tw, x),     ROW_TO_Y(tw, y),
			CHAR_WIDTH(tw) - 1,  CHAR_HEIGHT(tw),
			COL_TO_X(tw, x) + 1, ROW_TO_Y(tw, y));
		XSetFunction(tw->X->display, tw->X->gc_vt, GXcopy);
	}
}

/*------------------------------------------------------------------------
 * translate a key event into a keysym & buffer
 */
static int xvt_w_xlate_key (XKeyEvent *xe, int *kp)
{
	unsigned char	buffer[64];
	KeySym			keysym;
	int				nbytes;

	/*--------------------------------------------------------------------
	 * do initial lookup of key event
	 */
	nbytes = XLookupString(xe, (char *)buffer, sizeof(buffer), &keysym, 0);
	if (nbytes < 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * make sure buffer ends with a NULL
	 */
	buffer[nbytes] = 0;

	/*--------------------------------------------------------------------
	 * make sure keysym is in proper range
	 */
	keysym &= 0xffff;

	/*--------------------------------------------------------------------
	 * It seems that (Sun at least) sends wrong info for F11 & F12,
	 * so we will try to fix it here.
	 *
	 * Also, Linux (& maybe others) mix up the DEL & BS keys, so we will
	 * try to make the BS & DEL key correct here.
	 */
	switch (keysym)
	{
	case 0xff10:
		nbytes = 0;
		keysym = XK_F11;
		break;

	case 0xff11:
		nbytes = 0;
		keysym = XK_F12;
		break;

	case XK_BackSpace:
		nbytes = 1;
		buffer[0] = 0x08;
		buffer[1] = 0;
		break;

#if defined(XK_KP_Add)
	case XK_KP_Add:
#endif
#if defined(XK_KP_Subtract)
	case XK_KP_Subtract:
#endif
#if defined(XK_KP_Multiply)
	case XK_KP_Multiply:
#endif
#if defined(XK_KP_Divide)
	case XK_KP_Divide:
#endif
	case XK_Delete:
		nbytes = 0;
		break;
	}

	/*--------------------------------------------------------------------
	 * store keysym if requested
	 */
	if (kp != 0)
		*kp = keysym;

	/*--------------------------------------------------------------------
	 * check if translation is in buffer
	 */
	if (nbytes > 0)
	{
		return (buffer[0]);
	}

	/*--------------------------------------------------------------------
	 * translate special key to our internal representation
	 */
	switch (keysym)
	{
		/*----------------------------------------------------------------
		 * function keys
		 */
		case XK_F1:				return (XVT_KEY_FN_01);
		case XK_F2:				return (XVT_KEY_FN_02);
		case XK_F3:				return (XVT_KEY_FN_03);
		case XK_F4:				return (XVT_KEY_FN_04);
		case XK_F5:				return (XVT_KEY_FN_05);
		case XK_F6:				return (XVT_KEY_FN_06);
		case XK_F7:				return (XVT_KEY_FN_07);
		case XK_F8:				return (XVT_KEY_FN_08);
		case XK_F9:				return (XVT_KEY_FN_09);
		case XK_F10:			return (XVT_KEY_FN_10);
		case XK_F11:			return (XVT_KEY_FN_11);
		case XK_F12:			return (XVT_KEY_FN_12);
		case XK_F13:			return (XVT_KEY_FN_13);
		case XK_F14:			return (XVT_KEY_FN_14);
		case XK_F15:			return (XVT_KEY_FN_15);
		case XK_F16:			return (XVT_KEY_FN_16);

		/*----------------------------------------------------------------
		 * keypad keys
		 */
		case XK_Insert:			return (XVT_KEY_CS_0);
#if defined(XK_KP_Insert)
		case XK_KP_Insert:		return (XVT_KEY_KP_0);
#endif
		case XK_KP_0:			return (XVT_KEY_NL_0);

		case XK_End:			return (XVT_KEY_CS_1);
#if defined(XK_KP_End)
		case XK_KP_End:			return (XVT_KEY_KP_1);
#endif
		case XK_R13:			return (XVT_KEY_KP_1);
		case XK_KP_1:			return (XVT_KEY_NL_1);

		case XK_Down:			return (XVT_KEY_CS_2);
#if defined(XK_KP_Down)
		case XK_KP_Down:		return (XVT_KEY_KP_2);
#endif
		case XK_R14:			return (XVT_KEY_KP_2);
		case XK_KP_2:			return (XVT_KEY_NL_2);

#if defined(XK_Page_Down)
		case XK_Page_Down:		return (XVT_KEY_CS_3);
#endif
#if defined(XK_KP_Page_Down)
		case XK_KP_Page_Down:	return (XVT_KEY_KP_3);
#endif
		case XK_R15:			return (XVT_KEY_KP_3);
		case XK_KP_3:			return (XVT_KEY_NL_3);

		case XK_Left:			return (XVT_KEY_CS_4);
#if defined(XK_KP_Left)
		case XK_KP_Left:		return (XVT_KEY_KP_4);
#endif
		case XK_R10:			return (XVT_KEY_KP_4);
		case XK_KP_4:			return (XVT_KEY_NL_4);

#if defined(XK_KP_Begin)
		case XK_KP_Begin:		return (XVT_KEY_KP_5);
#endif
		case XK_R11:			return (XVT_KEY_KP_5);
		case XK_KP_5:			return (XVT_KEY_NL_5);

		case XK_Right:			return (XVT_KEY_CS_6);
#if defined(XK_KP_Right)
		case XK_KP_Right:		return (XVT_KEY_KP_6);
#endif
		case XK_R12:			return (XVT_KEY_KP_6);
		case XK_KP_6:			return (XVT_KEY_NL_6);
	
		case XK_Home:			return (XVT_KEY_CS_7);
#if defined(XK_KP_Home)
		case XK_KP_Home:		return (XVT_KEY_KP_7);
#endif
		case XK_R7:				return (XVT_KEY_KP_7);
		case XK_KP_7:			return (XVT_KEY_NL_7);

		case XK_Up:				return (XVT_KEY_CS_8);
#if defined(XK_KP_Up)
		case XK_KP_Up:			return (XVT_KEY_KP_8);
#endif
		case XK_R8:				return (XVT_KEY_KP_8);
		case XK_KP_8:			return (XVT_KEY_NL_8);

#if defined(XK_Page_Up)
		case XK_Page_Up:		return (XVT_KEY_CS_9);
#endif
#if defined(XK_KP_Page_up)
		case XK_KP_Page_Up:		return (XVT_KEY_KP_9);
#endif
		case XK_R9:				return (XVT_KEY_KP_9);
		case XK_KP_9:			return (XVT_KEY_KP_9);

		case XK_Delete:			return (XVT_KEY_CS_DELETE);
#if defined(XK_KP_Delete)
		case XK_KP_Delete:		return (XVT_KEY_KP_DELETE);
#endif
		case XK_KP_Decimal:		return (XVT_KEY_NL_DECIMAL);

#if defined(XK_KP_Subtract)
		case XK_KP_Subtract:	return (XVT_KEY_KP_SUB);
#endif
		case XK_R4:				return (XVT_KEY_KP_SUB);

#if defined(XK_KP_Add)
		case XK_KP_Add:			return (XVT_KEY_KP_ADD);
#endif

#if defined(XK_KP_Divide)
		case XK_KP_Divide:		return (XVT_KEY_KP_DIV);
#endif
		case XK_R5:				return (XVT_KEY_KP_DIV);

#if defined(XK_KP_Multiply)
		case XK_KP_Multiply:	return (XVT_KEY_KP_MUL);
#endif
		case XK_R6:				return (XVT_KEY_KP_MUL);

#if defined (XK_KP_ENTER)
		case XK_KP_ENTER:		return (XVT_KEY_KP_ENTER);
#endif

		/*----------------------------------------------------------------
		 * misc keys
		 */
#if defined(XK_Pause)
		case XK_Pause:
#endif
		case XK_R1:				return (XVT_KEY_MS_PAUSE);

#if defined(XK_Print)
		case XK_Print:
#endif
		case XK_R2:				return (XVT_KEY_MS_PRTSCR);

#if defined(XK_Scroll_Lock)
		case XK_Scroll_Lock:
#endif
		case XK_R3:				return (XVT_KEY_MS_SCRLCK);

#if defined(XK_Sys_Req)
		case XK_Sys_Req:		return (XVT_KEY_MS_SYSREQ);
#endif

#if defined(XK_Break)
		case XK_Break:			return (XVT_KEY_MS_BREAK);
#endif

		/*----------------------------------------------------------------
		 * prefix keys
		 */
		case XK_Shift_L:
		case XK_Shift_R:		return (XVT_KEY_PF_SHFT);

		case XK_Control_L:
		case XK_Control_R:		return (XVT_KEY_PF_CTRL);

		case XK_Meta_L:
		case XK_Meta_R:
		case XK_Alt_L:
		case XK_Alt_R:			return (XVT_KEY_PF_META);

		case XK_Num_Lock:		return (XVT_KEY_PF_NUML);
	}

	return (XVT_KEY_OT_UNKNOWN);
}

/*------------------------------------------------------------------------
 * convert a pointer name to a number
 */
int xvt_w_pointer_value (const char *name)
{
	const CURSOR_TYPES *	cp;

	/*--------------------------------------------------------------------
	 * check for null name
	 */
	if (name == 0 || *name == 0)
		return (-1);

	/*--------------------------------------------------------------------
	 * if string is numeric, just return numeric value
	 */
	if (isdigit(*name))
	{
		int code = atoi(name) & ~1;

		if (code < 0 || code > XC_num_glyphs)
			return (-1);

		return (code);
	}

	/*--------------------------------------------------------------------
	 * lookup name in cursor-name table
	 */
	for (cp = xvt_w_pointers; cp->ptr_name; cp++)
	{
		if (xvt_strccmpc(cp->ptr_name, name) == 0)
			return (cp->ptr_code);
	}

	return (-1);
}

/*------------------------------------------------------------------------
 * get a pointer name/code from a number
 */
const char * xvt_w_pointer_name (int n)
{
	if (n < 0 || n >= NUM_CURSORS)
		return (0);

	return (xvt_w_pointers[n].ptr_name);
}

int xvt_w_pointer_code (int n)
{
	if (n < 0 || n >= NUM_CURSORS)
		return (-1);

	return (xvt_w_pointers[n].ptr_code);
}

/*------------------------------------------------------------------------
 * convert an X event to our event
 */
static void xvt_w_cvt_event (TERMWIN *tw, XEvent*xe, XVT_EVENT *xp)
{
	switch (xe->type)
	{
	/*--------------------------------------------------------------------
	 * Key press/release events
	 */
	case KeyPress:
	case KeyRelease:
		xp->event_type = XVT_E_KEYPRESS;
		{
			XVT_EVENT_KEYPRESS *	xk = &xp->event_data.keypress;
			XKeyEvent *				xt = &xe->xkey;

			/*------------------------------------------------------------
			 * Get the state of the keyboard prefix keys.
			 *
			 * It seems that the Mod3Mask is used to indicate the
			 * state of the NumLock key, but this is not documented.
			 */
											xk->kbd_state  = KEY_MASK_NONE;
			if (xt->state & ShiftMask)		xk->kbd_state |= KEY_MASK_SHFT;
			if (xt->state & ControlMask)	xk->kbd_state |= KEY_MASK_CTRL;
			if (xt->state & Mod1Mask)		xk->kbd_state |= KEY_MASK_META;
			if (xt->state & Mod3Mask)		xk->kbd_state |= KEY_MASK_NUML;

			/*------------------------------------------------------------
			 * cache up/down state
			 */
			xk->key_state = (xt->type == KeyPress);

			/*------------------------------------------------------------
			 * now translate the key value
			 */
			xk->key_value = xvt_w_xlate_key(xt, &xk->key_symbol);
		}
		break;

	/*--------------------------------------------------------------------
	 * Keyboard state changed event
	 */
	case KeymapNotify:
		xp->event_type = XVT_E_KEYMAP;
		{
			XVT_EVENT_KEYMAP *		xk = &xp->event_data.keymap;
			XKeymapEvent *			xt = &xe->xkeymap;

			/*------------------------------------------------------------
			 * This event happens when we get focus, so we don't know
			 * what the state of the keyboard is.
			 */
			xk->kbd_state = xvt_w_get_kbd_state(tw, xt->key_vector);
		}
		break;

	/*--------------------------------------------------------------------
	 * Button press/release events
	 */
	case ButtonPress:
	case ButtonRelease:
		xp->event_type = XVT_E_BUTTON;
		{
			XVT_EVENT_BUTTON *		xk = &xp->event_data.button;
			XButtonEvent *			xt = &xe->xbutton;

			/*------------------------------------------------------------
			 * We are only interested in button events inside the window.
			 */
			if (X_ON_BORDER(tw, xt->x) || Y_ON_BORDER(tw, xt->y))
			{
				xp->event_type = XVT_E_IGNORE;
			}
			else
			{
				switch (xt->button)
				{
				case Button1:	xk->button = BUT_L;	break;
				case Button2:	xk->button = BUT_M;	break;
				case Button3:	xk->button = BUT_R;	break;
				case Button4:	xk->button = BUT_U;	break;
				case Button5:	xk->button = BUT_D;	break;
				}

				xk->btn_state = (xt->type == ButtonPress);

				xk->x	= X_TO_COL(tw, xt->x);
				xk->y	= Y_TO_ROW(tw, xt->y);
			}
		}
		break;

	/*--------------------------------------------------------------------
	 * Pointer movement events
	 */
	case MotionNotify:
		xp->event_type = XVT_E_MOTION;
		{
			XVT_EVENT_MOTION *		xk = &xp->event_data.motion;
			XMotionEvent *			xt = &xe->xmotion;

			/*------------------------------------------------------------
			 * We are only interested in motion events inside the window.
			 */
			if (X_ON_BORDER(tw, xt->x) || Y_ON_BORDER(tw, xt->y))
			{
				xp->event_type = XVT_E_IGNORE;
			}
			else
			{
				xk->x	= X_TO_COL(tw, xt->x);
				xk->y	= Y_TO_ROW(tw, xt->y);
			}
		}
		break;

	/*--------------------------------------------------------------------
	 * Focus change events
	 */
	case FocusIn:
	case FocusOut:
		xp->event_type = XVT_E_FOCUS;
		{
			XVT_EVENT_FOCUS *		xk = &xp->event_data.focus;
			XFocusChangeEvent *		xt = &xe->xfocus;

			xk->focus	= (xt->type == FocusIn);
		}
		break;

	/*--------------------------------------------------------------------
	 * Exposure (window area repaint) events
	 */
	case Expose:
		xp->event_type = XVT_E_EXPOSE;
		{
			XVT_EVENT_EXPOSE *		xk = &xp->event_data.expose;
			XExposeEvent *			xt = &xe->xexpose;
			int						need_border	= FALSE;
			int						x, y;
			int						w, h;

			if (X_ON_L_BORDER(tw, xt->x))
			{
				x = 0;
				need_border = TRUE;
			}
			else
			{
				x = X_TO_COL(tw, xt->x);
			}

			if (Y_ON_T_BORDER(tw, xt->y))
			{
				y = 0;
				need_border = TRUE;
			}
			else
			{
				y = Y_TO_ROW(tw, xt->y);
			}

			if (X_ON_R_BORDER(tw, xt->x + xt->width))
			{
				w = tw->win_cols; 
				need_border = TRUE;
			}
			else
			{
				w = X_TO_COL(tw, xt->x + xt->width  + COL_TO_W(tw, 1));
			}

			if (Y_ON_B_BORDER(tw, xt->y + xt->height))
			{
				h = tw->win_rows;
				need_border = TRUE;
			}
			else
			{
				h = Y_TO_ROW(tw, xt->y + xt->height + ROW_TO_H(tw, 1));
			}

			w = (w - x) + 1;
			h = (h - y) + 1;

			xk->x	= x;
			xk->y	= y;
			xk->w	= w;
			xk->h	= h;

			if (need_border)
				xvt_w_draw_border(tw);
		}
		break;

	/*--------------------------------------------------------------------
	 * Window size change events
	 */
	case ConfigureNotify:
		xp->event_type = XVT_E_CONFIG;
		{
			XVT_EVENT_CONFIG *		xk = &xp->event_data.config;
			XConfigureEvent *		xt = &xe->xconfigure;

			xk->w	= W_TO_COL(tw, xt->width  - (2 * tw->extra) - tw->sb_size);
			xk->h	= H_TO_ROW(tw, xt->height - (2 * tw->extra));

			tw->win_cols = xk->w;
			tw->win_rows = xk->h;

			xvt_w_resize_win(tw);
			xvt_w_draw_border(tw);
		}
		break;

	/*--------------------------------------------------------------------
	 * Iconify/Uniconify events
	 */
	case MapNotify:
	case UnmapNotify:
		xp->event_type = XVT_E_MAP;
		{
			XVT_EVENT_MAP *			xk = &xp->event_data.map;

			xk->mapped	= (xe->type == MapNotify);
		}
		tw->iconified = (xe->type == UnmapNotify);
		break;

	/*--------------------------------------------------------------------
	 * cut/paste events
	 */
	case SelectionClear:
	case SelectionNotify:
	case SelectionRequest:
		break;

	/*--------------------------------------------------------------------
	 * Window manager events
	 */
	case ClientMessage:
		xp->event_type = XVT_E_WINMSG;
		{
			XVT_EVENT_WINMSG *		xk = &xp->event_data.winmsg;
			XClientMessageEvent *	xt = &xe->xclient;

			/*------------------------------------------------------------
			 * These are window manager messages.  Currently, the only
			 * message we are interested in is "WM_DELETE_WINDOW".
			 */
			if (xt->format == 32)
			{
				if (xt->data.l[0] == tw->X->wm_delete)
				{
					xk->winmsg	= XVT_W_DELETE;
				}
				else
				{
					xk->winmsg	= XVT_W_IGNORE;
				}
			}
			else
			{
				xk->winmsg	= XVT_W_IGNORE;
			}
		}
		break;

	/*--------------------------------------------------------------------
	 * Unknown events
	 */
	default:
		xp->event_type = XVT_E_IGNORE;
		break;
	}
}

/*------------------------------------------------------------------------
 * get next event if there is one
 */
int xvt_w_event (TERMWIN *tw, XVT_EVENT *xe)
{
	XEvent	ev;

	if (XPending(tw->X->display) > 0)
	{
		XNextEvent(tw->X->display, &ev);
		xvt_w_cvt_event(tw, &ev, xe);
		return (TRUE);
	}

	return (FALSE);
}

/*------------------------------------------------------------------------
 * copy data to/from the clipboard
 */
#define PROP_DFLT	"CUT_BUFFER0"
#ifndef PROP_NAME
#define PROP_NAME	PROP_DFLT
#endif

static Atom xvt_w_property (TERMWIN *tw, const char *name)
{
	Atom	property;

	property = XInternAtom(tw->X->display, name, TRUE);
	if (property == None)
		property = XInternAtom(tw->X->display, PROP_DFLT, TRUE);

	return (property);
}

static int xvt_w_clipboard_copy_to_name (TERMWIN *tw, const char *prop_name,
	unsigned char *data, int len)
{
	Atom	property = xvt_w_property(tw, prop_name);

	XSetSelectionOwner(tw->X->display, XA_PRIMARY, tw->X->win_vt, CurrentTime);
	if (XGetSelectionOwner(tw->X->display, XA_PRIMARY) != tw->X->win_vt)
		return (-1);

	XChangeProperty(tw->X->display, tw->X->win_root, property,
		XA_STRING, 8, PropModeReplace, data, len);

	return (0);
}

int xvt_w_clipboard_copy_to (TERMWIN *tw, unsigned char *data, int len)
{
	return xvt_w_clipboard_copy_to_name(tw, PROP_NAME, data, len);
}

static unsigned char * xvt_w_clipboard_copy_from_atom (TERMWIN *tw,
	Atom property, int *plen)
{
	Window			win;
	long			nread;
	unsigned long	bytes_after;
	unsigned long	nitems;
	Atom			type;
	int				format;
	unsigned char *	prop;
	unsigned char *	data	= 0;
	int				len		= 0;
	int				old_len	= 0;

	win = XGetSelectionOwner(tw->X->display, XA_PRIMARY);
	if (win == None)
		win = tw->X->win_root;

	for (nread=0, bytes_after=1; bytes_after > 0; nread += nitems)
	{
		nitems = 0;
		XGetWindowProperty(tw->X->display, win, property,
			(nread / 4), 4096, FALSE, AnyPropertyType,
			&type, &format, &nitems, &bytes_after, &prop);

		if (nitems == 0)
			break;

		len += (int)nitems;
		data = xvt_mem_realloc(data, len);
		memcpy(data + old_len, prop, nitems);
		old_len = len;

		XFree(prop);
	}

	if (plen != 0)
		*plen = len;

	return (data);
}

static unsigned char * xvt_w_clipboard_copy_from_name (TERMWIN *tw,
	const char *prop_name, int *plen)
{
	Atom	property = xvt_w_property(tw, prop_name);

	return xvt_w_clipboard_copy_from_atom(tw, property, plen);
}

unsigned char * xvt_w_clipboard_copy_from (TERMWIN *tw, int *plen)
{
	return xvt_w_clipboard_copy_from_name(tw, PROP_NAME, plen);
}
