/*------------------------------------------------------------------------
 * header for X11 window routines
 */
#ifndef XVTX_H
#define XVTX_H

#include <X11/Intrinsic.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <X11/Xatom.h>

/*------------------------------------------------------------------------
 * font-info struct
 */
struct font_info
{
	const char *	reg_fname;				/* reg  font name			*/
	const char *	bld_fname;				/* bold font name			*/

	XFontStruct *	reg_font;				/* ptr to reg  font			*/
	XFontStruct *	bld_font;				/* ptr to bold font or 0	*/

	int				cell_width;				/* width  of font char		*/
	int				cell_height;			/* height of font char		*/

	int				cell_xhome;				/* x-offset of font pos		*/
	int				cell_yhome;				/* y-offset of font pos		*/
};
typedef struct font_info	FONT_INFO;

/*------------------------------------------------------------------------
 * color-table
 */
struct clr_tbl
{
	Pixel			pix;					/* pixel value				*/
	int				got;					/* TRUE if allocated		*/
};
typedef struct clr_tbl	CLR_TBL;

/*------------------------------------------------------------------------
 * x-data struct
 */
struct x_data
{
	/*--------------------------------------------------------------------
	 * display stuff
	 */
	Display *		display;				/* X display				*/
	int				screen;					/* X screen number			*/
	int				xfd;					/* X connection number		*/

	/*--------------------------------------------------------------------
	 * window stuff
	 */
	Window			win_root;				/* X root window			*/
	Window			win_parent;				/* X parent window			*/
	Window			win_vt;					/* X term window			*/
	Window			win_sb;					/* X scrollbar window		*/

	XSizeHints		s_hints;				/* X size hints				*/
	XWMHints		w_hints;				/* X wm hints				*/
	XClassHint		c_hints;				/* X class hints			*/

	/*--------------------------------------------------------------------
	 * other stuff
	 */
	GC				gc_vt;					/* X GC for vt window		*/
	GC				gc_sb;					/* X GC for sb window		*/
	GC				gc_sb_ts;				/* X GC for sb top shadow	*/
	GC				gc_sb_bs;				/* X GC for sb bot shadow	*/

	Cursor			cursor;					/* X Cursor being used		*/
	Pixmap			icon;					/* X icon of window			*/
	Pixmap			fill_stipple;			/* X stipple pixmap in GC	*/
	Atom			wm_delete;				/* X WM delete message		*/

	/*--------------------------------------------------------------------
	 * color stuff
	 */
	Colormap		cmap;					/* X colormap				*/
	CLR_TBL			clr_tbl[CMS_COLORSIZE];	/* color pixel entries		*/

	/*--------------------------------------------------------------------
	 * font stuff
	 */
	FONT_INFO		fonts[XVT_NUM_FONTS];	/* font info structs		*/

	/*--------------------------------------------------------------------
	 * window info
	 */
	int				vt_x;					/* vt x position			*/
	int				vt_y;					/* vt y position			*/
	int				vt_width;				/* vt width  in pixels		*/
	int				vt_height;				/* vt height in pixels		*/

	int				sb_x;					/* sb x position			*/
	int				sb_y;					/* sb y position			*/
	int				sb_width;				/* sb width  in pixels		*/
	int				sb_height;				/* sb height in pixels		*/
};

/*------------------------------------------------------------------------
 * misc macros
 */
#define COLOR_PIX(tw, n)		((tw)->X->clr_tbl[n].pix)
#define COLOR_GOT(tw, n)		((tw)->X->clr_tbl[n].got)
#define COLOR_SET(tw, n, p, g)	((tw)->X->clr_tbl[n].pix = (p), \
								 (tw)->X->clr_tbl[n].got = (g))

#define CHAR_WIDTH(tw)			((tw)->X->fonts[(tw)->cur_font_no].cell_width)
#define CHAR_HEIGHT(tw)			((tw)->X->fonts[(tw)->cur_font_no].cell_height)

#define CHAR_X_OFFSET(tw)		((tw)->X->fonts[(tw)->cur_font_no].cell_xhome)
#define CHAR_Y_OFFSET(tw)		((tw)->X->fonts[(tw)->cur_font_no].cell_yhome)

#define W_TO_COL(tw, w)			((w) / CHAR_WIDTH (tw))
#define H_TO_ROW(tw, h)			((h) / CHAR_HEIGHT(tw))

#define COL_TO_W(tw, x)			((x) * CHAR_WIDTH (tw))
#define ROW_TO_H(tw, y)			((y) * CHAR_HEIGHT(tw))

#define COL_TO_X(tw, x)			(COL_TO_W(tw, (x)) + (tw)->extra)
#define ROW_TO_Y(tw, y)			(ROW_TO_H(tw, (y)) + (tw)->extra)

#define X_TO_COL(tw, x)			(W_TO_COL(tw, (x) - (tw)->extra))
#define Y_TO_ROW(tw, y)			(H_TO_ROW(tw, (y) - (tw)->extra))

#define X_ON_L_BORDER(tw, x)	((x) <  (tw)->border)
#define X_ON_R_BORDER(tw, x)	((x) >= ((tw)->X->vt_width  - (tw)->extra))

#define Y_ON_T_BORDER(tw, y)	((y) <  (tw)->border)
#define Y_ON_B_BORDER(tw, y)	((y) >= ((tw)->X->vt_height - (tw)->extra))

#define X_ON_BORDER(tw, x)		(X_ON_L_BORDER(tw, x) || X_ON_R_BORDER(tw, x))
#define Y_ON_BORDER(tw, y)		(Y_ON_T_BORDER(tw, y) || Y_ON_B_BORDER(tw, y))

#endif /* XVTX_H */
